@@ -14,6 +14,7 @@ struct window_params {
     const char *title;
     int width;
     int height;
+    const char *forced_impl;
 
     // initial color space
     struct pl_swapchain_colors colors;
@@ -42,10 +43,13 @@ void window_get_scroll(const struct window *win, float *dx, float *dy);
 bool window_get_button(const struct window *win, enum button);
 bool window_get_key(const struct window *win, enum key);
 char *window_get_file(const struct window *win);
+bool window_toggle_fullscreen(const struct window *win, bool fullscreen);
+bool window_is_fullscreen(const struct window *win);
 
 // For implementations
 struct window_impl {
     const char *name;
+    const char *tag;
     __typeof__(window_create) *create;
     __typeof__(window_destroy) *destroy;
     __typeof__(window_poll) *poll;
@@ -54,4 +58,6 @@ struct window_impl {
     __typeof__(window_get_button) *get_button;
     __typeof__(window_get_key) *get_key;
     __typeof__(window_get_file) *get_file;
+    __typeof__(window_toggle_fullscreen) *toggle_fullscreen;
+    __typeof__(window_is_fullscreen) *is_fullscreen;
 };
