@@ -16,18 +16,21 @@
 #define GLFW_INCLUDE_VULKAN
 #define IMPL win_impl_glfw_vk
 #define IMPL_NAME "GLFW (vulkan)"
+#define IMPL_TAG "glfw-vk"
 #endif
 
 #ifdef USE_GL
 #include <libplacebo/opengl.h>
 #define IMPL win_impl_glfw_gl
 #define IMPL_NAME "GLFW (opengl)"
+#define IMPL_TAG "glfw-gl"
 #endif
 
 #ifdef USE_D3D11
 #include <libplacebo/d3d11.h>
 #define IMPL win_impl_glfw_d3d11
 #define IMPL_NAME "GLFW (D3D11)"
+#define IMPL_TAG "glfw-d3d11"
 #endif
 
 #include <GLFW/glfw3.h>
@@ -42,6 +45,10 @@
 #include <GLFW/glfw3native.h>
 #endif
 
+#ifdef _WIN32
+#define strdup _strdup
+#endif
+
 #ifdef NDEBUG
 #define DEBUG false
 #else
@@ -52,6 +59,13 @@
 
 const struct window_impl IMPL;
 
+struct window_pos {
+    int x;
+    int y;
+    int w;
+    int h;
+};
+
 struct priv {
     struct window w;
     GLFWwindow *win;
@@ -75,6 +89,8 @@ struct priv {
     size_t files_num;
     size_t files_size;
     bool file_seen;
+
+    struct window_pos windowed_pos;
 };
 
 static void err_cb(int code, const char *desc)
@@ -333,7 +349,11 @@ static struct window *glfw_create(pl_log log, const struct window_params *params
     p->w.gpu = p->d3d11->gpu;
 #endif // USE_D3D11
 
-    int w = params->width, h = params->height;
+    glfwGetWindowSize(p->win, &p->windowed_pos.w, &p->windowed_pos.h);
+    glfwGetWindowPos(p->win, &p->windowed_pos.x, &p->windowed_pos.y);
+
+    int w, h;
+    glfwGetFramebufferSize(p->win, &w, &h);
     pl_swapchain_colorspace_hint(p->w.swapchain, &params->colors);
     if (!pl_swapchain_resize(p->w.swapchain, &w, &h)) {
         fprintf(stderr, "libplacebo: Failed initializing swapchain\n");
@@ -450,8 +470,45 @@ static char *glfw_get_file(const struct window *window)
     return p->files[0];
 }
 
+static bool glfw_is_fullscreen(const struct window *window) {
+    const struct priv *p = (const struct priv *) window;
+    return glfwGetWindowMonitor(p->win);
+}
+
+static bool glfw_toggle_fullscreen(const struct window *window, bool fullscreen)
+{
+    struct priv *p = (struct priv *) window;
+    bool window_fullscreen = glfw_is_fullscreen(window);
+
+    if (window_fullscreen == fullscreen)
+        return true;
+
+    if (window_fullscreen) {
+        glfwSetWindowMonitor(p->win, NULL, p->windowed_pos.x, p->windowed_pos.y,
+                             p->windowed_pos.w, p->windowed_pos.h, GLFW_DONT_CARE);
+        return true;
+    }
+
+    // For simplicity sake use primary monitor
+    GLFWmonitor *monitor = glfwGetPrimaryMonitor();
+    if (!monitor)
+        return false;
+
+    const GLFWvidmode *mode = glfwGetVideoMode(monitor);
+    if (!mode)
+        return false;
+
+    glfwGetWindowPos(p->win, &p->windowed_pos.x, &p->windowed_pos.y);
+    glfwGetWindowSize(p->win, &p->windowed_pos.w, &p->windowed_pos.h);
+    glfwSetWindowMonitor(p->win, monitor, 0, 0, mode->width, mode->height,
+                         mode->refreshRate);
+
+    return true;
+}
+
 const struct window_impl IMPL = {
     .name = IMPL_NAME,
+    .tag = IMPL_TAG,
     .create = glfw_create,
     .destroy = glfw_destroy,
     .poll = glfw_poll,
@@ -460,4 +517,6 @@ const struct window_impl IMPL = {
     .get_key = glfw_get_key,
     .get_scroll = glfw_get_scroll,
     .get_file = glfw_get_file,
+    .toggle_fullscreen = glfw_toggle_fullscreen,
+    .is_fullscreen = glfw_is_fullscreen,
 };
