@@ -16,6 +16,7 @@
 #define WINFLAG_API SDL_WINDOW_VULKAN
 #define IMPL win_impl_sdl_vk
 #define IMPL_NAME "SDL2 (vulkan)"
+#define IMPL_TAG "sdl2-vk"
 #endif
 
 #ifdef USE_GL
@@ -23,6 +24,7 @@
 #define WINFLAG_API SDL_WINDOW_OPENGL
 #define IMPL win_impl_sdl_gl
 #define IMPL_NAME "SDL2 (opengl)"
+#define IMPL_TAG "sdl2-gl"
 #endif
 
 #ifdef NDEBUG
@@ -89,6 +91,8 @@ static struct window *sdl_create(pl_log log, const struct window_params *params)
         goto error;
     }
 
+    int w, h;
+
 #ifdef USE_VK
 
     unsigned int num = 0;
@@ -139,6 +143,8 @@ static struct window *sdl_create(pl_log log, const struct window_params *params)
     }
 
     p->w.gpu = p->vk->gpu;
+
+    SDL_Vulkan_GetDrawableSize(p->win, &w, &h);
 #endif // USE_VK
 
 #ifdef USE_GL
@@ -153,6 +159,7 @@ static struct window *sdl_create(pl_log log, const struct window_params *params)
         .debug = DEBUG,
         .make_current = make_current,
         .release_current = release_current,
+        .get_proc_addr = (void *) SDL_GL_GetProcAddress,
         .priv = p,
     ));
     if (!p->gl) {
@@ -171,9 +178,10 @@ static struct window *sdl_create(pl_log log, const struct window_params *params)
     }
 
     p->w.gpu = p->gl->gpu;
+
+    SDL_GL_GetDrawableSize(p->win, &w, &h);
 #endif // USE_GL
 
-    int w = params->width, h = params->height;
     pl_swapchain_colorspace_hint(p->w.swapchain, &params->colors);
     if (!pl_swapchain_resize(p->w.swapchain, &w, &h)) {
         fprintf(stderr, "libplacebo: Failed initializing swapchain\n");
@@ -326,8 +334,47 @@ static char *sdl_get_file(const struct window *window)
     return p->files[0];
 }
 
+static bool sdl_is_fullscreen(const struct window *window)
+{
+    const struct priv *p = (const struct priv *) window;
+    return SDL_GetWindowFlags(p->win) & SDL_WINDOW_FULLSCREEN;
+}
+
+static bool sdl_toggle_fullscreen(const struct window *window, bool fullscreen)
+{
+    struct priv *p = (struct priv *) window;
+    bool window_fullscreen = sdl_is_fullscreen(window);
+
+    if (window_fullscreen == fullscreen)
+        return true;
+
+    SDL_DisplayMode mode;
+    if (SDL_GetDesktopDisplayMode(0, &mode))
+    {
+        fprintf(stderr, "SDL2: Failed to get display mode: %s\n", SDL_GetError());
+        SDL_ClearError();
+        return false;
+    }
+
+    if (SDL_SetWindowDisplayMode(p->win, &mode))
+    {
+        fprintf(stderr, "SDL2: Failed to set window display mode: %s\n", SDL_GetError());
+        SDL_ClearError();
+        return false;
+    }
+
+    if (SDL_SetWindowFullscreen(p->win, fullscreen ? SDL_WINDOW_FULLSCREEN : 0)) {
+        fprintf(stderr, "SDL2: SetWindowFullscreen failed: %s\n", SDL_GetError());
+        SDL_ClearError();
+        return false;
+    }
+
+    return true;
+}
+
 const struct window_impl IMPL = {
     .name = IMPL_NAME,
+    .tag = IMPL_TAG,
     .create = sdl_create,
     .destroy = sdl_destroy,
     .poll = sdl_poll,
@@ -336,4 +383,6 @@ const struct window_impl IMPL = {
     .get_key = sdl_get_key,
     .get_scroll = sdl_get_scroll,
     .get_file = sdl_get_file,
+    .toggle_fullscreen = sdl_toggle_fullscreen,
+    .is_fullscreen = sdl_is_fullscreen,
 };
