@@ -233,6 +233,9 @@ enum pl_color_primaries pl_color_primaries_guess(int width, int height)
     }
 }
 
+// HLG 75% value (scene-referred)
+#define HLG_75 3.17955
+
 float pl_color_transfer_nominal_peak(enum pl_color_transfer trc)
 {
     switch (trc) {
@@ -250,7 +253,7 @@ float pl_color_transfer_nominal_peak(enum pl_color_transfer trc)
     case PL_COLOR_TRC_ST428:
         return 1.0;
     case PL_COLOR_TRC_PQ:       return 10000.0 / PL_COLOR_SDR_WHITE;
-    case PL_COLOR_TRC_HLG:      return 12.0 / PL_COLOR_SDR_WHITE_HLG;
+    case PL_COLOR_TRC_HLG:      return 12.0 / HLG_75;
     case PL_COLOR_TRC_V_LOG:    return 46.0855;
     case PL_COLOR_TRC_S_LOG1:   return 6.52;
     case PL_COLOR_TRC_S_LOG2:   return 9.212;
@@ -260,22 +263,6 @@ float pl_color_transfer_nominal_peak(enum pl_color_transfer trc)
     pl_unreachable();
 }
 
-bool pl_color_light_is_scene_referred(enum pl_color_light light)
-{
-    switch (light) {
-    case PL_COLOR_LIGHT_UNKNOWN:
-    case PL_COLOR_LIGHT_DISPLAY:
-        return false;
-    case PL_COLOR_LIGHT_SCENE_HLG:
-    case PL_COLOR_LIGHT_SCENE_709_1886:
-    case PL_COLOR_LIGHT_SCENE_1_2:
-        return true;
-    case PL_COLOR_LIGHT_COUNT: break;
-    }
-
-    pl_unreachable();
-}
-
 const struct pl_hdr_metadata pl_hdr_metadata_empty = {0};
 const struct pl_hdr_metadata pl_hdr_metadata_hdr10 ={
     .prim = {
@@ -303,6 +290,8 @@ float pl_hdr_rescale(enum pl_hdr_scaling from, enum pl_hdr_scaling to, float x)
     if (!x) // micro-optimization for common value
         return x;
 
+    x = fmaxf(x, 0.0f);
+
     // Convert input to PL_SCALE_RELATIVE
     switch (from) {
     case PL_HDR_PQ:
@@ -401,7 +390,7 @@ void pl_hdr_metadata_merge(struct pl_hdr_metadata *orig,
 bool pl_hdr_metadata_contains(const struct pl_hdr_metadata *data,
                               enum pl_hdr_metadata_type type)
 {
-    bool has_hdr10 = data->min_luma && data->max_luma;
+    bool has_hdr10 = data->max_luma;
     bool has_hdr10plus = data->scene_avg && (data->scene_max[0] ||
                                              data->scene_max[1] ||
                                              data->scene_max[2]);
@@ -513,7 +502,7 @@ static void luma_from_maxrgb(const struct pl_color_space *csp,
 
     struct pl_raw_primaries prim = csp->hdr.prim;
     pl_raw_primaries_merge(&prim, pl_raw_primaries_get(csp->primaries));
-    const struct pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(&prim);
+    const pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(&prim);
 
     const float max_luma = rgb2xyz.m[1][0] * csp->hdr.scene_max[0] +
                            rgb2xyz.m[1][1] * csp->hdr.scene_max[1] +
@@ -524,6 +513,12 @@ static void luma_from_maxrgb(const struct pl_color_space *csp,
     *out_avg = pl_hdr_rescale(PL_HDR_NITS, scaling, coef * csp->hdr.scene_avg);
 }
 
+static inline bool metadata_compat(enum pl_hdr_metadata_type metadata,
+                                   enum pl_hdr_metadata_type compat)
+{
+    return metadata == PL_HDR_METADATA_ANY || metadata == compat;
+}
+
 void pl_color_space_nominal_luma_ex(const struct pl_nominal_luma_params *params)
 {
     if (!params || (!params->out_min && !params->out_max && !params->out_avg))
@@ -532,84 +527,58 @@ void pl_color_space_nominal_luma_ex(const struct pl_nominal_luma_params *params)
     const struct pl_color_space *csp = params->color;
     const enum pl_hdr_scaling scaling = params->scaling;
 
-    // Baseline/fallback metadata, inferred entirely from the colorspace
-    // description and built-in default assumptions
-    if (params->metadata == PL_HDR_METADATA_NONE) {
-        float min_luma, max_luma; // in nits
-        max_luma = pl_color_transfer_nominal_peak(csp->transfer) *
-                   PL_COLOR_SDR_WHITE;
-
-        // Exception: For HLG content, we want to infer a value of 1000 cd/m²,
-        // a value which is considered the "reference" HLG display.
-        if (csp->transfer == PL_COLOR_TRC_HLG)
-            max_luma = 1000;
-
-        if (pl_color_transfer_is_hdr(csp->transfer)) {
-            // Use a slightly nonzero black level, for the following reasons:
-            // - 0 may be seen as 'missing/undefined' in HDR10 metadata structs
-            //
-            // - we need to clamp to 1e-7 to avoid issues with some tone
-            //   mapping functions anyway, so doing it here also advertises
-            //   this fact downstream
-            //
-            // - true infinite contrast does not exist in reality, even 1e-7
-            //   is extremely generous considering typical viewing environments
-            //   are not absolutely devoid of stray ambient light
-            min_luma = 1e-7;
-        } else {
-            min_luma = max_luma / 1000; // Typical SDR contrast
-        }
-
-        if (params->out_min)
-            *params->out_min = pl_hdr_rescale(PL_HDR_NITS, scaling, min_luma);
-        if (params->out_max)
-            *params->out_max = pl_hdr_rescale(PL_HDR_NITS, scaling, max_luma);
-        if (params->out_avg)
-            *params->out_avg = 0;
-        return;
+    float min_luma = 0, max_luma = 0, avg_luma = 0;
+    if (params->metadata != PL_HDR_METADATA_NONE) {
+        // Initialize from static HDR10 metadata, in all cases
+        min_luma = pl_hdr_rescale(PL_HDR_NITS, scaling, csp->hdr.min_luma);
+        max_luma = pl_hdr_rescale(PL_HDR_NITS, scaling, csp->hdr.max_luma);
     }
 
-    // Always take the minimum luminance from static mastering display
-    // metadata, because DV/HDR10+ don't encode this at all.
-    float min_luma = pl_hdr_rescale(PL_HDR_NITS, scaling, csp->hdr.min_luma);
-
-    // PQ is always scaled down to absolute black, ignoring HDR metadata
-    if (csp->transfer == PL_COLOR_TRC_PQ)
-        min_luma = 0;
-
-    float max_luma = 0, avg_luma = 0;
-    if (!params->metadata || params->metadata == PL_HDR_METADATA_HDR10) {
-        // This metadata is very often completely broken
-        if (csp->hdr.max_luma >= 100.0f)
-            max_luma = pl_hdr_rescale(PL_HDR_NITS, scaling, csp->hdr.max_luma);
-    }
-
-    if ((!params->metadata || params->metadata == PL_HDR_METADATA_HDR10PLUS) &&
+    if (metadata_compat(params->metadata, PL_HDR_METADATA_HDR10PLUS) &&
         pl_hdr_metadata_contains(&csp->hdr, PL_HDR_METADATA_HDR10PLUS))
     {
         luma_from_maxrgb(csp, scaling, &max_luma, &avg_luma);
     }
 
-    if ((!params->metadata || params->metadata == PL_HDR_METADATA_CIE_Y) &&
+    if (metadata_compat(params->metadata, PL_HDR_METADATA_CIE_Y) &&
         pl_hdr_metadata_contains(&csp->hdr, PL_HDR_METADATA_CIE_Y))
     {
         max_luma = pl_hdr_rescale(PL_HDR_PQ, scaling, csp->hdr.max_pq_y);
         avg_luma = pl_hdr_rescale(PL_HDR_PQ, scaling, csp->hdr.avg_pq_y);
     }
 
-    if (min_luma < 0)
-        min_luma = 0; // sanity
-    if (max_luma && min_luma > max_luma) // sanity
-        min_luma = max_luma = 0;
+    // Clamp to sane value range
+    const float hdr_min = pl_hdr_rescale(PL_HDR_NITS, scaling, PL_COLOR_HDR_BLACK);
+    const float hdr_max = pl_hdr_rescale(PL_HDR_PQ,   scaling, 1.0f);
+    max_luma = max_luma ? PL_CLAMP(max_luma, hdr_min, hdr_max) : 0;
+    min_luma = min_luma ? PL_CLAMP(min_luma, hdr_min, hdr_max) : 0;
+    if ((max_luma && min_luma >= max_luma) || min_luma >= hdr_max)
+        min_luma = max_luma = 0; // sanity
 
-    // Make sure this metadata is complete by inferring with fixed defaults
-    pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
-        .color      = csp,
-        .metadata   = PL_HDR_METADATA_NONE,
-        .scaling    = scaling,
-        .out_min    = min_luma ? NULL : &min_luma,
-        .out_max    = max_luma ? NULL : &max_luma,
-    ));
+    // PQ is always scaled down to absolute black, ignoring HDR metadata
+    if (csp->transfer == PL_COLOR_TRC_PQ)
+        min_luma = hdr_min;
+
+    // Baseline/fallback metadata, inferred entirely from the colorspace
+    // description and built-in default assumptions
+    if (!max_luma) {
+        if (csp->transfer == PL_COLOR_TRC_HLG) {
+            max_luma = pl_hdr_rescale(PL_HDR_NITS, scaling, PL_COLOR_HLG_PEAK);
+        } else {
+            const float peak = pl_color_transfer_nominal_peak(csp->transfer);
+            max_luma = pl_hdr_rescale(PL_HDR_NORM, scaling, peak);
+        }
+    }
+
+    if (!min_luma) {
+        if (pl_color_transfer_is_hdr(csp->transfer)) {
+            min_luma = hdr_min;
+        } else {
+            const float peak = pl_hdr_rescale(scaling, PL_HDR_NITS, max_luma);
+            min_luma = pl_hdr_rescale(PL_HDR_NITS, scaling,
+                                      peak / PL_COLOR_SDR_CONTRAST);
+        }
+    }
 
     if (avg_luma)
         avg_luma = PL_CLAMP(avg_luma, min_luma, max_luma); // sanity
@@ -974,9 +943,9 @@ const struct pl_raw_primaries *pl_raw_primaries_get(enum pl_color_primaries prim
 
 // Compute the RGB/XYZ matrix as described here:
 // http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
-struct pl_matrix3x3 pl_get_rgb2xyz_matrix(const struct pl_raw_primaries *prim)
+pl_matrix3x3 pl_get_rgb2xyz_matrix(const struct pl_raw_primaries *prim)
 {
-    struct pl_matrix3x3 out = {{{0}}};
+    pl_matrix3x3 out = {{{0}}};
     float S[3], X[4], Z[4];
 
     // Convert from CIE xyY to XYZ. Note that Y=1 holds true for all primaries
@@ -1012,17 +981,17 @@ struct pl_matrix3x3 pl_get_rgb2xyz_matrix(const struct pl_raw_primaries *prim)
     return out;
 }
 
-struct pl_matrix3x3 pl_get_xyz2rgb_matrix(const struct pl_raw_primaries *prim)
+pl_matrix3x3 pl_get_xyz2rgb_matrix(const struct pl_raw_primaries *prim)
 {
     // For simplicity, just invert the rgb2xyz matrix
-    struct pl_matrix3x3 out = pl_get_rgb2xyz_matrix(prim);
+    pl_matrix3x3 out = pl_get_rgb2xyz_matrix(prim);
     pl_matrix3x3_invert(&out);
     return out;
 }
 
 // LMS<-XYZ revised matrix from CIECAM97, based on a linear transform and
 // normalized for equal energy on monochrome inputs
-static const struct pl_matrix3x3 m_cat97 = {{
+static const pl_matrix3x3 m_cat97 = {{
     {  0.8562,  0.3372, -0.1934 },
     { -0.8360,  1.8327,  0.0033 },
     {  0.0357, -0.0469,  1.0112 },
@@ -1031,7 +1000,7 @@ static const struct pl_matrix3x3 m_cat97 = {{
 // M := M * XYZd<-XYZs
 static void apply_chromatic_adaptation(struct pl_cie_xy src,
                                        struct pl_cie_xy dest,
-                                       struct pl_matrix3x3 *mat)
+                                       pl_matrix3x3 *mat)
 {
     // If the white points are nearly identical, this is a wasteful identity
     // operation.
@@ -1056,27 +1025,27 @@ static void apply_chromatic_adaptation(struct pl_cie_xy src,
     }
 
     // tmp := I * [Cd/Cs] * Ma
-    struct pl_matrix3x3 tmp = {0};
+    pl_matrix3x3 tmp = {0};
     for (int i = 0; i < 3; i++)
         tmp.m[i][i] = C[i][1] / C[i][0];
 
     pl_matrix3x3_mul(&tmp, &m_cat97);
 
     // M := M * Ma^-1 * tmp
-    struct pl_matrix3x3 ma_inv = m_cat97;
+    pl_matrix3x3 ma_inv = m_cat97;
     pl_matrix3x3_invert(&ma_inv);
     pl_matrix3x3_mul(mat, &ma_inv);
     pl_matrix3x3_mul(mat, &tmp);
 }
 
-struct pl_matrix3x3 pl_get_adaptation_matrix(struct pl_cie_xy src, struct pl_cie_xy dst)
+pl_matrix3x3 pl_get_adaptation_matrix(struct pl_cie_xy src, struct pl_cie_xy dst)
 {
     // Use BT.709 primaries (with chosen white point) as an XYZ reference
     struct pl_raw_primaries csp = *pl_raw_primaries_get(PL_COLOR_PRIM_BT_709);
     csp.white = src;
 
-    struct pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(&csp);
-    struct pl_matrix3x3 xyz2rgb = rgb2xyz;
+    pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(&csp);
+    pl_matrix3x3 xyz2rgb = rgb2xyz;
     pl_matrix3x3_invert(&xyz2rgb);
 
     apply_chromatic_adaptation(src, dst, &xyz2rgb);
@@ -1084,6 +1053,53 @@ struct pl_matrix3x3 pl_get_adaptation_matrix(struct pl_cie_xy src, struct pl_cie
     return xyz2rgb;
 }
 
+pl_matrix3x3 pl_ipt_rgb2lms(const struct pl_raw_primaries *prim)
+{
+    static const pl_matrix3x3 hpe = {{ // HPE XYZ->LMS (D65) method
+        {  0.40024f, 0.70760f, -0.08081f },
+        { -0.22630f, 1.16532f,  0.04570f },
+        {  0.00000f, 0.00000f,  0.91822f },
+    }};
+
+    const float c = 0.04; // 4% crosstalk
+    pl_matrix3x3 m = {{
+        { 1 - 2*c,       c,       c },
+        {       c, 1 - 2*c,       c },
+        {       c,       c, 1 - 2*c },
+    }};
+
+    pl_matrix3x3_mul(&m, &hpe);
+
+    // Apply chromatic adaptation to D65 if the input white point differs
+    static const struct pl_cie_xy d65 = CIE_D65;
+    apply_chromatic_adaptation(prim->white, d65, &m);
+
+    const pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(prim);
+    pl_matrix3x3_mul(&m, &rgb2xyz);
+    return m;
+}
+
+pl_matrix3x3 pl_ipt_lms2rgb(const struct pl_raw_primaries *prim)
+{
+    pl_matrix3x3 m = pl_ipt_rgb2lms(prim);
+    pl_matrix3x3_invert(&m);
+    return m;
+}
+
+// As standardized in Ebner & Fairchild IPT (1998)
+const pl_matrix3x3 pl_ipt_lms2ipt = {{
+    { 0.4000,  0.4000,  0.2000 },
+    { 4.4550, -4.8510,  0.3960 },
+    { 0.8056,  0.3572, -1.1628 },
+}};
+
+// Numerically inverted from the matrix above
+const pl_matrix3x3 pl_ipt_ipt2lms = {{
+    { 1.0,  0.0975689,  0.205226 },
+    { 1.0, -0.1138760,  0.133217 },
+    { 1.0,  0.0326151, -0.676887 },
+}};
+
 const struct pl_cone_params pl_vision_normal        = {PL_CONE_NONE, 1.0};
 const struct pl_cone_params pl_vision_protanomaly   = {PL_CONE_L,    0.5};
 const struct pl_cone_params pl_vision_protanopia    = {PL_CONE_L,    0.0};
@@ -1094,12 +1110,12 @@ const struct pl_cone_params pl_vision_tritanopia    = {PL_CONE_S,    0.0};
 const struct pl_cone_params pl_vision_monochromacy  = {PL_CONE_LM,   0.0};
 const struct pl_cone_params pl_vision_achromatopsia = {PL_CONE_LMS,  0.0};
 
-struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
-                                       const struct pl_raw_primaries *prim)
+pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
+                                const struct pl_raw_primaries *prim)
 {
     // LMS<-RGB := LMS<-XYZ * XYZ<-RGB
-    struct pl_matrix3x3 rgb2lms = m_cat97;
-    struct pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(prim);
+    pl_matrix3x3 rgb2lms = m_cat97;
+    pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(prim);
     pl_matrix3x3_mul(&rgb2lms, &rgb2xyz);
 
     // LMS versions of the two opposing primaries, plus neutral
@@ -1112,7 +1128,7 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
     pl_matrix3x3_apply(&rgb2lms, lms_w);
 
     float a, b, c = params->strength;
-    struct pl_matrix3x3 distort;
+    pl_matrix3x3 distort;
 
     switch (params->cones) {
     case PL_CONE_NONE:
@@ -1126,7 +1142,7 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
             (lms_b[2] - lms_b[1] * lms_w[2] / lms_w[1]);
         assert(fabs(a * lms_w[1] + b * lms_w[2] - lms_w[0]) < 1e-6);
 
-        distort = (struct pl_matrix3x3) {{
+        distort = (pl_matrix3x3) {{
             {            c, (1.0 - c) * a, (1.0 - c) * b},
             {          0.0,           1.0,           0.0},
             {          0.0,           0.0,           1.0},
@@ -1141,7 +1157,7 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
             (lms_b[2] - lms_b[0] * lms_w[2] / lms_w[0]);
         assert(fabs(a * lms_w[0] + b * lms_w[2] - lms_w[1]) < 1e-6);
 
-        distort = (struct pl_matrix3x3) {{
+        distort = (pl_matrix3x3) {{
             {          1.0,           0.0,           0.0},
             {(1.0 - c) * a,             c, (1.0 - c) * b},
             {          0.0,           0.0,           1.0},
@@ -1156,7 +1172,7 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
             (lms_r[1] - lms_r[0] * lms_w[1] / lms_w[0]);
         assert(fabs(a * lms_w[0] + b * lms_w[1] - lms_w[2]) < 1e-6);
 
-        distort = (struct pl_matrix3x3) {{
+        distort = (pl_matrix3x3) {{
             {          1.0,           0.0,           0.0},
             {          0.0,           1.0,           0.0},
             {(1.0 - c) * a, (1.0 - c) * b,             c},
@@ -1168,7 +1184,7 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
         a = lms_w[0] / lms_w[2];
         b = lms_w[1] / lms_w[2];
 
-        distort = (struct pl_matrix3x3) {{
+        distort = (pl_matrix3x3) {{
             {            c,           0.0, (1.0 - c) * a},
             {          0.0,             c, (1.0 - c) * b},
             {          0.0,           0.0,           1.0},
@@ -1180,7 +1196,7 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
         a = lms_w[1] / lms_w[0];
         b = lms_w[2] / lms_w[0];
 
-        distort = (struct pl_matrix3x3) {{
+        distort = (pl_matrix3x3) {{
             {          1.0,           0.0,           0.0},
             {(1.0 - c) * a,             c,           0.0},
             {(1.0 - c) * b,           0.0,             c},
@@ -1192,7 +1208,7 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
         a = lms_w[0] / lms_w[1];
         b = lms_w[2] / lms_w[1];
 
-        distort = (struct pl_matrix3x3) {{
+        distort = (pl_matrix3x3) {{
             {            c, (1.0 - c) * a,           0.0},
             {          0.0,           1.0,           0.0},
             {          0.0, (1.0 - c) * b,             c},
@@ -1221,7 +1237,7 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
     }
 
     // out := RGB<-LMS * distort * LMS<-RGB
-    struct pl_matrix3x3 out = rgb2lms;
+    pl_matrix3x3 out = rgb2lms;
     pl_matrix3x3_invert(&out);
     pl_matrix3x3_mul(&out, &distort);
     pl_matrix3x3_mul(&out, &rgb2lms);
@@ -1229,9 +1245,9 @@ struct pl_matrix3x3 pl_get_cone_matrix(const struct pl_cone_params *params,
     return out;
 }
 
-struct pl_matrix3x3 pl_get_color_mapping_matrix(const struct pl_raw_primaries *src,
-                                                const struct pl_raw_primaries *dst,
-                                                enum pl_rendering_intent intent)
+pl_matrix3x3 pl_get_color_mapping_matrix(const struct pl_raw_primaries *src,
+                                         const struct pl_raw_primaries *dst,
+                                         enum pl_rendering_intent intent)
 {
     // In saturation mapping, we don't care about accuracy and just want
     // primaries to map to primaries, making this an identity transformation.
@@ -1244,14 +1260,14 @@ struct pl_matrix3x3 pl_get_color_mapping_matrix(const struct pl_raw_primaries *s
     // definition for perceptual other than "make it look good".
 
     // RGBd<-XYZd matrix
-    struct pl_matrix3x3 xyz2rgb_d = pl_get_xyz2rgb_matrix(dst);
+    pl_matrix3x3 xyz2rgb_d = pl_get_xyz2rgb_matrix(dst);
 
     // Chromatic adaptation, except in absolute colorimetric intent
     if (intent != PL_INTENT_ABSOLUTE_COLORIMETRIC)
         apply_chromatic_adaptation(src->white, dst->white, &xyz2rgb_d);
 
     // XYZs<-RGBs
-    struct pl_matrix3x3 rgb2xyz_s = pl_get_rgb2xyz_matrix(src);
+    pl_matrix3x3 rgb2xyz_s = pl_get_rgb2xyz_matrix(src);
     pl_matrix3x3_mul(&xyz2rgb_d, &rgb2xyz_s);
     return xyz2rgb_d;
 }
@@ -1315,10 +1331,10 @@ bool pl_primaries_valid(const struct pl_raw_primaries *prim)
  * Under these conditions the given parameters lr, lg, lb uniquely
  * determine the mapping of Y, U, V to R, G, B.
  */
-static struct pl_matrix3x3 luma_coeffs(float lr, float lg, float lb)
+static pl_matrix3x3 luma_coeffs(float lr, float lg, float lb)
 {
     pl_assert(fabs(lr+lg+lb - 1) < 1e-6);
-    return (struct pl_matrix3x3) {{
+    return (pl_matrix3x3) {{
         {1, 0,                    2 * (1-lr)          },
         {1, -2 * (1-lb) * lb/lg, -2 * (1-lr) * lr/lg  },
         {1,  2 * (1-lb),          0                   },
@@ -1326,7 +1342,7 @@ static struct pl_matrix3x3 luma_coeffs(float lr, float lg, float lb)
 }
 
 // Applies hue and saturation controls to a YCbCr->RGB matrix
-static inline void apply_hue_sat(struct pl_matrix3x3 *m,
+static inline void apply_hue_sat(pl_matrix3x3 *m,
                                  const struct pl_color_adjustment *params)
 {
     // Hue is equivalent to rotating input [U, V] subvector around the origin.
@@ -1340,12 +1356,12 @@ static inline void apply_hue_sat(struct pl_matrix3x3 *m,
     }
 }
 
-struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
-                                    const struct pl_color_adjustment *params)
+pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
+                                     const struct pl_color_adjustment *params)
 {
     params = PL_DEF(params, &pl_color_adjustment_neutral);
 
-    struct pl_matrix3x3 m;
+    pl_matrix3x3 m;
     switch (repr->sys) {
     case PL_COLOR_SYSTEM_BT_709:     m = luma_coeffs(0.2126, 0.7152, 0.0722); break;
     case PL_COLOR_SYSTEM_BT_601:     m = luma_coeffs(0.2990, 0.5870, 0.1140); break;
@@ -1353,7 +1369,7 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
     case PL_COLOR_SYSTEM_BT_2020_NC: m = luma_coeffs(0.2627, 0.6780, 0.0593); break;
     case PL_COLOR_SYSTEM_BT_2020_C:
         // Note: This outputs into the [-0.5,0.5] range for chroma information.
-        m = (struct pl_matrix3x3) {{
+        m = (pl_matrix3x3) {{
             {0, 0, 1},
             {1, 0, 0},
             {0, 1, 0},
@@ -1364,7 +1380,7 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         // and precision reasons. Exact values truncated from ITU-T H-series
         // Supplement 18.
         static const float lm_t = 0.008609, lm_p = 0.111029625;
-        m = (struct pl_matrix3x3) {{
+        m = (pl_matrix3x3) {{
             {1.0,  lm_t,  lm_p},
             {1.0, -lm_t, -lm_p},
             {1.0, 0.560031, -0.320627},
@@ -1374,7 +1390,7 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
     case PL_COLOR_SYSTEM_BT_2100_HLG: {
         // Similar to BT.2100 PQ, exact values truncated from WolframAlpha
         static const float lm_t = 0.01571858011, lm_p = 0.2095810681;
-        m = (struct pl_matrix3x3) {{
+        m = (pl_matrix3x3) {{
             {1.0,  lm_t,  lm_p},
             {1.0, -lm_t, -lm_p},
             {1.0, 1.02127108, -0.605274491},
@@ -1385,7 +1401,7 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         m = repr->dovi->nonlinear;
         break;
     case PL_COLOR_SYSTEM_YCGCO:
-        m = (struct pl_matrix3x3) {{
+        m = (pl_matrix3x3) {{
             {1,  -1,  1},
             {1,   1,  0},
             {1,  -1, -1},
@@ -1413,8 +1429,8 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         apply_hue_sat(&m, params);
     } else if (params->saturation != 1.0 || params->hue != 0.0) {
         // Arbitrarily simulate hue shifts using the BT.709 YCbCr model
-        struct pl_matrix3x3 yuv2rgb = luma_coeffs(0.2126, 0.7152, 0.0722);
-        struct pl_matrix3x3 rgb2yuv = yuv2rgb;
+        pl_matrix3x3 yuv2rgb = luma_coeffs(0.2126, 0.7152, 0.0722);
+        pl_matrix3x3 rgb2yuv = yuv2rgb;
         pl_matrix3x3_invert(&rgb2yuv);
         apply_hue_sat(&yuv2rgb, params);
         // M := RGB<-YUV * YUV<-RGB * M
@@ -1426,11 +1442,11 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
     if (params->temperature) {
         struct pl_cie_xy src = pl_white_from_temp(6500);
         struct pl_cie_xy dst = pl_white_from_temp(6500 + 3500 * params->temperature);
-        struct pl_matrix3x3 adapt = pl_get_adaptation_matrix(src, dst);
+        pl_matrix3x3 adapt = pl_get_adaptation_matrix(src, dst);
         pl_matrix3x3_rmul(&adapt, &m);
     }
 
-    struct pl_transform3x3 out = { .mat = m };
+    pl_transform3x3 out = { .mat = m };
     int bit_depth = PL_DEF(repr->bits.sample_depth,
                     PL_DEF(repr->bits.color_depth, 8));
 
