@@ -26,74 +26,39 @@
 #include <inttypes.h>
 
 #if defined(__MINGW32__) && !defined(__clang__)
-#define PL_PRINTF(fmt, va) __attribute__ ((format(gnu_printf, fmt, va)))
+#define PL_PRINTF(fmt, va) __attribute__ ((format(gnu_printf, fmt, va))) \
+                           __attribute__ ((nonnull))
 #elif defined(__GNUC__)
-#define PL_PRINTF(fmt, va) __attribute__ ((format(printf, fmt, va)))
+#define PL_PRINTF(fmt, va) __attribute__ ((format(printf, fmt, va))) \
+                           __attribute__ ((nonnull))
 #else
 #define PL_PRINTF(fmt, va)
 #endif
 
-#ifdef __unix__
-#define PL_HAVE_UNIX
-#endif
+#define PL_NOINLINE __attribute__((noinline))
 
-#ifdef _WIN32
-#define PL_HAVE_WIN32
-#endif
+#include "os.h"
 
 #include "config_internal.h"
 #include "pl_assert.h"
 #include "pl_alloc.h"
+#include "pl_clock.h"
 #include "pl_string.h"
 
-// Include all of the symbols that should be public in a way that marks them
-// as being externally visible. (Otherwise, all symbols are hidden by default)
-#pragma GCC visibility push(default)
+#define PL_DEPRECATED
 
 #include <libplacebo/config.h>
-#undef PL_DEPRECATED
-#define PL_DEPRECATED
 
 #if PL_API_VER != BUILD_API_VER
 #error Header mismatch? <libplacebo/config.h> pulled from elsewhere!
 #endif
 
-#include <libplacebo/colorspace.h>
-#include <libplacebo/common.h>
-#include <libplacebo/log.h>
-#include <libplacebo/dispatch.h>
-#include <libplacebo/dither.h>
-#include <libplacebo/dummy.h>
-#include <libplacebo/filters.h>
-#include <libplacebo/gpu.h>
-#include <libplacebo/renderer.h>
-#include <libplacebo/shaders.h>
-#include <libplacebo/shaders/colorspace.h>
-#include <libplacebo/shaders/custom.h>
-#include <libplacebo/shaders/film_grain.h>
-#include <libplacebo/shaders/icc.h>
-#include <libplacebo/shaders/lut.h>
-#include <libplacebo/shaders/sampling.h>
-#include <libplacebo/swapchain.h>
-#include <libplacebo/tone_mapping.h>
-#include <libplacebo/utils/frame_queue.h>
-#include <libplacebo/utils/upload.h>
-
-#ifdef PL_HAVE_VULKAN
-#include <libplacebo/vulkan.h>
-#endif
-#ifdef PL_HAVE_OPENGL
-#include <libplacebo/opengl.h>
-#endif
-#ifdef PL_HAVE_D3D11
-#include <libplacebo/d3d11.h>
-#endif
-
-#pragma GCC visibility pop
+// Divide a number while rounding up (careful: double-eval)
+#define PL_DIV_UP(x, y) (((x) + (y) - 1) / (y))
 
 // Align up to the nearest multiple of an arbitrary alignment, which may also
 // be 0 to signal no alignment requirements.
-#define PL_ALIGN(x, align) ((align) ? ((x) + (align) - 1) / (align) * (align) : (x))
+#define PL_ALIGN(x, align) ((align) ? PL_DIV_UP(x, align) * (align) : (x))
 
 // This is faster but must only be called on positive powers of two.
 #define PL_ALIGN2(x, align) (((x) + (align) - 1) & ~((align) - 1))
@@ -104,6 +69,9 @@
 // Rounds a number up to the nearest power of two
 #define PL_ALIGN_POT(x) (0x1LLU << (PL_LOG2((x) - 1) + 1))
 
+// Right shift a number while rounding up
+#define PL_RSHIFT_UP(x, s) -((-(x)) >> (s))
+
 // Returns whether or not a number is a power of two (or zero)
 #define PL_ISPOT(x) (((x) & ((x) - 1)) == 0)
 
@@ -111,11 +79,11 @@
 #define PL_ARRAY_SIZE(s) (sizeof(s) / sizeof((s)[0]))
 
 // Swaps two variables
-#define PL_SWAP(a, b)             \
-    do {                          \
-        __typeof__ (a) tmp = (a); \
-        (a) = (b);                \
-        (b) = tmp;                \
+#define PL_SWAP(a, b)              \
+    do {                           \
+        __typeof__ (a) _tmp = (a); \
+        (a) = (b);                 \
+        (b) = _tmp;                \
     } while (0)
 
 // Helper functions for transposing a matrix in-place.
@@ -138,14 +106,22 @@ static inline float *pl_transpose(int dim, float *out, const float *in)
 
 // Helper functions for some common numeric operations (careful: double-eval)
 #define PL_MAX(x, y) ((x) > (y) ? (x) : (y))
+#define PL_MAX3(x, y, z) PL_MAX(PL_MAX(x, y), z)
 #define PL_MIN(x, y) ((x) < (y) ? (x) : (y))
 #define PL_CLAMP(x, l, h) ((x) < (l) ? (l) : (x) > (h) ? (h) : (x))
-#define PL_CMP(a, b) ((a) < (b) ? -1 : (a) > (b) ? 1 : 0)
+#define PL_CMP(a, b) (((a) > (b)) - ((a) < (b)))
 #define PL_DEF(x, d) ((x) ? (x) : (d))
 #define PL_SQUARE(x) ((x) * (x))
 #define PL_CUBE(x) ((x) * (x) * (x))
 #define PL_MIX(a, b, x) ((x) * (b) + (1 - (x)) * (a))
 
+static inline float pl_smoothstep(float edge0, float edge1, float x)
+{
+    x = (x - edge0) / (edge1 - edge0);
+    x = PL_CLAMP(x, 0.0f, 1.0f);
+    return x * x * (3.0f - 2.0f * x);
+}
+
 // Helpers for doing alignment calculations
 static inline size_t pl_gcd(size_t x, size_t y)
 {
@@ -165,22 +141,14 @@ static inline size_t pl_lcm(size_t x, size_t y)
     return x * (y / pl_gcd(x, y));
 }
 
-// Error checking macro for stuff with integer errors, aborts on failure
-#define PL_CHECK_ERR(expr)                                                      \
-    do {                                                                        \
-        int _ret = (expr);                                                      \
-        if (_ret) {                                                             \
-            fprintf(stderr, "libplacebo: internal error: %s (%s:%d)\n",         \
-                    strerror(_ret), __FILE__, __LINE__);                        \
-            abort();                                                            \
-        }                                                                       \
-    } while (0)
-
 // Conditional abort() macro that depends on the configuration option
 #ifdef PL_DEBUG_ABORT
-# define pl_debug_abort() abort()
+# define pl_debug_abort() do {                          \
+    fprintf(stderr, "pl_debug_abort() triggered!\n");   \
+    abort();                                            \
+} while (0)
 #else
-# define pl_debug_abort()
+# define pl_debug_abort() do {} while (0)
 #endif
 
 // Refcounting helpers
@@ -191,3 +159,15 @@ typedef _Atomic uint32_t pl_rc_t;
 #define pl_rc_count(rc)  atomic_load(rc)
 
 #define pl_unreachable() (assert(!"unreachable"), __builtin_unreachable())
+
+// Helper for parameter validation
+#define pl_require(ctx, expr)                                   \
+  do {                                                          \
+      if (!(expr)) {                                            \
+          PL_ERR(ctx, "Validation failed: %s (%s:%d)",          \
+                  #expr, __FILE__, __LINE__);                   \
+          pl_log_stack_trace(ctx->log, PL_LOG_ERR);             \
+          pl_debug_abort();                                     \
+          goto error;                                           \
+      }                                                         \
+  } while (0)
