@@ -15,6 +15,7 @@
  * License along with libplacebo. If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <initguid.h>
 #include <windows.h>
 #include <versionhelpers.h>
 
@@ -274,7 +275,6 @@ static void d3d11_gpu_destroy(pl_gpu gpu)
 
     pl_buf_destroy(gpu, &p->finish_buf_src);
     pl_buf_destroy(gpu, &p->finish_buf_dst);
-    pl_dispatch_destroy(&p->dp);
 
     // Release everything except the immediate context
     SAFE_RELEASE(p->dev);
@@ -298,9 +298,11 @@ static void d3d11_gpu_destroy(pl_gpu gpu)
 
     // Destroy the immediate context synchronously so referenced objects don't
     // show up in the leak check
-    ID3D11DeviceContext_ClearState(p->imm);
-    ID3D11DeviceContext_Flush(p->imm);
-    SAFE_RELEASE(p->imm);
+    if (p->imm) {
+        ID3D11DeviceContext_ClearState(p->imm);
+        ID3D11DeviceContext_Flush(p->imm);
+        SAFE_RELEASE(p->imm);
+    }
 
     pl_free((void *) gpu);
 }
@@ -397,11 +399,15 @@ pl_gpu pl_gpu_create_d3d11(struct d3d11_ctx *ctx)
     gpu->log = ctx->log;
 
     struct pl_gpu_d3d11 *p = PL_PRIV(gpu);
+    uint32_t spirv_ver = PL_MIN(SPV_VERSION, PL_MAX_SPIRV_VER);
     *p = (struct pl_gpu_d3d11) {
         .ctx = ctx,
         .impl = pl_fns_d3d11,
         .dev = ctx->dev,
-        .spirv = spirv_compiler_create(ctx->log),
+        .spirv = spirv_compiler_create(ctx->log, &(const struct pl_spirv_version) {
+                                          .env_version = pl_spirv_version_to_vulkan(spirv_ver),
+                                          .spv_version = spirv_ver,
+                                      }),
         .vbuf.bind_flags = D3D11_BIND_VERTEX_BUFFER,
         .ibuf.bind_flags = D3D11_BIND_INDEX_BUFFER,
     };
@@ -494,8 +500,14 @@ pl_gpu pl_gpu_create_d3d11(struct d3d11_ctx *ctx)
     }
 
     if (p->fl >= D3D_FEATURE_LEVEL_10_0) {
-        gpu->limits.max_buffer_texels =
-            1 << D3D11_REQ_BUFFER_RESOURCE_TEXEL_COUNT_2_TO_EXP;
+        // Direct3D 11 runtime only guarantees that allocations within the
+        // maximum resource size are supported, but an app can try to allocate
+        // more memory for a resource than the maximum resource size specifies.
+        // That is, the Direct3D 11 runtime allows these memory allocation
+        // attempts in the event that the hardware might support them.
+        // Instead hardcoding limit here, allow it to be resolved during runtime.
+        gpu->limits.max_buffer_texels = UINT32_MAX;
+        // Use UINT32_MAX for parity with Vulkan.
     }
 
     if (p->fl >= D3D_FEATURE_LEVEL_11_0) {
@@ -634,7 +646,7 @@ pl_gpu pl_gpu_create_d3d11(struct d3d11_ctx *ctx)
     if (p->dev5 && p->has_monitored_fences) {
         hr = ID3D11Device5_CreateFence(p->dev5, 0, D3D11_FENCE_FLAG_NONE,
                                        &IID_ID3D11Fence,
-                                       (void **) p->finish_fence);
+                                       (void **) &p->finish_fence);
         if (SUCCEEDED(hr)) {
             p->finish_event = CreateEventW(NULL, FALSE, FALSE, NULL);
             if (!p->finish_event) {
@@ -662,15 +674,13 @@ pl_gpu pl_gpu_create_d3d11(struct d3d11_ctx *ctx)
             &(D3D11_QUERY_DESC) { D3D11_QUERY_EVENT }, &p->finish_query));
     }
 
-    // Create the dispatch last, after any setup of `gpu` is done
-    p->dp = pl_dispatch_create(ctx->log, gpu);
-
     pl_d3d11_flush_message_queue(ctx, "After gpu create");
 
     success = true;
 error:
     SAFE_RELEASE(dxgi_dev);
     SAFE_RELEASE(adapter);
+    SAFE_RELEASE(adapter4);
     if (success) {
         return pl_gpu_finalize(gpu);
     } else {
