@@ -40,6 +40,10 @@
 #include "filters.h"
 #include "log.h"
 
+#ifdef PL_HAVE_WIN32
+#define j1 _j1
+#endif
+
 bool pl_filter_function_eq(const struct pl_filter_function *a,
                            const struct pl_filter_function *b)
 {
@@ -102,7 +106,7 @@ double pl_filter_sample(const struct pl_filter_config *c, double x)
 
 // Compute a single row of weights for a given filter in one dimension, indexed
 // by the indicated subpixel offset. Writes `f->row_size` values to `out`.
-static void compute_row(struct pl_filter *f, double offset, float *out)
+static void compute_row(struct pl_filter_t *f, double offset, float *out)
 {
     double wsum = 0.0;
     for (int i = 0; i < f->row_size; i++) {
@@ -144,7 +148,7 @@ pl_filter pl_filter_generate(pl_log log, const struct pl_filter_params *params)
         return NULL;
     }
 
-    struct pl_filter *f = pl_zalloc_ptr(NULL, f);
+    struct pl_filter_t *f = pl_zalloc_ptr(NULL, f);
     f->params = *params;
     f->params.config.kernel = dupfilter(f, params->config.kernel);
     f->params.config.window = dupfilter(f, params->config.window);
@@ -674,17 +678,6 @@ const struct pl_filter_config pl_filter_ewa_hann = {
     .name   = "ewa_hann",
 };
 
-const struct pl_filter_config pl_filter_haasnsoft = {
-    .kernel = &jinc3,
-    .window = &pl_filter_function_hann,
-    // The blur is tuned to equal out orthogonal and diagonal contributions
-    // on a regular grid. This has the effect of almost completely killing
-    // aliasing.
-    .blur   = 1.11,
-    .polar  = true,
-    .name   = "haasnsoft",
-};
-
 // Spline family
 const struct pl_filter_config pl_filter_bicubic = {
     .kernel = &pl_filter_function_bicubic,
