@@ -20,6 +20,10 @@
 #include <stdlib.h>
 #include <stdbool.h>
 
+typedef struct TLimits TLimits;
+typedef struct TBuiltInResource TBuiltInResource;
+#include <glslang/Include/ResourceLimits.h>
+
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -42,9 +46,12 @@ struct pl_glslang_res {
 // Compile GLSL into a SPIRV stream, if possible. The resulting
 // pl_glslang_res can simply be freed with pl_free() when done.
 struct pl_glslang_res *pl_glslang_compile(const struct pl_glsl_version *glsl,
+                                          const struct pl_spirv_version *spirv_ver,
                                           enum glsl_shader_stage stage,
                                           const char *shader);
 
+extern const TBuiltInResource DefaultTBuiltInResource;
+
 #ifdef __cplusplus
 }
 #endif
