@@ -20,6 +20,9 @@
 #include "common.h"
 #include "log.h"
 
+#include <libplacebo/gpu.h>
+#include <libplacebo/dispatch.h>
+
 // To avoid having to include drm_fourcc.h
 #ifndef DRM_FORMAT_MOD_LINEAR
 #define DRM_FORMAT_MOD_LINEAR   UINT64_C(0x0)
@@ -31,6 +34,13 @@
 
 #define GPU_PFN(name) __typeof__(pl_##name) *name
 struct pl_gpu_fns {
+    // This is a pl_dispatch used (on the pl_gpu itself!) for the purposes of
+    // dispatching compute shaders for performing various emulation tasks (e.g.
+    // partial clears, blits or emulated texture transfers, see below).
+    //
+    // Warning: Care must be taken to avoid recursive calls.
+    pl_dispatch dp;
+
     // Destructors: These also free the corresponding objects, but they
     // must not be called on NULL. (The NULL checks are done by the pl_*_destroy
     // wrappers)
@@ -83,6 +93,9 @@ static inline bool pl_gpu_supports_interop(pl_gpu gpu)
            gpu->import_caps.sync;
 }
 
+// Returns the GPU-internal `pl_dispatch` object.
+pl_dispatch pl_gpu_dispatch(pl_gpu gpu);
+
 // GPU-internal helpers: these should not be used outside of GPU implementations
 
 // This performs several tasks. It sorts the format list, logs GPU metadata,
@@ -109,19 +122,37 @@ bool pl_tex_upload_pbo(pl_gpu gpu, const struct pl_tex_transfer_params *params);
 bool pl_tex_download_pbo(pl_gpu gpu, const struct pl_tex_transfer_params *params);
 
 // This requires that params.buf has been set and is of type PL_BUF_TEXEL_*
-bool pl_tex_upload_texel(pl_gpu gpu, pl_dispatch dp,
-                         const struct pl_tex_transfer_params *params);
-bool pl_tex_download_texel(pl_gpu gpu, pl_dispatch dp,
-                           const struct pl_tex_transfer_params *params);
+bool pl_tex_upload_texel(pl_gpu gpu, const struct pl_tex_transfer_params *params);
+bool pl_tex_download_texel(pl_gpu gpu, const struct pl_tex_transfer_params *params);
 
 // Both `src` and `dst must be storable. `src` must also be sampleable, if the
 // blit requires linear sampling. Returns false if these conditions are unmet.
-bool pl_tex_blit_compute(pl_gpu gpu, pl_dispatch dp,
-                         const struct pl_tex_blit_params *params);
+bool pl_tex_blit_compute(pl_gpu gpu, const struct pl_tex_blit_params *params);
 
 // Helper to do a 2D blit with stretch and scale using a raster pass
-void pl_tex_blit_raster(pl_gpu gpu, pl_dispatch dp,
-                        const struct pl_tex_blit_params *params);
+void pl_tex_blit_raster(pl_gpu gpu, const struct pl_tex_blit_params *params);
+
+// Helper for GPU-accelerated endian swapping
+//
+// Note: `src` and `dst` can be the same buffer, for an in-place operation. In
+// this case, `src_offset` and `dst_offset` must be the same.
+struct pl_buf_copy_swap_params {
+    // Source of the copy operation. Must be `storable`.
+    pl_buf src;
+    size_t src_offset;
+
+    // Destination of the copy operation. Must be `storable`.
+    pl_buf dst;
+    size_t dst_offset;
+
+    // Number of bytes to copy. Must be a multiple of 4.
+    size_t size;
+
+    // Underlying word size. Must be 2 (for 16-bit swap) or 4 (for 32-bit swap)
+    int wordsize;
+};
+
+bool pl_buf_copy_swap(pl_gpu gpu, const struct pl_buf_copy_swap_params *params);
 
 void pl_pass_run_vbo(pl_gpu gpu, const struct pl_pass_run_params *params);
 
@@ -142,27 +173,7 @@ static inline size_t pl_index_buf_size(const struct pl_pass_run_params *params)
 }
 
 // Helper to compute the size of a vertex buffer required to fit all indices
-static inline size_t pl_vertex_buf_size(const struct pl_pass_run_params *params)
-{
-    if (!params->index_data)
-        return params->vertex_count * params->pass->params.vertex_stride;
-
-    int num_vertices = 0;
-    const void *idx = params->index_data;
-    switch (params->index_fmt) {
-    case PL_INDEX_UINT16:
-        for (int i = 0; i < params->vertex_count; i++)
-            num_vertices = PL_MAX(num_vertices, ((const uint16_t *) idx)[i]);
-        break;
-    case PL_INDEX_UINT32:
-        for (int i = 0; i < params->vertex_count; i++)
-            num_vertices = PL_MAX(num_vertices, ((const uint32_t *) idx)[i]);
-        break;
-    case PL_INDEX_FORMAT_COUNT: pl_unreachable();
-    }
-
-    return (num_vertices + 1) * params->pass->params.vertex_stride;
-}
+size_t pl_vertex_buf_size(const struct pl_pass_run_params *params);
 
 // Utility function for pretty-printing UUIDs
 #define UUID_SIZE 16
