@@ -339,10 +339,6 @@ static void add_format(pl_gpu pgpu, const struct gl_format *gl_fmt)
     if (fmt->caps & PL_FMT_CAP_SAMPLEABLE)
         fmt->gatherable = p->gather_comps >= fmt->num_components;
 
-    // Mask renderable/blittable if no FBOs available
-    if (!p->has_fbos)
-        fmt->caps &= ~(PL_FMT_CAP_RENDERABLE | PL_FMT_CAP_BLITTABLE);
-
     // Reading from textures on GLES requires FBO support for this fmt
     if (p->has_readback && (p->gl_ver || (fmt->caps & PL_FMT_CAP_RENDERABLE)))
         fmt->caps |= PL_FMT_CAP_HOST_READABLE;
