@@ -22,13 +22,13 @@
 #include <sys/time.h>
 #include <time.h>
 
+#include <pl_assert.h>
+
 typedef pthread_mutex_t pl_mutex;
 typedef pthread_cond_t  pl_cond;
-
-enum pl_mutex_type {
-    PL_MUTEX_NORMAL = 0,
-    PL_MUTEX_RECURSIVE,
-};
+typedef pthread_mutex_t pl_static_mutex;
+typedef pthread_t       pl_thread;
+#define PL_STATIC_MUTEX_INITIALIZER PTHREAD_MUTEX_INITIALIZER
 
 static inline int pl_mutex_init_type_internal(pl_mutex *mutex, enum pl_mutex_type mtype)
 {
@@ -60,7 +60,7 @@ static inline int pl_mutex_init_type_internal(pl_mutex *mutex, enum pl_mutex_typ
 }
 
 #define pl_mutex_init_type(mutex, mtype) \
-    PL_CHECK_ERR(pl_mutex_init_type_internal(mutex, mtype))
+    pl_assert(!pl_mutex_init_type_internal(mutex, mtype))
 
 #define pl_mutex_destroy    pthread_mutex_destroy
 #define pl_mutex_lock       pthread_mutex_lock
@@ -107,10 +107,19 @@ static inline int pl_cond_timedwait(pl_cond *cond, pl_mutex *mutex, uint64_t tim
     ts.tv_sec  += timeout / 1000000000LLU;
     ts.tv_nsec += timeout % 1000000000LLU;
 
-    if (ts.tv_nsec > 1000000000LLU) {
-        ts.tv_nsec -= 1000000000LLU;
+    if (ts.tv_nsec > 1000000000L) {
+        ts.tv_nsec -= 1000000000L;
         ts.tv_sec++;
     }
 
     return pthread_cond_timedwait(cond, mutex, &ts);
 }
+
+#define pl_static_mutex_lock    pthread_mutex_lock
+#define pl_static_mutex_unlock  pthread_mutex_unlock
+
+#define PL_THREAD_VOID void *
+#define PL_THREAD_RETURN() return NULL
+
+#define pl_thread_create(t, f, a) pthread_create(t, NULL, f, a)
+#define pl_thread_join(t)         pthread_join(t, NULL)
