@@ -22,34 +22,107 @@
 #include "log.h"
 #include "shaders.h"
 
+pl_shader_info pl_shader_info_ref(pl_shader_info pinfo)
+{
+    struct sh_info *info = (struct sh_info *) pinfo;
+    if (!info)
+        return NULL;
+
+    pl_rc_ref(&info->rc);
+    return &info->info;
+}
+
+void pl_shader_info_deref(pl_shader_info *pinfo)
+{
+    struct sh_info *info = (struct sh_info *) *pinfo;
+    if (!info)
+        return;
+
+    if (pl_rc_deref(&info->rc))
+        pl_free(info);
+    *pinfo = NULL;
+}
+
+static struct sh_info *sh_info_alloc(void *alloc)
+{
+    struct sh_info *info = pl_zalloc_ptr(alloc, info);
+    info->tmp = pl_tmp(info);
+    pl_rc_init(&info->rc);
+    return info;
+}
+
+// Re-use `sh_info` allocation if possible, allocate new otherwise
+static struct sh_info *sh_info_recycle(struct sh_info *info)
+{
+    if (!pl_rc_deref(&info->rc))
+        return sh_info_alloc(NULL);
+
+    memset(&info->info, 0, sizeof(info->info)); // reset public fields
+    pl_free_children(info->tmp);
+    pl_rc_ref(&info->rc);
+    info->desc.len = 0;
+    info->steps.num = 0;
+    return info;
+}
+
+static uint8_t reverse_bits(uint8_t x)
+{
+    static const uint8_t reverse_nibble[16] = {
+        0x0, 0x8, 0x4, 0xc, 0x2, 0xa, 0x6, 0xe,
+        0x1, 0x9, 0x5, 0xd, 0x3, 0xb, 0x7, 0xf,
+    };
+
+    return reverse_nibble[x & 0xF] << 4 | reverse_nibble[x >> 4];
+}
+
+static void init_shader(pl_shader sh, const struct pl_shader_params *params)
+{
+    if (params) {
+        sh->info->info.params = *params;
+
+        // To avoid collisions for shaders with very high number of
+        // identifiers, pack the shader ID into the highest bits (MSB -> LSB)
+        pl_static_assert(sizeof(sh->prefix) > sizeof(params->id));
+        const int shift = 8 * (sizeof(sh->prefix) - sizeof(params->id));
+        sh->prefix = reverse_bits(params->id) << shift;
+    }
+
+    sh->name = sh_fresh(sh, "main");
+}
+
 pl_shader pl_shader_alloc(pl_log log, const struct pl_shader_params *params)
 {
+    static const int glsl_ver_req = 130;
+    if (params && params->glsl.version && params->glsl.version < 130) {
+        pl_err(log, "Requested GLSL version %d too low (required: %d)",
+               params->glsl.version, glsl_ver_req);
+        return NULL;
+    }
+
     pl_shader sh = pl_alloc_ptr(NULL, sh);
     *sh = (struct pl_shader_t) {
-        .log = log,
-        .mutable = true,
+        .log        = log,
+        .tmp        = pl_tmp(sh),
+        .info       = sh_info_alloc(NULL),
+        .mutable    = true,
     };
 
     for (int i = 0; i < PL_ARRAY_SIZE(sh->buffers); i++)
         sh->buffers[i] = pl_str_builder_alloc(sh);
 
-    // Ensure there's always at least one `tmp` object
-    PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_new(NULL));
-
-    if (params)
-        sh->res.params = *params;
-
+    init_shader(sh, params);
     return sh;
 }
 
 static void sh_obj_deref(pl_shader_obj obj);
 
-static void sh_deref(pl_shader sh)
+void sh_deref(pl_shader sh)
 {
-    for (int i = 0; i < sh->tmp.num; i++)
-        pl_ref_deref(&sh->tmp.elem[i]);
+    pl_free_children(sh->tmp);
+
     for (int i = 0; i < sh->obj.num; i++)
         sh_obj_deref(sh->obj.elem[i]);
+    sh->obj.num = 0;
 }
 
 void pl_shader_free(pl_shader *psh)
@@ -59,6 +132,7 @@ void pl_shader_free(pl_shader *psh)
         return;
 
     sh_deref(sh);
+    pl_shader_info_deref((pl_shader_info *) &sh->info);
     pl_free_ptr(psh);
 }
 
@@ -67,29 +141,58 @@ void pl_shader_reset(pl_shader sh, const struct pl_shader_params *params)
     sh_deref(sh);
 
     struct pl_shader_t new = {
-        .log = sh->log,
-        .mutable = true,
+        .log            = sh->log,
+        .tmp            = sh->tmp,
+        .info           = sh_info_recycle(sh->info),
+        .data.buf       = sh->data.buf,
+        .mutable        = true,
 
         // Preserve array allocations
-        .tmp.elem       = sh->tmp.elem,
         .obj.elem       = sh->obj.elem,
         .vas.elem       = sh->vas.elem,
         .vars.elem      = sh->vars.elem,
         .descs.elem     = sh->descs.elem,
         .consts.elem    = sh->consts.elem,
-        .steps.elem     = sh->steps.elem,
     };
 
-    if (params)
-        new.res.params = *params;
-
     // Preserve buffer allocations
     memcpy(new.buffers, sh->buffers, sizeof(new.buffers));
     for (int i = 0; i < PL_ARRAY_SIZE(new.buffers); i++)
         pl_str_builder_reset(new.buffers[i]);
 
     *sh = new;
-    PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_new(NULL));
+    init_shader(sh, params);
+}
+
+static void *sh_alloc(pl_shader sh, size_t size, size_t align)
+{
+    const size_t offset = PL_ALIGN2(sh->data.len, align);
+    const size_t req_size = offset + size;
+    if (req_size <= pl_get_size(sh->data.buf)) {
+        sh->data.len = offset + size;
+        return sh->data.buf + offset;
+    }
+
+    // We can't realloc this buffer because various pointers will be left
+    // dangling, so just reparent it onto `sh->tmp` (so it will be cleaned
+    // up when the shader is next reset) and allocate a new, larger buffer
+    // in its place
+    const size_t new_size = PL_MAX(req_size << 1, 256);
+    pl_steal(sh->tmp, sh->data.buf);
+    sh->data.buf = pl_alloc(sh, new_size);
+    sh->data.len = size;
+    return sh->data.buf;
+}
+
+static void *sh_memdup(pl_shader sh, const void *data, size_t size, size_t align)
+{
+    if (!size)
+        return NULL;
+
+    void *dst = sh_alloc(sh, size, align);
+    assert(data);
+    memcpy(dst, data, size);
+    return dst;
 }
 
 bool pl_shader_is_failed(const pl_shader sh)
@@ -111,8 +214,8 @@ struct pl_glsl_version sh_glsl(const pl_shader sh)
 bool sh_try_compute(pl_shader sh, int bw, int bh, bool flex, size_t mem)
 {
     pl_assert(bw && bh);
-    int *sh_bw = &sh->res.compute_group_size[0];
-    int *sh_bh = &sh->res.compute_group_size[1];
+    int *sh_bw = &sh->group_size[0];
+    int *sh_bh = &sh->group_size[1];
 
     struct pl_glsl_version glsl = sh_glsl(sh);
     if (!glsl.compute) {
@@ -120,7 +223,7 @@ bool sh_try_compute(pl_shader sh, int bw, int bh, bool flex, size_t mem)
         return false;
     }
 
-    if (sh->res.compute_shmem + mem > glsl.max_shmem_size) {
+    if (sh->shmem + mem > glsl.max_shmem_size) {
         PL_TRACE(sh, "Disabling compute shader due to insufficient shmem");
         return false;
     }
@@ -146,7 +249,7 @@ bool sh_try_compute(pl_shader sh, int bw, int bh, bool flex, size_t mem)
         }
     }
 
-    sh->res.compute_shmem += mem;
+    sh->shmem += mem;
 
     // If the current shader is either not a compute shader, or we have no
     // choice but to override the metadata, always do so
@@ -154,6 +257,7 @@ bool sh_try_compute(pl_shader sh, int bw, int bh, bool flex, size_t mem)
         *sh_bw = bw;
         *sh_bh = bh;
         sh->type = SH_COMPUTE;
+        sh->flexible_work_groups = flex;
         return true;
     }
 
@@ -165,17 +269,16 @@ bool sh_try_compute(pl_shader sh, int bw, int bh, bool flex, size_t mem)
         return true;
     }
 
-    // If the other shader is rigid but this is flexible, change nothing
-    if (flex)
-        return true;
-
-    // If neither are flexible, make sure the parameters match
-    pl_assert(!flex && !sh->flexible_work_groups);
-    if (bw != *sh_bw || bh != *sh_bh) {
-        PL_TRACE(sh, "Disabling compute shader due to incompatible group "
-                 "sizes %dx%d and %dx%d", *sh_bw, *sh_bh, bw, bh);
-        sh->res.compute_shmem -= mem;
-        return false;
+    // At this point we're looking only at a non-flexible compute shader
+    pl_assert(sh->type == SH_COMPUTE && !sh->flexible_work_groups);
+    if (!flex) {
+        // Ensure parameters match
+        if (bw != *sh_bw || bh != *sh_bh) {
+            PL_TRACE(sh, "Disabling compute shader due to incompatible group "
+                     "sizes %dx%d and %dx%d", *sh_bw, *sh_bh, bw, bh);
+            sh->shmem -= mem;
+            return false;
+        }
     }
 
     return true;
@@ -198,16 +301,28 @@ bool pl_shader_output_size(const pl_shader sh, int *w, int *h)
 
 ident_t sh_fresh(pl_shader sh, const char *name)
 {
-    return pl_asprintf(SH_TMP(sh), "_%s_%d_%u", PL_DEF(name, "var"),
-                       sh->fresh++, SH_PARAMS(sh).id);
+    unsigned short id = ++sh->fresh;
+    assert(!(sh->prefix & id));
+    id |= sh->prefix;
+
+    assert(name);
+    return sh_mkident(id, name);
+}
+
+static inline ident_t sh_fresh_name(pl_shader sh, const char **pname)
+{
+    ident_t id = sh_fresh(sh, *pname);
+    *pname = sh_ident_pack(id);
+    return id;
 }
 
 ident_t sh_var(pl_shader sh, struct pl_shader_var sv)
 {
-    sv.var.name = sh_fresh(sh, sv.var.name);
-    sv.data = pl_memdup(SH_TMP(sh), sv.data, pl_var_host_layout(0, &sv.var).size);
+    ident_t id = sh_fresh_name(sh, &sv.var.name);
+    struct pl_var_layout layout = pl_var_host_layout(0, &sv.var);
+    sv.data = sh_memdup(sh, sv.data, layout.size, layout.stride);
     PL_ARRAY_APPEND(sh, sh->vars, sv);
-    return (ident_t) sv.var.name;
+    return id;
 }
 
 ident_t sh_var_int(pl_shader sh, const char *name, int val, bool dynamic)
@@ -237,10 +352,12 @@ ident_t sh_var_float(pl_shader sh, const char *name, float val, bool dynamic)
     });
 }
 
-static void merge_access(enum pl_desc_access *a, enum pl_desc_access b)
+ident_t sh_var_mat3(pl_shader sh, const char *name, pl_matrix3x3 val)
 {
-    if (*a != b)
-        *a = PL_DESC_ACCESS_READWRITE;
+    return sh_var(sh, (struct pl_shader_var) {
+        .var     = pl_var_mat3(name),
+        .data    = PL_TRANSPOSE_3X3(val.m),
+    });
 }
 
 ident_t sh_desc(pl_shader sh, struct pl_shader_desc sd)
@@ -248,18 +365,16 @@ ident_t sh_desc(pl_shader sh, struct pl_shader_desc sd)
     switch (sd.desc.type) {
     case PL_DESC_BUF_UNIFORM:
     case PL_DESC_BUF_STORAGE:
-        // Skip re-attaching the same buffer desc twice
-        // FIXME: define aliases if the variable names differ
-        for (int i = 0; i < sh->descs.num; i++) {
-            if (sh->descs.elem[i].binding.object == sd.binding.object) {
-                merge_access(&sh->descs.elem[i].desc.access, sd.desc.access);
-                sh->descs.elem[i].memory |= sd.memory;
-                return (ident_t) sh->descs.elem[i].desc.name;
-            }
-        }
-
+        for (int i = 0; i < sh->descs.num; i++) // ensure uniqueness
+            pl_assert(sh->descs.elem[i].binding.object != sd.binding.object);
         size_t bsize = sizeof(sd.buffer_vars[0]) * sd.num_buffer_vars;
-        sd.buffer_vars = pl_memdup(SH_TMP(sh), sd.buffer_vars, bsize);
+        sd.buffer_vars = sh_memdup(sh, sd.buffer_vars, bsize,
+                                   alignof(struct pl_buffer_var));
+        for (int i = 0; i < sd.num_buffer_vars; i++) {
+            struct pl_var *bv = &sd.buffer_vars[i].var;
+            const char *name = bv->name;
+            GLSLP("#define %s "$"\n", name, sh_fresh_name(sh, &bv->name));
+        }
         break;
 
     case PL_DESC_BUF_TEXEL_UNIFORM:
@@ -274,14 +389,14 @@ ident_t sh_desc(pl_shader sh, struct pl_shader_desc sd)
         pl_unreachable();
     }
 
-    sd.desc.name = sh_fresh(sh, sd.desc.name);
+    ident_t id = sh_fresh_name(sh, &sd.desc.name);
     PL_ARRAY_APPEND(sh, sh->descs, sd);
-    return (ident_t) sd.desc.name;
+    return id;
 }
 
 ident_t sh_const(pl_shader sh, struct pl_shader_const sc)
 {
-    if (sh->res.params.dynamic_constants && !sc.compile_time) {
+    if (SH_PARAMS(sh).dynamic_constants && !sc.compile_time) {
         return sh_var(sh, (struct pl_shader_var) {
             .var = {
                 .name = sc.name,
@@ -294,28 +409,29 @@ ident_t sh_const(pl_shader sh, struct pl_shader_const sc)
         });
     }
 
-    sc.name = sh_fresh(sh, sc.name);
+    ident_t id = sh_fresh_name(sh, &sc.name);
 
     pl_gpu gpu = SH_GPU(sh);
     if (gpu && gpu->limits.max_constants) {
         if (!sc.compile_time || gpu->limits.array_size_constants) {
-            sc.data = pl_memdup(SH_TMP(sh), sc.data, pl_var_type_size(sc.type));
+            size_t size = pl_var_type_size(sc.type);
+            sc.data = sh_memdup(sh, sc.data, size, size);
             PL_ARRAY_APPEND(sh, sh->consts, sc);
-            return (ident_t) sc.name;
+            return id;
         }
     }
 
     // Fallback for GPUs without specialization constants
     switch (sc.type) {
     case PL_VAR_SINT:
-        GLSLH("const int %s = %d; \n", sc.name, *(int *) sc.data);
-        return (ident_t) sc.name;
+        GLSLH("const int "$" = %d; \n", id, *(int *) sc.data);
+        return id;
     case PL_VAR_UINT:
-        GLSLH("const uint %s = %uu; \n", sc.name, *(unsigned int *) sc.data);
-        return (ident_t) sc.name;
+        GLSLH("const uint "$" = %uu; \n", id, *(unsigned int *) sc.data);
+        return id;
     case PL_VAR_FLOAT:
-        GLSLH("const float %s = %f; \n", sc.name, *(float *) sc.data);
-        return (ident_t) sc.name;
+        GLSLH("const float "$" = %f; \n", id, *(float *) sc.data);
+        return id;
     case PL_VAR_INVALID:
     case PL_VAR_TYPE_COUNT:
         break;
@@ -351,55 +467,65 @@ ident_t sh_const_float(pl_shader sh, const char *name, float val)
     });
 }
 
-ident_t sh_attr_vec2(pl_shader sh, const char *name,
-                     const struct pl_rect2df *rc)
+ident_t sh_attr(pl_shader sh, struct pl_shader_va sva)
+{
+    const size_t vsize = sva.attr.fmt->texel_size;
+    uint8_t *data = sh_alloc(sh, vsize * 4, vsize);
+    for (int i = 0; i < 4; i++) {
+        memcpy(data, sva.data[i], vsize);
+        sva.data[i] = data;
+        data += vsize;
+    }
+
+    ident_t id = sh_fresh_name(sh, &sva.attr.name);
+    PL_ARRAY_APPEND(sh, sh->vas, sva);
+    return id;
+}
+
+ident_t sh_attr_vec2(pl_shader sh, const char *name, const pl_rect2df *rc)
 {
     pl_gpu gpu = SH_GPU(sh);
     if (!gpu) {
         SH_FAIL(sh, "Failed adding vertex attr '%s': No GPU available!", name);
-        return NULL;
+        return NULL_IDENT;
     }
 
     pl_fmt fmt = pl_find_vertex_fmt(gpu, PL_FMT_FLOAT, 2);
     if (!fmt) {
         SH_FAIL(sh, "Failed adding vertex attr '%s': no vertex fmt!", name);
-        return NULL;
+        return NULL_IDENT;
     }
 
-    float vals[4][2] = {
+    float verts[4][2] = {
         { rc->x0, rc->y0 },
         { rc->x1, rc->y0 },
         { rc->x0, rc->y1 },
         { rc->x1, rc->y1 },
     };
 
-    float *data = pl_memdup(SH_TMP(sh), &vals[0][0], sizeof(vals));
-    struct pl_shader_va va = {
+    return sh_attr(sh, (struct pl_shader_va) {
         .attr = {
-            .name     = sh_fresh(sh, name),
+            .name     = name,
             .fmt      = pl_find_vertex_fmt(gpu, PL_FMT_FLOAT, 2),
         },
-        .data = { &data[0], &data[2], &data[4], &data[6] },
-    };
-
-    PL_ARRAY_APPEND(sh, sh->vas, va);
-    return (ident_t) va.attr.name;
+        .data = { verts[0], verts[1], verts[2], verts[3] },
+    });
 }
 
 ident_t sh_bind(pl_shader sh, pl_tex tex,
                 enum pl_tex_address_mode address_mode,
                 enum pl_tex_sample_mode sample_mode,
-                const char *name, const struct pl_rect2df *rect,
-                ident_t *out_pos, ident_t *out_size, ident_t *out_pt)
+                const char *name, const pl_rect2df *rect,
+                ident_t *out_pos, ident_t *out_pt)
 {
     if (pl_tex_params_dimension(tex->params) != 2) {
         SH_FAIL(sh, "Failed binding texture '%s': not a 2D texture!", name);
-        return NULL;
+        return NULL_IDENT;
     }
 
     if (!tex->params.sampleable) {
         SH_FAIL(sh, "Failed binding texture '%s': texture not sampleable!", name);
-        return NULL;
+        return NULL_IDENT;
     }
 
     ident_t itex = sh_desc(sh, (struct pl_shader_desc) {
@@ -424,25 +550,18 @@ ident_t sh_bind(pl_shader sh, pl_tex tex,
     }
 
     if (out_pos) {
-        struct pl_rect2df full = {
+        pl_rect2df full = {
             .x1 = tex->params.w,
             .y1 = tex->params.h,
         };
 
         rect = PL_DEF(rect, &full);
-        *out_pos = sh_attr_vec2(sh, "tex_coord", &(struct pl_rect2df) {
+        *out_pos = sh_attr_vec2(sh, "tex_coord", &(pl_rect2df) {
             .x0 = sx * rect->x0, .y0 = sy * rect->y0,
             .x1 = sx * rect->x1, .y1 = sy * rect->y1,
         });
     }
 
-    if (out_size) {
-        *out_size = sh_var(sh, (struct pl_shader_var) {
-            .var  = pl_var_vec2("tex_size"),
-            .data = &(float[2]) {tex->params.w, tex->params.h},
-        });
-    }
-
     if (out_pt) {
         *out_pt = sh_var(sh, (struct pl_shader_var) {
             .var  = pl_var_vec2("tex_pt"),
@@ -501,7 +620,7 @@ void sh_describef(pl_shader sh, const char *fmt, ...)
 {
     va_list ap;
     va_start(ap, fmt);
-    sh_describe(sh, pl_vasprintf(SH_TMP(sh), fmt, ap));
+    sh_describe(sh, pl_vasprintf(sh->info->tmp, fmt, ap));
     va_end(ap);
 }
 
@@ -527,13 +646,13 @@ static const char *samplers2D[] = {
     [PL_SAMPLER_EXTERNAL]   = "samplerExternalOES",
 };
 
-ident_t sh_subpass(pl_shader sh, const pl_shader sub)
+ident_t sh_subpass(pl_shader sh, pl_shader sub)
 {
     pl_assert(sh->mutable);
 
-    if (SH_PARAMS(sh).id == SH_PARAMS(sub).id) {
+    if (sh->prefix == sub->prefix) {
         PL_TRACE(sh, "Can't merge shaders: conflicting identifiers!");
-        return NULL;
+        return NULL_IDENT;
     }
 
     // Check for shader compatibility
@@ -545,18 +664,18 @@ ident_t sh_subpass(pl_shader sh, const pl_shader sub)
     {
         PL_TRACE(sh, "Can't merge shaders: incompatible sizes: %dx%d and %dx%d",
                  sh->output_w, sh->output_h, sub->output_w, sub->output_h);
-        return NULL;
+        return NULL_IDENT;
     }
 
     if (sub->type == SH_COMPUTE) {
-        int subw = sub->res.compute_group_size[0],
-            subh = sub->res.compute_group_size[1];
+        int subw = sub->group_size[0],
+            subh = sub->group_size[1];
         bool flex = sub->flexible_work_groups;
 
-        if (!sh_try_compute(sh, subw, subh, flex, sub->res.compute_shmem)) {
+        if (!sh_try_compute(sh, subw, subh, flex, sub->shmem)) {
             PL_TRACE(sh, "Can't merge shaders: incompatible block sizes or "
                      "exceeded shared memory resource capabilities");
-            return NULL;
+            return NULL_IDENT;
         }
     }
 
@@ -568,38 +687,56 @@ ident_t sh_subpass(pl_shader sh, const pl_shader sub)
     pl_str_builder_concat(sh->buffers[SH_BUF_HEADER], sub->buffers[SH_BUF_HEADER]);
 
     // Append the body as a new header function
-    ident_t name = sh_fresh(sh, "sub");
-    if (sub->res.input == PL_SHADER_SIG_SAMPLER) {
+    if (sub->input == PL_SHADER_SIG_SAMPLER) {
         pl_assert(sub->sampler_prefix);
-        GLSLH("%s %s(%c%s src_tex, vec2 tex_coord) {\n",
-              outsigs[sub->res.output], name,
+        GLSLH("%s "$"(%c%s src_tex, vec2 tex_coord) {\n",
+              outsigs[sub->output], sub->name,
               sub->sampler_prefix, samplers2D[sub->sampler_type]);
     } else {
-        GLSLH("%s %s(%s) {\n", outsigs[sub->res.output], name, insigs[sub->res.input]);
+        GLSLH("%s "$"(%s) {\n",
+              outsigs[sub->output], sub->name, insigs[sub->input]);
     }
     pl_str_builder_concat(sh->buffers[SH_BUF_HEADER], sub->buffers[SH_BUF_BODY]);
-    GLSLH("%s\n}\n\n", retvals[sub->res.output]);
+    GLSLH("%s\n}\n\n", retvals[sub->output]);
 
-    // Ref all objects
-    PL_ARRAY_CONCAT(sh, sh->obj, sub->obj);
-    for (int i = 0; i < sub->obj.num; i++)
-        pl_rc_ref(&sub->obj.elem[i]->rc);
+    // Steal all inputs and objects from the subpass
+#define ARRAY_STEAL(arr) do                 \
+{                                           \
+    PL_ARRAY_CONCAT(sh, sh->arr, sub->arr); \
+    sub->arr.num = 0;                       \
+} while (0)
 
-    // Copy over all of the descriptors etc.
-    for (int i = 0; i < sub->tmp.num; i++)
-        PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_dup(sub->tmp.elem[i]));
-    PL_ARRAY_CONCAT(sh, sh->vas, sub->vas);
-    PL_ARRAY_CONCAT(sh, sh->vars, sub->vars);
-    PL_ARRAY_CONCAT(sh, sh->descs, sub->descs);
-    PL_ARRAY_CONCAT(sh, sh->consts, sub->consts);
-    PL_ARRAY_CONCAT(sh, sh->steps, sub->steps);
+    ARRAY_STEAL(obj);
+    ARRAY_STEAL(vas);
+    ARRAY_STEAL(vars);
+    ARRAY_STEAL(descs);
+    ARRAY_STEAL(consts);
+#undef ARRAY_STEAL
 
-    return name;
+    // Steal the scratch buffer (if it holds data)
+    if (sub->data.len) {
+        pl_steal(sh->tmp, sub->data.buf);
+        sub->data = (pl_str) {0};
+    }
+
+    // Steal all temporary allocations and mark the child as unusable
+    pl_steal(sh->tmp, sub->tmp);
+    sub->tmp = pl_tmp(sub);
+    sub->failed = true;
+
+    // Steal the shader steps array (and allocations)
+    pl_assert(pl_rc_count(&sub->info->rc) == 1);
+    PL_ARRAY_CONCAT(sh->info, sh->info->steps, sub->info->steps);
+    pl_steal(sh->info->tmp, sub->info->tmp);
+    sub->info->tmp = pl_tmp(sub->info);
+    sub->info->steps.num = 0; // sanity
+
+    return sub->name;
 }
 
 pl_str_builder sh_finalize_internal(pl_shader sh)
 {
-    pl_assert(sh->mutable);
+    pl_assert(sh->mutable); // this function should only ever be called once
     if (sh->failed)
         return NULL;
 
@@ -609,81 +746,113 @@ pl_str_builder sh_finalize_internal(pl_shader sh)
     // Concatenate everything onto the prelude to form the final output
     pl_str_builder_concat(sh->buffers[SH_BUF_PRELUDE], sh->buffers[SH_BUF_HEADER]);
 
-    sh->res.name = sh_fresh(sh, "main");
-    if (sh->res.input == PL_SHADER_SIG_SAMPLER) {
+    if (sh->input == PL_SHADER_SIG_SAMPLER) {
         pl_assert(sh->sampler_prefix);
-        GLSLP("%s %s(%c%s src_tex, vec2 tex_coord) {\n",
-              outsigs[sh->res.output], sh->res.name,
-              sh->sampler_prefix, samplers2D[sh->sampler_type]);
+        GLSLP("%s "$"(%c%s src_tex, vec2 tex_coord) {\n",
+              outsigs[sh->output], sh->name,
+              sh->sampler_prefix,
+              samplers2D[sh->sampler_type]);
     } else {
-        GLSLP("%s %s(%s) {\n", outsigs[sh->res.output], sh->res.name, insigs[sh->res.input]);
+        GLSLP("%s "$"(%s) {\n", outsigs[sh->output], sh->name, insigs[sh->input]);
     }
 
     pl_str_builder_concat(sh->buffers[SH_BUF_PRELUDE], sh->buffers[SH_BUF_BODY]);
     pl_str_builder_concat(sh->buffers[SH_BUF_PRELUDE], sh->buffers[SH_BUF_FOOTER]);
-    GLSLP("%s\n}\n\n", retvals[sh->res.output]);
+    GLSLP("%s\n}\n\n", retvals[sh->output]);
 
-    // Generate the pretty description
-    sh->res.description = "(unknown shader)";
-    if (sh->steps.num) {
-        // Reuse this builder
-        pl_str_builder desc = sh->buffers[SH_BUF_BODY];
-        pl_str_builder_reset(desc);
+    // Generate the shader info
+    struct sh_info *info = sh->info;
+    info->info.steps = info->steps.elem;
+    info->info.num_steps = info->steps.num;
+    info->info.description = "(unknown shader)";
 
-        for (int i = 0; i < sh->steps.num; i++) {
-            const char *step = sh->steps.elem[i];
-            if (!step)
-                continue;
+    // Generate pretty description
+    for (int i = 0; i < info->steps.num; i++) {
+        const char *step = info->steps.elem[i];
 
-            // Group together duplicates. We're okay using a weak equality
-            // check here because all pass descriptions are static strings.
-            int count = 1;
-            for (int j = i+1; j < sh->steps.num; j++) {
-                if (sh->steps.elem[j] == step) {
-                    sh->steps.elem[j] = NULL;
-                    count++;
-                }
-            }
-
-            if (i > 0)
-                pl_str_builder_const_str(desc, ", ");
-            pl_str_builder_str0(desc, step);
-            if (count > 1)
-                pl_str_builder_printf_c(desc, " x%d", count);
+        // Prevent duplicates. We're okay using a weak equality check here
+        // because most pass descriptions are static strings.
+        for (int j = 0; j < i; j++) {
+            if (info->steps.elem[j] == step)
+                goto next_step;
         }
 
-        sh->res.description = (char *) pl_str_builder_exec(desc).buf;
+        int count = 1;
+        for (int j = i+1; j < info->steps.num; j++) {
+            if (info->steps.elem[j] == step)
+                count++;
+        }
+
+        const char *prefix = i > 0 ? ", " : "";
+        if (count > 1) {
+            pl_str_append_asprintf(info, &info->desc, "%s%s x%d",
+                                   prefix, step, count);
+        } else {
+            pl_str_append_asprintf(info, &info->desc, "%s%s", prefix, step);
+        }
+
+next_step: ;
     }
 
-    // Set the vas/vars/descs
-    sh->res.vertex_attribs = sh->vas.elem;
-    sh->res.num_vertex_attribs = sh->vas.num;
-    sh->res.variables = sh->vars.elem;
-    sh->res.num_variables = sh->vars.num;
-    sh->res.descriptors = sh->descs.elem;
-    sh->res.num_descriptors = sh->descs.num;
-    sh->res.constants = sh->consts.elem;
-    sh->res.num_constants = sh->consts.num;
-    sh->res.steps = sh->steps.elem;
-    sh->res.num_steps = sh->steps.num;
+    if (info->desc.len)
+        info->info.description = (char *) info->desc.buf;
+
     sh->mutable = false;
     return sh->buffers[SH_BUF_PRELUDE];
 }
 
 const struct pl_shader_res *pl_shader_finalize(pl_shader sh)
 {
-    pl_str_builder glsl = NULL;
-    if (sh->mutable) {
-        glsl = sh_finalize_internal(sh);
-        if (!glsl)
-            return NULL;
+    if (sh->failed) {
+        return NULL;
+    } else if (!sh->mutable) {
+        return &sh->result;
     }
 
-    pl_assert(!sh->mutable);
-    if (!sh->res.glsl)
-        sh->res.glsl = (char *) pl_str_builder_exec(glsl).buf;
+    pl_shader_info info = &sh->info->info;
+    pl_str_builder glsl = sh_finalize_internal(sh);
 
-    return &sh->res;
+    // Turn ident_t into friendly strings before passing it to users
+#define FIX_IDENT(name) \
+    name = sh_ident_tostr(sh_ident_unpack(name))
+    for (int i = 0; i < sh->vas.num; i++)
+        FIX_IDENT(sh->vas.elem[i].attr.name);
+    for (int i = 0; i < sh->vars.num; i++)
+        FIX_IDENT(sh->vars.elem[i].var.name);
+    for (int i = 0; i < sh->consts.num; i++)
+        FIX_IDENT(sh->consts.elem[i].name);
+    for (int i = 0; i < sh->descs.num; i++) {
+        struct pl_shader_desc *sd = &sh->descs.elem[i];
+        FIX_IDENT(sd->desc.name);
+        for (int j = 0; j < sd->num_buffer_vars; sd++)
+            FIX_IDENT(sd->buffer_vars[j].var.name);
+    }
+#undef FIX_IDENT
+
+    sh->result = (struct pl_shader_res) {
+        .info               = info,
+        .glsl               = (char *) pl_str_builder_exec(glsl).buf,
+        .name               = sh_ident_tostr(sh->name),
+        .input              = sh->input,
+        .output             = sh->output,
+        .compute_group_size = { sh->group_size[0], sh->group_size[1] },
+        .compute_shmem      = sh->shmem,
+        .vertex_attribs     = sh->vas.elem,
+        .num_vertex_attribs = sh->vas.num,
+        .variables          = sh->vars.elem,
+        .num_variables      = sh->vars.num,
+        .descriptors        = sh->descs.elem,
+        .num_descriptors    = sh->descs.num,
+        .constants          = sh->consts.elem,
+        .num_constants      = sh->consts.num,
+        // deprecated fields
+        .params             = info->params,
+        .steps              = info->steps,
+        .num_steps          = info->num_steps,
+        .description        = info->description,
+    };
+
+    return &sh->result;
 }
 
 bool sh_require(pl_shader sh, enum pl_shader_sig insig, int w, int h)
@@ -714,18 +883,18 @@ bool sh_require(pl_shader sh, enum pl_shader_sig insig, int w, int h)
 
     // If we require an input, but there is none available - just get it from
     // the user by turning it into an explicit input signature.
-    if (!sh->res.output && insig) {
-        pl_assert(!sh->res.input);
-        sh->res.input = insig;
-    } else if (sh->res.output != insig) {
+    if (!sh->output && insig) {
+        pl_assert(!sh->input);
+        sh->input = insig;
+    } else if (sh->output != insig) {
         SH_FAIL(sh, "Illegal sequence of shader operations! Current output "
                 "signature is '%s', but called operation expects '%s'!",
-                names[sh->res.output], names[insig]);
+                names[sh->output], names[insig]);
         return false;
     }
 
     // All of our shaders end up returning a vec4 color
-    sh->res.output = PL_SHADER_SIG_COLOR;
+    sh->output = PL_SHADER_SIG_COLOR;
     sh->output_w = PL_DEF(sh->output_w, w);
     sh->output_h = PL_DEF(sh->output_h, h);
     return true;
@@ -792,96 +961,32 @@ ident_t sh_prng(pl_shader sh, bool temporal, ident_t *p_state)
     ident_t randfun = sh_fresh(sh, "rand"),
             state = sh_fresh(sh, "state");
 
-    if (sh_glsl(sh).version >= 130) {
-
-        // Based on pcg3d (http://jcgt.org/published/0009/03/02/)
-        GLSLP("#define prng_t uvec3\n");
-        GLSLH("vec3 %s(inout uvec3 s) {                     \n"
-              "    s = 1664525u * s + uvec3(1013904223u);   \n"
-              "    s.x += s.y * s.z;                        \n"
-              "    s.y += s.z * s.x;                        \n"
-              "    s.z += s.x * s.y;                        \n"
-              "    s ^= s >> 16u;                           \n"
-              "    s.x += s.y * s.z;                        \n"
-              "    s.y += s.z * s.x;                        \n"
-              "    s.z += s.x * s.y;                        \n"
-              "    return vec3(s) * 1.0/float(0xFFFFFFFFu); \n"
-              "}                                            \n",
-              randfun);
-
-        const char *seed = "0u";
-        if (temporal) {
-            seed = sh_var(sh, (struct pl_shader_var) {
-                .var  = pl_var_uint("seed"),
-                .data = &(unsigned int){ SH_PARAMS(sh).index },
-                .dynamic = true,
-            });
-        };
-
-        GLSL("uvec3 %s = uvec3(gl_FragCoord.xy, %s); \n", state, seed);
+    // Based on pcg3d (http://jcgt.org/published/0009/03/02/)
+    GLSLP("#define prng_t uvec3\n");
+    GLSLH("vec3 "$"(inout uvec3 s) {                    \n"
+          "    s = 1664525u * s + uvec3(1013904223u);   \n"
+          "    s.x += s.y * s.z;                        \n"
+          "    s.y += s.z * s.x;                        \n"
+          "    s.z += s.x * s.y;                        \n"
+          "    s ^= s >> 16u;                           \n"
+          "    s.x += s.y * s.z;                        \n"
+          "    s.y += s.z * s.x;                        \n"
+          "    s.z += s.x * s.y;                        \n"
+          "    return vec3(s) * 1.0/float(0xFFFFFFFFu); \n"
+          "}                                            \n",
+          randfun);
 
+    if (temporal) {
+        GLSL("uvec3 "$" = uvec3(gl_FragCoord.xy, "$"); \n",
+             state, SH_UINT_DYN(SH_PARAMS(sh).index));
     } else {
-
-        // Based on SGGP (https://briansharpe.wordpress.com/2011/10/01/gpu-texture-free-noise/)
-        ident_t permute = sh_fresh(sh, "permute");
-        GLSLP("#define prng_t float\n");
-        GLSLH("float %s(float x) {                          \n"
-              "    x = (34.0 * x + 1.0) * x;                \n"
-              "    return fract(x * 1.0/289.0) * 289.0;     \n"
-              "}                                            \n"
-              "vec3 %s(inout float s) {                     \n"
-              "    vec3 ret;                                \n"
-              "    ret.x = %s(s);                           \n"
-              "    ret.y = %s(ret.x);                       \n"
-              "    ret.z = %s(ret.y);                       \n"
-              "    s = ret.z;                               \n"
-              "    return fract(ret * 1.0/41.0);            \n"
-              "}                                            \n",
-              permute, randfun, permute, permute, permute);
-
-        static const double phi = 1.618033988749895;
-        const char *seed = "0.0";
-        if (temporal) {
-            seed = sh_var(sh, (struct pl_shader_var) {
-                .var  = pl_var_float("seed"),
-                .data = &(float){ modff(phi * SH_PARAMS(sh).index, &(float){0}) },
-                .dynamic = true,
-            });
-        };
-
-        GLSL("vec3 %s_m = vec3(fract(gl_FragCoord.xy * vec2(%f)), %s);  \n"
-             "%s_m += vec3(1.0);                                        \n"
-             "float %s = %s(%s(%s(%s_m.x) + %s_m.y) + %s_m.z);          \n",
-             state, phi, seed,
-             state,
-             state, permute, permute, permute, state, state, state);
-
+        GLSL("uvec3 "$" = uvec3(gl_FragCoord.xy, 0.0); \n", state);
     }
 
     if (p_state)
         *p_state = state;
 
     ident_t res = sh_fresh(sh, "RAND");
-    GLSLH("#define %s (%s(%s))\n", res, randfun, state);
+    GLSLH("#define "$" ("$"("$"))\n", res, randfun, state);
     return res;
 }
-
-const char *sh_bvec(const pl_shader sh, int dims)
-{
-    static const char *bvecs[] = {
-        [1] = "bool",
-        [2] = "bvec2",
-        [3] = "bvec3",
-        [4] = "bvec4",
-    };
-
-    static const char *vecs[] = {
-        [1] = "float",
-        [2] = "vec2",
-        [3] = "vec3",
-        [4] = "vec4",
-    };
-
-    pl_assert(dims > 0 && dims < PL_ARRAY_SIZE(bvecs));
-    return sh_glsl(sh).version >= 130 ? bvecs[dims] : vecs[dims];
-}
