@@ -18,15 +18,62 @@
 #pragma once
 
 #include <stdio.h>
+#include <limits.h>
 
 #include "common.h"
 #include "log.h"
 #include "gpu.h"
 
+#include <libplacebo/shaders.h>
+
 // This represents an identifier (e.g. name of function, uniform etc.) for
 // a shader resource. The generated identifiers are immutable, but only live
 // until pl_shader_reset - so make copies when passing to external stuff.
-typedef const char * ident_t;
+//
+// Note: When PL_STRIP_SHADERS is set during compile time, identifiers will be
+// treated directly as integers, with no human-readable names.
+
+#ifdef PL_STRIP_SHADERS
+
+typedef unsigned short ident_t;
+#define $           "_%hx"
+#define NULL_IDENT  0u
+
+#define sh_mkident(id, name) ((ident_t) id)
+#define sh_ident_tostr(id)   pl_asprintf(sh->tmp, $, id)
+
+enum {
+    IDENT_BITS     = 8 * sizeof(ident_t),
+    IDENT_MASK     = (uintptr_t) USHRT_MAX,
+    IDENT_SENTINEL = (uintptr_t) 0x20230319 << IDENT_BITS,
+};
+
+// Functions to pack/unpack an identifier into a `const char *` name field.
+// Used to defer string templating of friendly names until actually necessary
+static inline const char *sh_ident_pack(ident_t id)
+{
+    return (const char *) (IDENT_SENTINEL | id);
+}
+
+static inline ident_t sh_ident_unpack(const char *name)
+{
+    uintptr_t uname = (uintptr_t) name;
+    assert((uname & ~IDENT_MASK) == IDENT_SENTINEL);
+    return uname & IDENT_MASK;
+}
+
+#else // !PL_STRIP_SHADERS
+
+typedef const char *ident_t;
+#define $           "%s"
+#define NULL_IDENT  NULL
+
+#define sh_mkident(id, name)  pl_asprintf(sh->tmp, "_%hx_%s", id, name)
+#define sh_ident_tostr(id)    ((const char *) id)
+#define sh_ident_pack(id)     ((const char *) id)
+#define sh_ident_unpack(name) ((ident_t) name)
+
+#endif // PL_STRIP_SHADERS
 
 enum pl_shader_buf {
     SH_BUF_PRELUDE, // extra #defines etc.
@@ -42,34 +89,65 @@ enum pl_shader_type {
     SH_FRAGMENT
 };
 
-struct pl_shader {
+struct sh_info {
+    // public-facing struct
+    struct pl_shader_info_t info;
+
+    // internal fields
+    void *tmp;
+    pl_rc_t rc;
+    pl_str desc;
+    PL_ARRAY(const char *) steps;
+};
+
+struct pl_shader_t {
     pl_log log;
-    struct pl_shader_res res; // for accumulating some of the fields
-    PL_ARRAY(struct pl_ref *) tmp; // only used for var/va/desc names and data
+    void *tmp; // temporary allocations (freed on pl_shader_reset)
+    struct sh_info *info;
+    pl_str data; // pooled/recycled scratch buffer for small allocations
+    PL_ARRAY(pl_shader_obj) obj;
     bool failed;
     bool mutable;
+    ident_t name;
+    enum pl_shader_sig input, output;
     int output_w;
     int output_h;
     bool transpose;
-    pl_str buffers[SH_BUF_COUNT];
+    pl_str_builder buffers[SH_BUF_COUNT];
     enum pl_shader_type type;
     bool flexible_work_groups;
+    int group_size[2];
+    size_t shmem;
     enum pl_sampler_type sampler_type;
     char sampler_prefix;
-    int fresh;
+    unsigned short prefix; // pre-processed version of res.params.id
+    unsigned short fresh;
 
-    // mutable versions of the fields from pl_shader_res
+    // Note: internally, these `pl_shader_va` etc. use raw ident_t fields
+    // instead of `const char *` wherever a name is required! These are
+    // translated to legal strings either in `pl_shader_finalize`, or inside
+    // the `pl_dispatch` shader compilation step.
     PL_ARRAY(struct pl_shader_va) vas;
     PL_ARRAY(struct pl_shader_var) vars;
     PL_ARRAY(struct pl_shader_desc) descs;
     PL_ARRAY(struct pl_shader_const) consts;
-    PL_ARRAY(const char *) steps;
+
+    // cached result of `pl_shader_finalize`
+    struct pl_shader_res result;
 };
 
+// Free temporary resources associated with a shader. Normally called by
+// pl_shader_reset(), but used internally to reduce memory waste.
+void sh_deref(pl_shader sh);
+
+// Same as `pl_shader_finalize` but doesn't generate `sh->res`, instead returns
+// the string builder to be used to finalize the shader. Assumes the caller
+// will access the shader's internal fields directly.
+pl_str_builder sh_finalize_internal(pl_shader sh);
+
 // Helper functions for convenience
-#define SH_PARAMS(sh) ((sh)->res.params)
+#define SH_PARAMS(sh) ((sh)->info->info.params)
 #define SH_GPU(sh) (SH_PARAMS(sh).gpu)
-#define SH_TMP(sh) ((sh)->tmp.elem[0])
 
 // Returns the GLSL version, defaulting to desktop 130.
 struct pl_glsl_version sh_glsl(const pl_shader sh);
@@ -85,7 +163,10 @@ bool sh_try_compute(pl_shader sh, int bw, int bh, bool flex, size_t mem);
 // Attempt merging a secondary shader into the current shader. Returns NULL if
 // merging fails (e.g. incompatible signatures); otherwise returns an identifier
 // corresponding to the generated subpass function.
-ident_t sh_subpass(pl_shader sh, const pl_shader sub);
+//
+// If successful, the subpass shader is set to an undefined failure state and
+// must be explicitly reset/aborted before being re-used.
+ident_t sh_subpass(pl_shader sh, pl_shader sub);
 
 // Helpers for adding new variables/descriptors/etc. with fresh, unique
 // identifier names. These will never conflict with other identifiers, even
@@ -95,6 +176,16 @@ ident_t sh_fresh(pl_shader sh, const char *name);
 // Add a new shader var and return its identifier
 ident_t sh_var(pl_shader sh, struct pl_shader_var sv);
 
+// Helper functions for `sh_var`
+ident_t sh_var_int(pl_shader sh, const char *name, int val, bool dynamic);
+ident_t sh_var_uint(pl_shader sh, const char *name, unsigned int val, bool dynamic);
+ident_t sh_var_float(pl_shader sh, const char *name, float val, bool dynamic);
+ident_t sh_var_mat3(pl_shader sh, const char *name, pl_matrix3x3 val);
+#define SH_INT_DYN(val)   sh_var_int(sh, "const", val, true)
+#define SH_UINT_DYN(val)  sh_var_uint(sh, "const", val, true)
+#define SH_FLOAT_DYN(val) sh_var_float(sh, "const", val, true)
+#define SH_MAT3(val) sh_var_mat3(sh, "mat", val)
+
 // Add a new shader desc and return its identifier.
 ident_t sh_desc(pl_shader sh, struct pl_shader_desc sd);
 
@@ -109,9 +200,11 @@ ident_t sh_const_float(pl_shader sh, const char *name, float val);
 #define SH_UINT(val)    sh_const_uint(sh, "const", val)
 #define SH_FLOAT(val)   sh_const_float(sh, "const", val)
 
-// Add a new vec2 vertex attribute from a pl_rect2df, or returns NULL on failure.
-ident_t sh_attr_vec2(pl_shader sh, const char *name,
-                     const struct pl_rect2df *rc);
+// Add a new shader va and return its identifier
+ident_t sh_attr(pl_shader sh, struct pl_shader_va sva);
+
+// Helper to add a a vec2 VA from a pl_rect2df. Returns NULL_IDENT on failure.
+ident_t sh_attr_vec2(pl_shader sh, const char *name, const pl_rect2df *rc);
 
 // Bind a texture under a given transformation and make its attributes
 // available as well. If an output pointer for one of the attributes is left
@@ -123,8 +216,8 @@ ident_t sh_attr_vec2(pl_shader sh, const char *name,
 ident_t sh_bind(pl_shader sh, pl_tex tex,
                 enum pl_tex_address_mode address_mode,
                 enum pl_tex_sample_mode sample_mode,
-                const char *name, const struct pl_rect2df *rect,
-                ident_t *out_pos, ident_t *out_size, ident_t *out_pt);
+                const char *name, const pl_rect2df *rect,
+                ident_t *out_pos, ident_t *out_pt);
 
 // Incrementally build up a buffer by adding new variable elements to the
 // buffer, resizing buf.buffer_vars if necessary. Returns whether or not the
@@ -140,10 +233,11 @@ size_t sh_buf_desc_size(const struct pl_shader_desc *buf_desc);
 
 
 // Underlying function for appending text to a shader
-void sh_append(pl_shader sh, enum pl_shader_buf buf, const char *fmt, ...)
-    PL_PRINTF(3, 4);
+#define sh_append(sh, buf, ...) \
+    pl_str_builder_addf((sh)->buffers[buf], __VA_ARGS__)
 
-void sh_append_str(pl_shader sh, enum pl_shader_buf buf, pl_str str);
+#define sh_append_str(sh, buf, str) \
+    pl_str_builder_str((sh)->buffers[buf], str)
 
 #define GLSLP(...) sh_append(sh, SH_BUF_PRELUDE, __VA_ARGS__)
 #define GLSLH(...) sh_append(sh, SH_BUF_HEADER, __VA_ARGS__)
@@ -156,7 +250,7 @@ void sh_describef(pl_shader sh, const char *fmt, ...)
 
 static inline void sh_describe(pl_shader sh, const char *desc)
 {
-    PL_ARRAY_APPEND(sh, sh->steps, desc);
+    PL_ARRAY_APPEND(sh->info, sh->info->steps, desc);
 };
 
 // Requires that the share is mutable, has an output signature compatible
@@ -168,7 +262,7 @@ bool sh_require(pl_shader sh, enum pl_shader_sig insig, int w, int h);
 
 enum pl_shader_obj_type {
     PL_SHADER_OBJ_INVALID = 0,
-    PL_SHADER_OBJ_TONE_MAP,
+    PL_SHADER_OBJ_COLOR_MAP,
     PL_SHADER_OBJ_SAMPLER,
     PL_SHADER_OBJ_DITHER,
     PL_SHADER_OBJ_LUT,
@@ -177,8 +271,9 @@ enum pl_shader_obj_type {
     PL_SHADER_OBJ_RESHAPE,
 };
 
-struct pl_shader_obj {
+struct pl_shader_obj_t {
     enum pl_shader_obj_type type;
+    pl_rc_t rc;
     pl_gpu gpu;
     void (*uninit)(pl_gpu gpu, void *priv);
     void *priv;
@@ -201,19 +296,32 @@ void *sh_require_obj(pl_shader sh, pl_shader_obj *ptr,
 // If `temporal` is set, the PRNG will vary across frames.
 ident_t sh_prng(pl_shader sh, bool temporal, ident_t *state);
 
-enum sh_lut_method {
+// Backing memory type
+enum sh_lut_type {
     SH_LUT_AUTO = 0, // pick whatever makes the most sense
     SH_LUT_TEXTURE,  // upload as texture
     SH_LUT_UNIFORM,  // uniform array
     SH_LUT_LITERAL,  // constant / literal array in shader source (fallback)
 };
 
+// Interpolation method
+enum sh_lut_method {
+    SH_LUT_NONE = 0,    // no interpolation, integer indices
+    SH_LUT_LINEAR,      // linear interpolation, vecN indices in range [0,1]
+    SH_LUT_CUBIC,       // (bi/tri)cubic interpolation
+    SH_LUT_TETRAHEDRAL, // tetrahedral interpolation for vec3, equivalent to
+                        // SH_LUT_LINEAR for lower dimensions
+};
+
 struct sh_lut_params {
     pl_shader_obj *object;
 
-    // Method and type of the LUT we intend to generate.
+    // Type of the LUT we intend to generate.
+    //
+    // Note: If `var_type` is PL_VAR_*INT, `method` must be SH_LUT_NONE.
+    enum pl_var_type var_type;
+    enum sh_lut_type lut_type;
     enum sh_lut_method method;
-    enum pl_var_type type;
 
     // For SH_LUT_TEXTURE, this can be used to override the texture's internal
     // format, in which case it takes precedence over the default for `type`.
@@ -225,10 +333,6 @@ struct sh_lut_params {
     int depth;
     int comps;
 
-    // If true, LUT takes a vecN coordinate and linearly interpolates values,
-    // rather than taking an ivecN. Requires `type == PL_VAR_FLOAT`!
-    bool linear;
-
     // If true, the LUT will always be regenerated, even if the dimensions have
     // not changed.
     bool update;
@@ -260,22 +364,28 @@ struct sh_lut_params {
 // gets interpolated and clamped as needed. Returns NULL on error.
 ident_t sh_lut(pl_shader sh, const struct sh_lut_params *params);
 
-// Returns a GLSL-version appropriate "bvec"-like type. For GLSL 130+, this
-// returns bvecN. For GLSL 120, this returns vecN instead. The intended use of
-// this function is with mix(), which only accepts bvec in GLSL 130+.
-const char *sh_bvec(const pl_shader sh, int dims);
-
-// Returns the appropriate `texture`-equivalent function for the shader and
-// given texture.
-static inline const char *sh_tex_fn(const pl_shader sh,
-                                    const struct pl_tex_params params)
+static inline const char *sh_float_type(uint8_t num_comps)
 {
-    static const char *suffixed[] = {
-        [1] = "texture1D",
-        [2] = "texture2D",
-        [3] = "texture3D",
-    };
+    switch (num_comps) {
+    case 1: return "float";
+    case 2: return "vec2";
+    case 3: return "vec3";
+    case 4: return "vec4";
+    }
 
-    int dims = pl_tex_params_dimension(params);
-    return sh_glsl(sh).version >= 130 ? "texture" : suffixed[dims];
+    pl_unreachable();
+}
+
+static inline uint8_t sh_tex_swiz(char swiz[5], uint8_t comp_mask)
+{
+    uint8_t num_comps = 0;
+    for (uint8_t comps = comp_mask; comps;) {
+        uint8_t c = __builtin_ctz(comps);
+        assert(c < 4 && num_comps < 4);
+        swiz[num_comps++] = "xyzw"[c];
+        comps &= ~(1u << c);
+    }
+
+    swiz[num_comps] = '\0';
+    return num_comps;
 }
