@@ -841,6 +841,8 @@ static enum pl_hook_stage mp_stage_to_pl(pl_str stage)
 
     if (pl_str_equals0(stage, "SCALED"))
         return PL_HOOK_SCALED;
+    if (pl_str_equals0(stage, "PREOUTPUT"))
+        return PL_HOOK_PRE_OUTPUT;
     if (pl_str_equals0(stage, "OUTPUT"))
         return PL_HOOK_OUTPUT;
 
@@ -868,6 +870,7 @@ static pl_str pl_stage_to_mp(enum pl_hook_stage stage)
     case PL_HOOK_POST_KERNEL:   return pl_str0("POSTKERNEL");
 
     case PL_HOOK_SCALED:        return pl_str0("SCALED");
+    case PL_HOOK_PRE_OUTPUT:    return pl_str0("PREOUTPUT");
     case PL_HOOK_OUTPUT:        return pl_str0("OUTPUT");
     };
 
@@ -884,7 +887,7 @@ struct pass_tex {
     pl_tex tex;
 
     // Metadata
-    struct pl_rect2df rect;
+    pl_rect2df rect;
     struct pl_color_repr repr;
     struct pl_color_space color;
     int comps;
@@ -1104,25 +1107,25 @@ static double prng_step(uint64_t s[4])
 
 static bool bind_pass_tex(pl_shader sh, pl_str name,
                           const struct pass_tex *ptex,
-                          const struct pl_rect2df *rect,
+                          const pl_rect2df *rect,
                           bool hooked, bool mainpresub)
 {
-    ident_t id, pos, size, pt;
+    ident_t id, pos, pt;
 
     // Compatibility with mpv texture binding semantics
     id = sh_bind(sh, ptex->tex, PL_TEX_ADDRESS_CLAMP, PL_TEX_SAMPLE_LINEAR,
-                 "hook_tex", rect, &pos, &size, &pt);
+                 "hook_tex", rect, &pos, &pt);
     if (!id)
         return false;
 
-    GLSLH("#define %.*s_raw %s \n", PL_STR_FMT(name), id);
-    GLSLH("#define %.*s_pos %s \n", PL_STR_FMT(name), pos);
-    GLSLH("#define %.*s_map %s_map \n", PL_STR_FMT(name), pos);
-    GLSLH("#define %.*s_size %s \n", PL_STR_FMT(name), size);
-    GLSLH("#define %.*s_pt %s \n", PL_STR_FMT(name), pt);
+    GLSLH("#define %.*s_raw "$" \n", PL_STR_FMT(name), id);
+    GLSLH("#define %.*s_pos "$" \n", PL_STR_FMT(name), pos);
+    GLSLH("#define %.*s_map "$"_map \n", PL_STR_FMT(name), pos);
+    GLSLH("#define %.*s_size vec2(textureSize("$", 0)) \n", PL_STR_FMT(name), id);
+    GLSLH("#define %.*s_pt "$" \n", PL_STR_FMT(name), pt);
 
     float off[2] = { ptex->rect.x0, ptex->rect.y0 };
-    GLSLH("#define %.*s_off %s \n", PL_STR_FMT(name),
+    GLSLH("#define %.*s_off "$" \n", PL_STR_FMT(name),
           sh_var(sh, (struct pl_shader_var) {
               .var = pl_var_vec2("offset"),
               .data = off,
@@ -1130,20 +1133,20 @@ static bool bind_pass_tex(pl_shader sh, pl_str name,
 
     struct pl_color_repr repr = ptex->repr;
     ident_t scale = SH_FLOAT(pl_color_repr_normalize(&repr));
-    GLSLH("#define %.*s_mul %s \n", PL_STR_FMT(name), scale);
+    GLSLH("#define %.*s_mul "$" \n", PL_STR_FMT(name), scale);
 
     // Compatibility with mpv
     GLSLH("#define %.*s_rot mat2(1.0, 0.0, 0.0, 1.0) \n", PL_STR_FMT(name));
 
     // Sampling function boilerplate
-    GLSLH("#define %.*s_tex(pos) (%s * vec4(%s(%s, pos))) \n",
-          PL_STR_FMT(name), scale, sh_tex_fn(sh, ptex->tex->params), id);
-    GLSLH("#define %.*s_texOff(off) (%.*s_tex(%s + %s * vec2(off))) \n",
+    GLSLH("#define %.*s_tex(pos) ("$" * vec4(textureLod("$", pos, 0.0))) \n",
+          PL_STR_FMT(name), scale, id);
+    GLSLH("#define %.*s_texOff(off) (%.*s_tex("$" + "$" * vec2(off))) \n",
           PL_STR_FMT(name), PL_STR_FMT(name), pos, pt);
 
     bool can_gather = ptex->tex->params.format->gatherable;
     if (can_gather) {
-        GLSLH("#define %.*s_gather(pos, c) (%s * vec4(textureGather(%s, pos, c))) \n",
+        GLSLH("#define %.*s_gather(pos, c) ("$" * vec4(textureGather("$", pos, c))) \n",
               PL_STR_FMT(name), scale, id);
     }
 
@@ -1275,12 +1278,11 @@ static struct pl_hook_res hook_hook(void *priv, const struct pl_hook_params *par
                     // Directly bind this, no need to bother with all the
                     // `bind_pass_tex` boilerplate
                     ident_t id = sh_desc(sh, p->descriptors.elem[j]);
-                    GLSLH("#define %.*s %s \n", PL_STR_FMT(texname), id);
+                    GLSLH("#define %.*s "$" \n", PL_STR_FMT(texname), id);
 
                     if (p->descriptors.elem[j].desc.type == PL_DESC_SAMPLED_TEX) {
-                        pl_tex tex = p->descriptors.elem[j].binding.object;
-                        GLSLH("#define %.*s_tex(pos) (%s(%s, pos)) \n",
-                              PL_STR_FMT(texname), sh_tex_fn(sh, tex->params), id);
+                        GLSLH("#define %.*s_tex(pos) (textureLod("$", pos, 0.0)) \n",
+                              PL_STR_FMT(texname), id);
                     }
                     goto next_bind;
                 }
@@ -1292,7 +1294,7 @@ static struct pl_hook_res hook_hook(void *priv, const struct pl_hook_params *par
                     // params->rect, because user shaders in general are not
                     // designed to handle cropped input textures.
                     const struct pass_tex *ptex = &p->pass_textures.elem[j];
-                    struct pl_rect2df rect = {
+                    pl_rect2df rect = {
                         0, 0, ptex->tex->params.w, ptex->tex->params.h,
                     };
 
@@ -1327,33 +1329,33 @@ static struct pl_hook_res hook_hook(void *priv, const struct pl_hook_params *par
 
         // Set up the input variables
         p->frame_count++;
-        GLSLH("#define frame %s \n", sh_var(sh, (struct pl_shader_var) {
+        GLSLH("#define frame "$" \n", sh_var(sh, (struct pl_shader_var) {
             .var = pl_var_int("frame"),
             .data = &p->frame_count,
             .dynamic = true,
         }));
 
         float random = prng_step(p->prng_state);
-        GLSLH("#define random %s \n", sh_var(sh, (struct pl_shader_var) {
+        GLSLH("#define random "$" \n", sh_var(sh, (struct pl_shader_var) {
             .var = pl_var_float("random"),
             .data = &random,
             .dynamic = true,
         }));
 
         float src_size[2] = { pl_rect_w(params->src_rect), pl_rect_h(params->src_rect) };
-        GLSLH("#define input_size %s \n", sh_var(sh, (struct pl_shader_var) {
+        GLSLH("#define input_size "$" \n", sh_var(sh, (struct pl_shader_var) {
             .var = pl_var_vec2("input_size"),
             .data = src_size,
         }));
 
         float dst_size[2] = { pl_rect_w(params->dst_rect), pl_rect_h(params->dst_rect) };
-        GLSLH("#define target_size %s \n", sh_var(sh, (struct pl_shader_var) {
+        GLSLH("#define target_size "$" \n", sh_var(sh, (struct pl_shader_var) {
             .var = pl_var_vec2("target_size"),
             .data = dst_size,
         }));
 
         float tex_off[2] = { params->src_rect.x0, params->src_rect.y0 };
-        GLSLH("#define tex_offset %s \n", sh_var(sh, (struct pl_shader_var) {
+        GLSLH("#define tex_offset "$" \n", sh_var(sh, (struct pl_shader_var) {
             .var = pl_var_vec2("tex_offset"),
             .data = tex_off,
         }));
@@ -1364,7 +1366,7 @@ static struct pl_hook_res hook_hook(void *priv, const struct pl_hook_params *par
             switch (hp->mode) {
             case PL_HOOK_PAR_VARIABLE:
             case PL_HOOK_PAR_DYNAMIC:
-                GLSLH("#define %s %s \n", hp->name,
+                GLSLH("#define %s "$" \n", hp->name,
                       sh_var(sh, (struct pl_shader_var) {
                         .var = {
                             .name = hp->name,
@@ -1379,7 +1381,7 @@ static struct pl_hook_res hook_hook(void *priv, const struct pl_hook_params *par
                 break;
 
             case PL_HOOK_PAR_CONSTANT:
-                GLSLH("#define %s %s \n", hp->name,
+                GLSLH("#define %s "$" \n", hp->name,
                       sh_const(sh, (struct pl_shader_const) {
                         .name = hp->name,
                         .type = hp->type,
@@ -1404,18 +1406,18 @@ static struct pl_hook_res hook_hook(void *priv, const struct pl_hook_params *par
             .gpu = p->gpu,
         ));
         pl_shader_linearize(p->trc_helper, params->orig_color);
-        GLSLH("#define linearize %s \n", sh_subpass(sh, p->trc_helper));
+        GLSLH("#define linearize "$" \n", sh_subpass(sh, p->trc_helper));
 
         pl_shader_reset(p->trc_helper, pl_shader_params(
             .id = ++sh_id,
             .gpu = p->gpu,
         ));
         pl_shader_delinearize(p->trc_helper, params->orig_color);
-        GLSLH("#define delinearize %s \n", sh_subpass(sh, p->trc_helper));
+        GLSLH("#define delinearize "$" \n", sh_subpass(sh, p->trc_helper));
 
         // Load and run the user shader itself
         sh_append_str(sh, SH_BUF_HEADER, hook->pass_body);
-        sh_describe(sh, pl_strdup0(SH_TMP(sh), hook->pass_desc));
+        sh_describef(sh, "%.*s", PL_STR_FMT(hook->pass_desc));
 
         // Resolve output size and create framebuffer
         float out_size[2] = {0};
@@ -1449,7 +1451,7 @@ static struct pl_hook_res hook_hook(void *priv, const struct pl_hook_params *par
                 goto error;
             }
 
-            GLSLP("#define out_image %s \n", sh_desc(sh, (struct pl_shader_desc) {
+            GLSLP("#define out_image "$" \n", sh_desc(sh, (struct pl_shader_desc) {
                 .binding.object = fbo,
                 .desc = {
                     .name = "out_image",
@@ -1458,7 +1460,7 @@ static struct pl_hook_res hook_hook(void *priv, const struct pl_hook_params *par
                 },
             }));
 
-            sh->res.output = PL_SHADER_SIG_NONE;
+            sh->output = PL_SHADER_SIG_NONE;
 
             GLSL("hook(); \n");
             ok = pl_dispatch_compute(params->dispatch, pl_dispatch_compute_params(
@@ -1495,7 +1497,7 @@ static struct pl_hook_res hook_hook(void *priv, const struct pl_hook_params *par
               x0 = sx * params->rect.x0 + hook->offset[0],
               y0 = sy * params->rect.y0 + hook->offset[1];
 
-        struct pl_rect2df new_rect = {
+        pl_rect2df new_rect = {
             x0,
             y0,
             x0 + sx * pl_rect_w(params->rect),
@@ -1678,10 +1680,17 @@ const struct pl_hook *pl_mpv_user_shader_parse(pl_gpu gpu,
 
     hook->parameters = p->hook_params.elem;
     hook->num_parameters = p->hook_params.num;
+
+    PL_MSG(gpu, PL_LOG_DEBUG, "Loaded user shader:");
+    pl_msg_source(gpu->log, PL_LOG_DEBUG, shader_text);
+
     return hook;
 
 error:
     pl_mpv_user_shader_destroy((const struct pl_hook **) &hook);
+    PL_MSG(gpu, PL_LOG_ERR, "Failed to parse user shader:");
+    pl_msg_source(gpu->log, PL_LOG_ERR, shader_text);
+    pl_log_stack_trace(gpu->log, PL_LOG_ERR);
     return NULL;
 }
 
