@@ -15,10 +15,11 @@
  * License along with libplacebo. If not, see <http://www.gnu.org/licenses/>.
  */
 
-#define _USE_MATH_DEFINES
 #include <math.h>
 #include "shaders.h"
 
+#include <libplacebo/shaders/dithering.h>
+
 const struct pl_dither_params pl_dither_default_params = { PL_DITHER_DEFAULTS };
 
 struct sh_dither_obj {
@@ -84,6 +85,7 @@ static inline float approx_gamma(enum pl_color_transfer trc)
     case PL_COLOR_TRC_GAMMA20:  return 2.0f;
     case PL_COLOR_TRC_GAMMA24:  return 2.4f;
     case PL_COLOR_TRC_GAMMA26:  return 2.6f;
+    case PL_COLOR_TRC_ST428:    return 2.6f;
     case PL_COLOR_TRC_GAMMA28:  return 2.8f;
 
     case PL_COLOR_TRC_SRGB:
@@ -116,7 +118,7 @@ void pl_shader_dither(pl_shader sh, int new_depth,
         return;
     }
 
-    sh_describe(sh, "dithering");
+    sh_describef(sh, "dithering (%d bits)", new_depth);
     GLSL("// pl_shader_dither \n"
         "{                    \n"
         "float bias;          \n");
@@ -128,16 +130,9 @@ void pl_shader_dither(pl_shader sh, int new_depth,
     }
 
     enum pl_dither_method method = params->method;
-    bool can_fixed = sh_glsl(sh).version >= 130;
-    ident_t lut = NULL;
+    ident_t lut = NULL_IDENT;
     int lut_size = 0;
 
-    if (method == PL_DITHER_ORDERED_FIXED && !can_fixed) {
-        PL_WARN(sh, "PL_DITHER_ORDERED_FIXED requires glsl version >= 130.."
-                " falling back.");
-        goto fallback;
-    }
-
     if (dither_method_is_lut(method)) {
         if (!dither_state) {
             PL_WARN(sh, "LUT-based dither method specified but no dither state "
@@ -172,7 +167,7 @@ void pl_shader_dither(pl_shader sh, int new_depth,
     goto done;
 
 fallback:
-    method = can_fixed ? PL_DITHER_ORDERED_FIXED : PL_DITHER_WHITE_NOISE;
+    method = PL_DITHER_ORDERED_FIXED;
     // fall through
 
 done: ;
@@ -186,7 +181,7 @@ done: ;
 
     if (size) {
         // Transform the screen position to the cyclic range [0,1)
-        GLSL("vec2 pos = fract(gl_FragCoord.xy * 1.0/%s);\n", SH_FLOAT(size));
+        GLSL("vec2 pos = fract(gl_FragCoord.xy * 1.0/"$"); \n", SH_FLOAT(size));
 
         if (params->temporal) {
             int phase = SH_PARAMS(sh).index % 8;
@@ -202,14 +197,14 @@ done: ;
                 .data = &mat[0][0],
                 .dynamic = true,
             });
-            GLSL("pos = fract(%s * pos + vec2(1.0));\n", rot);
+            GLSL("pos = fract("$" * pos + vec2(1.0));\n", rot);
         }
     }
 
     switch (method) {
     case PL_DITHER_WHITE_NOISE: {
         ident_t prng = sh_prng(sh, params->temporal, NULL);
-        GLSL("bias = %s.x;\n", prng);
+        GLSL("bias = "$".x;\n", prng);
         break;
     }
 
@@ -233,7 +228,7 @@ done: ;
     case PL_DITHER_BLUE_NOISE:
     case PL_DITHER_ORDERED_LUT:
         pl_assert(lut);
-        GLSL("bias = %s(ivec2(pos * %s));\n", lut, SH_FLOAT(lut_size));
+        GLSL("bias = "$"(ivec2(pos * "$"));\n", lut, SH_FLOAT(lut_size));
         break;
 
     case PL_DITHER_METHOD_COUNT:
@@ -245,7 +240,7 @@ done: ;
 
     const float gamma = approx_gamma(params->transfer);
     if (gamma != 1.0f && new_depth <= 4) {
-        GLSL("const float gamma = %s;                   \n"
+        GLSL("const float gamma = "$";                  \n"
              "vec4 color_lin = pow(color, vec4(gamma)); \n",
              SH_FLOAT(gamma));
 
@@ -268,8 +263,7 @@ done: ;
         }
 
         // Mix in the correct ratio corresponding to the offset and bias
-        GLSL("color = mix(low, high, %s(greaterThan(offset, vec4(bias)))); \n",
-             sh_bvec(sh, 4));
+        GLSL("color = mix(low, high, greaterThan(offset, vec4(bias))); \n");
     } else {
         // Approximate each gamma segment as a straight line, this simplifies
         // the process of dithering down to a single scale and (biased) round.
@@ -365,7 +359,7 @@ bool pl_shader_error_diffusion(pl_shader sh, const struct pl_error_diffusion_par
     // Figuring out how many block are required to process all pixels. We need
     // this explicitly to make the number of barrier() calls match.
     int block_size = PL_MIN(glsl.max_group_threads, height);
-    int blocks = (height * shifted_width + block_size - 1) / block_size;
+    int blocks = PL_DIV_UP(height * shifted_width, block_size);
 
     // If we figure out how many of the next columns will be affected while the
     // current columns is being processed. We can store errors of only a few
@@ -406,34 +400,36 @@ bool pl_shader_error_diffusion(pl_shader sh, const struct pl_error_diffusion_par
         },
     });
 
+    sh->output = PL_SHADER_SIG_NONE;
+    sh_describef(sh, "error diffusion (%s, %d bits)",
+                 kernel->name, params->new_depth);
+
     // Defines the ring buffer in shared memory.
-    sh->res.output = PL_SHADER_SIG_NONE;
-    sh_describe(sh, "error diffusion");
-    GLSLH("shared uint err_rgb8[%s]; \n", ring_buffer_size);
+    GLSLH("shared uint err_rgb8["$"]; \n", ring_buffer_size);
     GLSL("// pl_shader_error_diffusion                                          \n"
          // Safeguard against accidental over-execution
          "if (gl_WorkGroupID != uvec3(0))                                       \n"
          "    return;                                                           \n"
          // Initialize the ring buffer.
-         "for (uint i = gl_LocalInvocationIndex; i < %s; i+=gl_WorkGroupSize.x) \n"
+         "for (uint i = gl_LocalInvocationIndex; i < "$"; i+=gl_WorkGroupSize.x)\n"
          "    err_rgb8[i] = 0u;                                                 \n"
 
         // Main block loop, add barrier here to have previous block all
         // processed before starting the processing of the next.
-         "for (uint block_id = 0; block_id < %s; block_id++) {                  \n"
+         "for (uint block_id = 0; block_id < "$"; block_id++) {                 \n"
          "barrier();                                                            \n"
         // Compute the coordinate of the pixel we are currently processing,
         // both before and after the shift mapping.
          "uint id = block_id * gl_WorkGroupSize.x + gl_LocalInvocationIndex;    \n"
-         "const uint height = %s;                                               \n"
+         "const uint height = "$";                                              \n"
          "int y = int(id %% height), x_shifted = int(id / height);              \n"
          "int x = x_shifted - y * %d;                                           \n"
          // Proceed only if we are processing a valid pixel.
-         "if (x >= 0 && x < %s) {                                               \n"
+         "if (x >= 0 && x < "$") {                                              \n"
          // The index that the current pixel have on the ring buffer.
-         "uint idx = uint(x_shifted * %s + y) %% %s;                            \n"
+         "uint idx = uint(x_shifted * "$" + y) %% "$";                          \n"
          // Fetch the current pixel.
-         "vec4 pix_orig = texelFetch(%s, ivec2(x, y), 0);                       \n"
+         "vec4 pix_orig = texelFetch("$", ivec2(x, y), 0);                      \n"
          "vec3 pix = pix_orig.rgb;                                              \n",
          ring_buffer_size,
          SH_UINT(blocks),
@@ -480,7 +476,7 @@ bool pl_shader_error_diffusion(pl_shader sh, const struct pl_error_diffusion_par
          "err_rgb8[idx] = 0u;                                                   \n"
          // Write the dithered pixel.
          "vec3 dithered = round(pix);                                           \n"
-         "imageStore(%s, ivec2(x, y), vec4(dithered / %d.0, pix_orig.a));       \n"
+         "imageStore("$", ivec2(x, y), vec4(dithered / %d.0, pix_orig.a));      \n"
          // Prepare for error propagation pass
          "vec3 err_divided = (pix - dithered) * %d.0 / %d.0;                    \n"
          "ivec3 tmp;                                                            \n",
@@ -522,7 +518,7 @@ bool pl_shader_error_diffusion(pl_shader sh, const struct pl_error_diffusion_par
                 // Calculate the new position in the ring buffer to propagate
                 // the error into.
                 int ring_buffer_delta = shifted_x * ring_buffer_rows + y;
-                GLSL("atomicAdd(err_rgb8[(idx + %du) %% %s], err_u32); \n",
+                GLSL("atomicAdd(err_rgb8[(idx + %du) %% "$"], err_u32); \n",
                      ring_buffer_delta, ring_buffer_size);
             }
         }
