@@ -235,14 +235,6 @@ static inline int round2(int x, int shift)
     return (x + (1 << (shift - 1))) >> shift;
 }
 
-static ident_t sh_lut_quiet(pl_shader sh, const struct sh_lut_params *params)
-{
-    pl_log_level_cap(sh->log, PL_LOG_TRACE);
-    ident_t ret = sh_lut(sh, params);
-    pl_log_level_cap(sh->log, PL_LOG_NONE);
-    return ret;
-}
-
 enum {
     BLOCK_SIZE = 32,
     SCALING_LUT_SIZE = 256,
@@ -518,7 +510,7 @@ static void sample(pl_shader sh, enum offset off, ident_t lut, int idx,
     GLSL("offset = uvec2(%du, %du) * uvec2((data >> %d) & 0xFu, \n"
          "                                 (data >> %d) & 0xFu);\n"
          "pos = offset + local_id.xy + uvec2(%d, %d);           \n"
-         "val = %s(pos)%s;                                      \n",
+         "val = "$"(pos)%s;                                     \n",
          sub_x ? 1 : 2, sub_y ? 1 : 2, off + 4, off,
          (BLOCK_SIZE >> sub_x) * dx,
          (BLOCK_SIZE >> sub_y) * dy,
@@ -641,11 +633,6 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
         return false;
     }
 
-    if (sh_glsl(sh).version < 130) {
-        PL_ERR(sh, "AV1 film grain synthesis requires GLSL >= 130!");
-        return false;
-    }
-
     // Disable generation for unneeded component types
     fg_has_y &= tex_is_y;
     fg_has_u &= tex_is_cb;
@@ -680,7 +667,7 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
     int idx[3] = {-1};
 
     if (fg_has_y) {
-        lut[0] = sh_lut_quiet(sh, sh_lut_params(
+        lut[0] = sh_lut(sh, sh_lut_params(
             .object     = &obj->lut_grain[0],
             .var_type   = PL_VAR_FLOAT,
             .lut_type   = SH_LUT_TEXTURE,
@@ -715,7 +702,7 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
     }
 
     if (chroma_comps > 0) {
-        lut[1] = lut[2] = sh_lut_quiet(sh, sh_lut_params(
+        lut[1] = lut[2] = sh_lut(sh, sh_lut_params(
             .object     = &obj->lut_grain[1],
             .var_type   = PL_VAR_FLOAT,
             .lut_type   = SH_LUT_TEXTURE,
@@ -737,7 +724,7 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
             idx[1] = idx[2] = -1;
     }
 
-    ident_t offsets = sh_lut_quiet(sh, sh_lut_params(
+    ident_t offsets = sh_lut(sh, sh_lut_params(
         .object     = &obj->lut_offsets,
         .var_type   = PL_VAR_UINT,
         .lut_type   = SH_LUT_AUTO,
@@ -750,6 +737,11 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
         .priv       = (void *) &params->data,
     ));
 
+    if (!offsets) {
+        SH_FAIL(sh, "Failed generating/uploading block offsets LUT!");
+        return false;
+    }
+
     // For the scaling LUTs, we assume they'll be relatively constant
     // throughout the video so doing some extra work to avoid reinitializing
     // them constantly is probably worth it. Probably.
@@ -795,7 +787,7 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
         // Skip scaling for unneeded channels
         bool has_c[3] = { fg_has_y, fg_has_u, fg_has_v };
         if (has_c[i] && priv.num > 0) {
-            scaling[i] = sh_lut_quiet(sh, sh_lut_params(
+            scaling[i] = sh_lut(sh, sh_lut_params(
                 .object     = &obj->lut_scaling[i],
                 .var_type   = PL_VAR_FLOAT,
                 .method     = SH_LUT_LINEAR,
@@ -844,12 +836,12 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
     // Load the data vector which holds the offsets
     if (is_compute) {
         GLSLH("shared uint data; \n");
-        GLSL("if (gl_LocalInvocationIndex == 0u) \n"
-             "    data = uint(%s(block_id));     \n"
-             "barrier();                         \n",
+        GLSL("if (gl_LocalInvocationIndex == 0u)    \n"
+             "    data = uint("$"(block_id));       \n"
+             "barrier();                            \n",
              offsets);
     } else {
-        GLSL("uint data = uint(%s(block_id)); \n", offsets);
+        GLSL("uint data = uint("$"(block_id)); \n", offsets);
     }
 
     struct grain_scale scale = get_grain_scale(params);
@@ -881,7 +873,7 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
     });
 
     ident_t tex_scale = SH_FLOAT(scale.texture_scale);
-    GLSL("color = vec4(%s) * texelFetch(%s, ivec2(global_id), 0); \n",
+    GLSL("color = vec4("$") * texelFetch("$", ivec2(global_id), 0); \n",
          tex_scale, tex);
 
     // If we need access to the external luma plane, load it now
@@ -891,7 +883,7 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
             // We already have the luma channel as part of the pre-sampled color
             for (int i = 0; i < 3; i++) {
                 if (channel_map(i, params) == PL_CHANNEL_Y) {
-                    GLSL("averageLuma = color[%s]; \n", SH_INT(i));
+                    GLSL("averageLuma = color["$"]; \n", SH_INT(i));
                     break;
                 }
             }
@@ -907,9 +899,11 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
             });
 
             GLSL("pos = global_id * uvec2(%du, %du);                    \n"
-                 "averageLuma = %s * texelFetch(%s, ivec2(pos), 0)[%s]; \n",
-                 1 << sub_x, 1 << sub_y, tex_scale, luma,
-                 SH_INT(params->luma_comp));
+                 "averageLuma = texelFetch("$", ivec2(pos), 0)["$"];    \n"
+                 "averageLuma *= "$";                                   \n",
+                 1 << sub_x, 1 << sub_y,
+                 luma, SH_INT(params->luma_comp),
+                 tex_scale);
         }
     }
 
@@ -961,12 +955,12 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
                  "}                                             \n");
 
             // Correctly clip the interpolated grain
-            GLSL("grain = clamp(grain, %s, %s); \n", grain_min, grain_max);
+            GLSL("grain = clamp(grain, "$", "$"); \n", grain_min, grain_max);
         }
 
         if (c == PL_CHANNEL_Y) {
-            GLSL("color[%d] += %s(color[%d]) * grain;   \n"
-                 "color[%d] = clamp(color[%d], %s, %s); \n",
+            GLSL("color[%d] += "$"(color[%d]) * grain;      \n"
+                 "color[%d] = clamp(color[%d], "$", "$");   \n",
                  i, scaling[c], i,
                  i, i, minValue, maxLuma);
         } else {
@@ -983,17 +977,18 @@ bool pl_shader_fg_av1(pl_shader sh, pl_shader_obj *grain_state,
                     },
                 });
 
-                int c_offset = data->uv_offset[c - 1] << (bits - 8);
+                int c_offset = (unsigned) data->uv_offset[c - 1] << (bits - 8);
                 ident_t offset = sh_var(sh, (struct pl_shader_var) {
                     .var = pl_var_float("offset"),
                     .data = &(float) { c_offset * scale.grain_scale },
                 });
 
-                GLSL("val = dot(vec2(val, color[%d]), %s);  \n", i, mult);
-                GLSL("val += %s; \n", offset);
+                GLSL("val = dot(vec2(val, color[%d]), "$"); \n"
+                     "val += "$";                           \n",
+                     i, mult, offset);
             }
-            GLSL("color[%d] += %s(val) * grain;         \n"
-                 "color[%d] = clamp(color[%d], %s, %s); \n",
+            GLSL("color[%d] += "$"(val) * grain;            \n"
+                 "color[%d] = clamp(color[%d], "$", "$");   \n",
                  i, scaling[c],
                  i, i, minValue, maxChroma);
         }
