@@ -147,7 +147,7 @@ static bool detect_csp(pl_icc_object icc, struct pl_raw_primaries *prim,
     prim->white = pl_cie_from_XYZ(dst[WHITE].X, dst[WHITE].Y, dst[WHITE].Z);
 
     // Rough estimate of overall gamma and starting point for curve black point
-    const float y_approx = log(dst[GRAY].Y) / log(0.5);
+    const float y_approx = dst[GRAY].Y ? log(dst[GRAY].Y) / log(0.5) : 1.0f;
     const float kb = fmaxf(dst[BLACK].Y, 0.0f);
     float b = powf(kb, 1 / y_approx);
 
@@ -168,6 +168,8 @@ static bool detect_csp(pl_icc_object icc, struct pl_raw_primaries *prim,
         b = powf(kb, 1 / M);
     }
     S = sqrt(S / (k - 1));
+
+    PL_INFO(p, "Detected profile approximation gamma %.3f", M);
     if (S > 0.5) {
         PL_WARN(p, "Detected profile gamma (%.3f) very far from pure power "
                 "response (stddev=%.1f), suspected unusual or broken profile. "
@@ -184,27 +186,45 @@ static bool detect_csp(pl_icc_object icc, struct pl_raw_primaries *prim,
 }
 
 static bool detect_contrast(pl_icc_object icc, struct pl_hdr_metadata *hdr,
-                            float max_luma)
+                            struct pl_icc_params *params, float max_luma)
 {
     struct icc_priv *p = PL_PRIV(icc);
     cmsCIEXYZ *white = cmsReadTag(p->profile, cmsSigLuminanceTag);
-    enum pl_rendering_intent intent = icc->params.intent;
+    enum pl_rendering_intent intent = params->intent;
     /* LittleCMS refuses to detect an intent in absolute colorimetric intent,
      * so fall back to relative colorimetric since we only care about the
      * brightness value here */
     if (intent == PL_INTENT_ABSOLUTE_COLORIMETRIC)
         intent = PL_INTENT_RELATIVE_COLORIMETRIC;
     if (!cmsDetectDestinationBlackPoint(&p->black, p->profile, intent, 0)) {
+        /*
+         * v4 ICC profiles have a black point tag but only for
+         * perceptual/saturation intents. So we change the rendering intent
+         * to perceptual if we are provided a v4 ICC profile.
+         */
+        if (cmsGetEncodedICCversion(p->profile) >= 0x4000000 && intent != PL_INTENT_PERCEPTUAL) {
+            params->intent = PL_INTENT_PERCEPTUAL;
+            return detect_contrast(icc, hdr, params, max_luma);
+        }
+
         PL_ERR(p, "Failed detecting ICC profile black point!");
         return false;
     }
 
+    if (white) {
+        PL_DEBUG(p, "Detected raw white point X=%.2f Y=%.2f Z=%.2f cd/m^2",
+                 white->X, white->Y, white->Z);
+    }
+    PL_DEBUG(p, "Detected raw black point X=%.6f%% Y=%.6f%% Z=%.6f%%",
+             p->black.X * 100, p->black.Y * 100, p->black.Z * 100);
+
     if (max_luma <= 0)
         max_luma = white ? white->Y : PL_COLOR_SDR_WHITE;
 
     hdr->max_luma = max_luma;
     hdr->min_luma = p->black.Y * max_luma;
     hdr->min_luma = PL_MAX(hdr->min_luma, 1e-6); // prevent true 0
+    PL_INFO(p, "Using ICC contrast %.0f:1", hdr->max_luma / hdr->min_luma);
     return true;
 }
 
@@ -358,7 +378,7 @@ pl_icc_object pl_icc_open(pl_log log, const struct pl_icc_profile *profile,
     struct pl_raw_primaries *out_prim = &icc->csp.hdr.prim;
     if (!detect_csp(icc, out_prim, &icc->gamma))
         goto error;
-    if (!detect_contrast(icc, &icc->csp.hdr, params->max_luma))
+    if (!detect_contrast(icc, &icc->csp.hdr, params, params->max_luma))
         goto error;
     infer_clut_size(icc);
 
@@ -460,7 +480,7 @@ static void fill_lut(void *datap, const struct sh_lut_params *params, bool decod
         return;
     }
 
-    clock_t start = clock();
+    pl_clock_t start = pl_clock_now();
     cmsHTRANSFORM tf = cmsCreateTransformTHR(p->cms, srcp, TYPE_RGB_16,
                                              dstp, TYPE_RGBA_16,
                                              icc->params.intent,
@@ -469,7 +489,7 @@ static void fill_lut(void *datap, const struct sh_lut_params *params, bool decod
     if (!tf)
         return;
 
-    clock_t after_transform = clock();
+    pl_clock_t after_transform = pl_clock_now();
     pl_log_cpu_time(p->log, start, after_transform, "creating ICC transform");
 
     uint16_t *tmp = pl_alloc(NULL, s_r * 3 * sizeof(tmp[0]));
@@ -505,7 +525,7 @@ static void fill_lut(void *datap, const struct sh_lut_params *params, bool decod
         }
     }
 
-    pl_log_cpu_time(p->log, after_transform, clock(), "generating ICC 3DLUT");
+    pl_log_cpu_time(p->log, after_transform, pl_clock_now(), "generating ICC 3DLUT");
     cmsDeleteTransform(tf);
     pl_free(tmp);
 
@@ -554,13 +574,13 @@ void pl_icc_decode(pl_shader sh, pl_icc_object icc, pl_shader_obj *lut_obj,
 
     // Y = scale * (aX + b)^y
     sh_describe(sh, "ICC 3DLUT");
-    GLSL("// pl_icc_decode                      \n"
-         "{                                     \n"
-         "color.rgb = %s(color.rgb).rgb;        \n"
-         "color.rgb = %s * color.rgb + vec3(%s);\n"
-         "color.rgb = pow(color.rgb, vec3(%s)); \n"
-         "color.rgb = %s * color.rgb;           \n"
-         "}                                     \n",
+    GLSL("// pl_icc_decode                          \n"
+         "{                                         \n"
+         "color.rgb = "$"(color.rgb).rgb;           \n"
+         "color.rgb = "$" * color.rgb + vec3("$");  \n"
+         "color.rgb = pow(color.rgb, vec3("$"));    \n"
+         "color.rgb = "$" * color.rgb;              \n"
+         "}                                         \n",
          lut,
          SH_FLOAT(p->a), SH_FLOAT(p->b),
          SH_FLOAT(icc->gamma),
@@ -603,14 +623,14 @@ void pl_icc_encode(pl_shader sh, pl_icc_object icc, pl_shader_obj *lut_obj)
 
     // X = 1/a * (Y/scale)^(1/y) - b/a
     sh_describe(sh, "ICC 3DLUT");
-    GLSL("// pl_icc_encode                      \n"
-         "{                                     \n"
-         "color.rgb = max(color.rgb, 0.0);      \n"
-         "color.rgb = 1.0/%s * color.rgb;       \n"
-         "color.rgb = pow(color.rgb, vec3(%s)); \n"
-         "color.rgb = 1.0/%s * color.rgb - %s;  \n"
-         "color.rgb = %s(color.rgb).rgb;        \n"
-         "}                                     \n",
+    GLSL("// pl_icc_encode                          \n"
+         "{                                         \n"
+         "color.rgb = max(color.rgb, 0.0);          \n"
+         "color.rgb = 1.0/"$" * color.rgb;          \n"
+         "color.rgb = pow(color.rgb, vec3("$"));    \n"
+         "color.rgb = 1.0/"$" * color.rgb - "$";    \n"
+         "color.rgb = "$"(color.rgb).rgb;           \n"
+         "}                                         \n",
          SH_FLOAT(p->scale),
          SH_FLOAT(1.0f / icc->gamma),
          SH_FLOAT(p->a), SH_FLOAT(p->b / p->a),
