@@ -137,7 +137,7 @@ struct pl_custom_lut *pl_lut_parse_cube(pl_log log, const char *cstr, size_t cst
     lut->data = data;
 
     // Parse LUT body
-    clock_t start = clock();
+    pl_clock_t start = pl_clock_now();
     for (int n = 0; n < entries; n++) {
         for (int c = 0; c < 3; c++) {
             static const char * const digits = "0123456789.-+e";
@@ -177,7 +177,7 @@ struct pl_custom_lut *pl_lut_parse_cube(pl_log log, const char *cstr, size_t cst
     if (str.len)
         pl_warn(log, "Extra data after LUT?... ignoring '%c'", str.buf[0]);
 
-    pl_log_cpu_time(log, start, clock(), "parsing .cube LUT");
+    pl_log_cpu_time(log, start, pl_clock_now(), "parsing .cube LUT");
     return lut;
 
 error:
@@ -249,9 +249,9 @@ void pl_shader_custom_lut(pl_shader sh, const struct pl_custom_lut *lut,
 
     GLSL("// pl_shader_custom_lut \n");
 
-    static const struct pl_matrix3x3 zero = {0};
+    static const pl_matrix3x3 zero = {0};
     if (memcmp(&lut->shaper_in, &zero, sizeof(zero)) != 0) {
-        GLSL("color.rgb = %s * color.rgb; \n", sh_var(sh, (struct pl_shader_var) {
+        GLSL("color.rgb = "$" * color.rgb; \n", sh_var(sh, (struct pl_shader_var) {
             .var = pl_var_mat3("shaper_in"),
             .data = PL_TRANSPOSE_3X3(lut->shaper_in.m),
         }));
@@ -260,17 +260,19 @@ void pl_shader_custom_lut(pl_shader sh, const struct pl_custom_lut *lut,
     switch (dims) {
     case 1:
         sh_describe(sh, "custom 1DLUT");
-        GLSL("color.rgb = vec3(%s(color.r).r, %s(color.g).g, %s(color.b).b); \n",
+        GLSL("color.rgb = vec3("$"(color.r).r,  \n"
+             "                 "$"(color.g).g,  \n"
+             "                 "$"(color.b).b); \n",
              fun, fun, fun);
         break;
     case 3:
         sh_describe(sh, "custom 3DLUT");
-        GLSL("color.rgb = %s(color.rgb).rgb; \n", fun);
+        GLSL("color.rgb = "$"(color.rgb).rgb; \n", fun);
         break;
     }
 
     if (memcmp(&lut->shaper_out, &zero, sizeof(zero)) != 0) {
-        GLSL("color.rgb = %s * color.rgb; \n", sh_var(sh, (struct pl_shader_var) {
+        GLSL("color.rgb = "$" * color.rgb; \n", sh_var(sh, (struct pl_shader_var) {
             .var = pl_var_mat3("shaper_out"),
             .data = PL_TRANSPOSE_3X3(lut->shaper_out.m),
         }));
@@ -287,7 +289,8 @@ static ident_t texel_scale(pl_shader sh, int lut_size, bool normalized)
     const float scale = (end - base) / (normalized ? 1.0f : (lut_size - 1));
 
     ident_t name = sh_fresh(sh, "LUT_SCALE");
-    GLSLH("#define %s(x) (%s * (x) + %s) \n", name, SH_FLOAT(scale), SH_FLOAT(base));
+    GLSLH("#define "$"(x) ("$" * (x) + "$") \n",
+          name, SH_FLOAT(scale), SH_FLOAT(base));
     return name;
 }
 
@@ -337,6 +340,8 @@ ident_t sh_lut(pl_shader sh, const struct sh_lut_params *params)
     enum sh_lut_method method = params->method;
     if (method == SH_LUT_TETRAHEDRAL && dims != 3)
         method = SH_LUT_LINEAR;
+    if (method == SH_LUT_CUBIC && dims != 3)
+        method = SH_LUT_LINEAR;
 
     int texdim = 0;
     uint32_t max_tex_dim[] = {
@@ -349,7 +354,7 @@ ident_t sh_lut(pl_shader sh, const struct sh_lut_params *params)
                                     struct sh_lut_obj, sh_lut_uninit);
 
     if (!lut)
-        return NULL;
+        return NULL_IDENT;
 
     bool update = params->update || lut->signature != params->signature ||
                   vartype != lut->vartype || params->fmt != lut->fmt ||
@@ -357,7 +362,7 @@ ident_t sh_lut(pl_shader sh, const struct sh_lut_params *params)
                   params->depth != lut->depth || params->comps != lut->comps;
 
     if (lut->error && !update)
-        return NULL; // suppress error spam until something changes
+        return NULL_IDENT; // suppress error spam until something changes
 
     // Try picking the right number of dimensions for the texture LUT. This
     // allows e.g. falling back to 2D textures if 1D textures are unsupported.
@@ -383,7 +388,8 @@ next_dim: ; // `continue` out of the inner loop
     };
 
     enum pl_fmt_caps texcaps = PL_FMT_CAP_SAMPLEABLE;
-    if (method == SH_LUT_LINEAR)
+    bool is_linear = method == SH_LUT_LINEAR || method == SH_LUT_CUBIC;
+    if (is_linear)
         texcaps |= PL_FMT_CAP_LINEAR;
 
     pl_fmt texfmt = params->fmt;
@@ -418,7 +424,7 @@ next_dim: ; // `continue` out of the inner loop
     enum sh_lut_type type = params->lut_type;
 
     // The linear sampling code currently only supports 1D linear interpolation
-    if (method == SH_LUT_LINEAR && dims > 1) {
+    if (is_linear && dims > 1) {
         if (texfmt) {
             type = SH_LUT_TEXTURE;
         } else {
@@ -460,12 +466,19 @@ next_dim: ; // `continue` out of the inner loop
     update |= method != lut->method;
 
     if (update) {
-        PL_MSG(sh, params->dynamic ? PL_LOG_TRACE : PL_LOG_DEBUG,
-               "LUT cache invalidated, regenerating..");
+        if (params->dynamic)
+            pl_log_level_cap(sh->log, PL_LOG_TRACE);
 
+        PL_DEBUG(sh, "LUT cache invalidated, regenerating..");
         size_t buf_size = size * params->comps * pl_var_type_size(vartype);
         tmp = pl_zalloc(NULL, buf_size);
+
+        pl_clock_t start = pl_clock_now();
         params->fill(tmp, params);
+        pl_log_cpu_time(sh->log, start, pl_clock_now(), "generating shader LUT");
+
+        if (params->dynamic)
+            pl_log_level_cap(sh->log, PL_LOG_NONE);
 
         switch (type) {
         case SH_LUT_TEXTURE: {
@@ -579,7 +592,7 @@ next_dim: ; // `continue` out of the inner loop
 
     // Done updating, generate the GLSL
     ident_t name = sh_fresh(sh, "lut");
-    ident_t arr_name = NULL;
+    ident_t arr_name = NULL_IDENT;
 
     static const char * const swizzles[] = {"x", "xy", "xyz", "xyzw"};
     static const char * const vartypes[PL_VAR_TYPE_COUNT][4] = {
@@ -598,38 +611,35 @@ next_dim: ; // `continue` out of the inner loop
             },
             .binding = {
                 .object = lut->tex,
-                .sample_mode = method == SH_LUT_LINEAR
-                                    ? PL_TEX_SAMPLE_LINEAR
-                                    : PL_TEX_SAMPLE_NEAREST,
+                .sample_mode = is_linear ? PL_TEX_SAMPLE_LINEAR
+                                         : PL_TEX_SAMPLE_NEAREST,
             }
         });
 
-        bool can_fetch = sh_glsl(sh).version >= 130;
-        if (method == SH_LUT_LINEAR || !can_fetch) {
+        if (is_linear) {
             ident_t pos_macros[PL_ARRAY_SIZE(sizes)] = {0};
             for (int i = 0; i < dims; i++)
-                pos_macros[i] = texel_scale(sh, sizes[i], method == SH_LUT_LINEAR);
+                pos_macros[i] = texel_scale(sh, sizes[i], true);
 
-            GLSLH("#define %s(pos) (%s(%s, %s(\\\n",
-                  name, sh_tex_fn(sh, lut->tex->params),
-                  tex, vartypes[PL_VAR_FLOAT][texdim - 1]);
+            GLSLH("#define "$"(pos) (textureLod("$", %s(\\\n",
+                  name, tex, vartypes[PL_VAR_FLOAT][texdim - 1]);
 
             for (int i = 0; i < texdim; i++) {
                 char sep = i == 0 ? ' ' : ',';
                 if (pos_macros[i]) {
                     if (dims > 1) {
-                        GLSLH("   %c%s(%s(pos).%c)\\\n", sep, pos_macros[i],
+                        GLSLH("   %c"$"(%s(pos).%c)\\\n", sep, pos_macros[i],
                               vartypes[PL_VAR_FLOAT][dims - 1], "xyzw"[i]);
                     } else {
-                        GLSLH("   %c%s(float(pos))\\\n", sep, pos_macros[i]);
+                        GLSLH("   %c"$"(float(pos))\\\n", sep, pos_macros[i]);
                     }
                 } else {
                     GLSLH("   %c%f\\\n", sep, 0.5);
                 }
             }
-            GLSLH("  )).%s)\n", swizzles[params->comps - 1]);
+            GLSLH("  ), 0.0).%s)\n", swizzles[params->comps - 1]);
         } else {
-            GLSLH("#define %s(pos) (texelFetch(%s, %s(pos",
+            GLSLH("#define "$"(pos) (texelFetch("$", %s(pos",
                   name, tex, vartypes[PL_VAR_SINT][texdim - 1]);
 
             // Fill up extra components of the index
@@ -656,7 +666,7 @@ next_dim: ; // `continue` out of the inner loop
 
     case SH_LUT_LITERAL:
         arr_name = sh_fresh(sh, "weights");
-        GLSLH("const %s %s[%d] = %s[](\n  ",
+        GLSLH("const %s "$"[%d] = %s[](\n  ",
               vartypes[vartype][params->comps - 1], arr_name, size,
               vartypes[vartype][params->comps - 1]);
         sh_append_str(sh, SH_BUF_HEADER, lut->str);
@@ -668,7 +678,7 @@ next_dim: ; // `continue` out of the inner loop
     }
 
     if (arr_name) {
-        GLSLH("#define %s(pos) (%s[int((pos)%s)\\\n",
+        GLSLH("#define "$"(pos) ("$"[int((pos)%s)\\\n",
               name, arr_name, dims > 1 ? "[0]" : "");
         int shift = params->width;
         for (int i = 1; i < dims; i++) {
@@ -677,17 +687,17 @@ next_dim: ; // `continue` out of the inner loop
         }
         GLSLH("  ])\n");
 
-        if (method == SH_LUT_LINEAR) {
+        if (is_linear) {
             pl_assert(dims == 1);
             pl_assert(vartype == PL_VAR_FLOAT);
             ident_t arr_lut = name;
             name = sh_fresh(sh, "lut_lin");
-            GLSLH("%s %s(float fpos) {                              \n"
+            GLSLH("%s "$"(float fpos) {                             \n"
                   "    fpos = clamp(fpos, 0.0, 1.0) * %d.0;         \n"
                   "    float fbase = floor(fpos);                   \n"
                   "    float fceil = ceil(fpos);                    \n"
                   "    float fcoord = fpos - fbase;                 \n"
-                  "    return mix(%s(fbase), %s(fceil), fcoord);    \n"
+                  "    return mix("$"(fbase), "$"(fceil), fcoord);  \n"
                   "}                                                \n",
                   vartypes[PL_VAR_FLOAT][params->comps - 1], name,
                   size - 1,
@@ -695,25 +705,69 @@ next_dim: ; // `continue` out of the inner loop
         }
     }
 
+    if (method == SH_LUT_CUBIC && dims == 3) {
+        ident_t lin_lut = name;
+        name = sh_fresh(sh, "lut_tricubic");
+        GLSLH("%s "$"(vec3 pos) {                                       \n"
+              "    vec3 scale = vec3(%d.0, %d.0, %d.0);                 \n"
+              "    vec3 scale_inv = 1.0 / scale;                        \n"
+              "    pos *= scale;                                        \n"
+              "    vec3 fpos = fract(pos);                              \n"
+              "    vec3 base = pos - fpos;                              \n"
+              "    vec3 fpos2 = fpos * fpos;                            \n"
+              "    vec3 inv = 1.0 - fpos;                               \n"
+              "    vec3 inv2 = inv * inv;                               \n"
+              "    vec3 w0 = 1.0/6.0 * inv2 * inv;                      \n"
+              "    vec3 w1 = 2.0/3.0 - 0.5 * fpos2 * (2.0 - fpos);      \n"
+              "    vec3 w2 = 2.0/3.0 - 0.5 * inv2 * (2.0 - inv);        \n"
+              "    vec3 w3 = 1.0/6.0 * fpos2 * fpos;                    \n"
+              "    vec3 g0 = w0 + w1;                                   \n"
+              "    vec3 g1 = w2 + w3;                                   \n"
+              "    vec3 h0 = scale_inv * ((w1 / g0) - 1.0 + base);      \n"
+              "    vec3 h1 = scale_inv * ((w3 / g1) + 1.0 + base);      \n"
+              "    %s c000, c001, c010, c011, c100, c101, c110, c111;   \n"
+              "    c000 = "$"(h0);                                      \n"
+              "    c100 = "$"(vec3(h1.x, h0.y, h0.z));                  \n"
+              "    c000 = mix(c100, c000, g0.x);                        \n"
+              "    c010 = "$"(vec3(h0.x, h1.y, h0.z));                  \n"
+              "    c110 = "$"(vec3(h1.x, h1.y, h0.z));                  \n"
+              "    c010 = mix(c110, c010, g0.x);                        \n"
+              "    c000 = mix(c010, c000, g0.y);                        \n"
+              "    c001 = "$"(vec3(h0.x, h0.y, h1.z));                  \n"
+              "    c101 = "$"(vec3(h1.x, h0.y, h1.z));                  \n"
+              "    c001 = mix(c101, c001, g0.x);                        \n"
+              "    c011 = "$"(vec3(h0.x, h1.y, h1.z));                  \n"
+              "    c111 = "$"(h1);                                      \n"
+              "    c011 = mix(c111, c011, g0.x);                        \n"
+              "    c001 = mix(c011, c001, g0.y);                        \n"
+              "    return mix(c001, c000, g0.z);                        \n"
+              "}                                                        \n",
+              vartypes[PL_VAR_FLOAT][params->comps - 1], name,
+              sizes[0] - 1, sizes[1] - 1, sizes[2] - 1,
+              vartypes[PL_VAR_FLOAT][params->comps - 1],
+              lin_lut, lin_lut, lin_lut, lin_lut,
+              lin_lut, lin_lut, lin_lut, lin_lut);
+    }
+
     if (method == SH_LUT_TETRAHEDRAL) {
         ident_t int_lut = name;
         name = sh_fresh(sh, "lut_barycentric");
-        GLSLH("%s %s(vec3 pos) {                                            \n"
-              // Compute bounding vertices and fractinoal part
-              "    pos = clamp(pos, 0.0, 1.0) * vec3(%d.0, %d.0, %d.0);     \n"
-              "    vec3 base = floor(pos);                                  \n"
-              "    vec3 fpart = pos - base;                                 \n"
+        GLSLH("%s "$"(vec3 pos) {                                       \n"
+              // Compute bounding vertices and fractional part
+              "    pos = clamp(pos, 0.0, 1.0) * vec3(%d.0, %d.0, %d.0); \n"
+              "    vec3 base = floor(pos);                              \n"
+              "    vec3 fpart = pos - base;                             \n"
               // v0 and v3 are always 'black' and 'white', respectively
               // v1 and v2 are the closest RGB and CMY vertices, respectively
-              "    ivec3 v0 = ivec3(base), v3 = ivec3(ceil(pos));           \n"
-              "    ivec3 v1 = v0, v2 = v3;                                  \n"
+              "    ivec3 v0 = ivec3(base), v3 = ivec3(ceil(pos));       \n"
+              "    ivec3 v1 = v0, v2 = v3;                              \n"
               // Table of boolean checks to simplify following math
-              "    bvec3 c = greaterThanEqual(fpart.xyz, fpart.yzx);        \n"
-              "    bool c_xy = c.x, c_yx = !c.x,                            \n"
-              "       c_yz = c.y, c_zy = !c.y,                              \n"
-              "       c_zx = c.z, c_xz = !c.z;                              \n"
-              "    vec3 s = fpart.xyz;                                      \n"
-              "    bool cond;                                               \n",
+              "    bvec3 c = greaterThanEqual(fpart.xyz, fpart.yzx);    \n"
+              "    bool c_xy = c.x, c_yx = !c.x,                        \n"
+              "       c_yz = c.y, c_zy = !c.y,                          \n"
+              "       c_zx = c.z, c_xz = !c.z;                          \n"
+              "    vec3 s = fpart.xyz;                                  \n"
+              "    bool cond;                                           \n",
               vartypes[PL_VAR_FLOAT][params->comps - 1], name,
               sizes[0] - 1, sizes[1] - 1, sizes[2] - 1);
 
@@ -736,10 +790,10 @@ next_dim: ; // `continue` out of the inner loop
         }
 
         // Interpolate in barycentric coordinates, with four texel fetches
-        GLSLH("    return (1.0 - s.x) * %s(v0) +    \n"
-              "           (s.x - s.y) * %s(v1) +    \n"
-              "           (s.y - s.z) * %s(v2) +    \n"
-              "           (s.z)       * %s(v3);     \n"
+        GLSLH("    return (1.0 - s.x) * "$"(v0) +   \n"
+              "           (s.x - s.y) * "$"(v1) +   \n"
+              "           (s.y - s.z) * "$"(v2) +   \n"
+              "           (s.z)       * "$"(v3);    \n"
               "}                                    \n",
               int_lut, int_lut, int_lut, int_lut);
     }
@@ -752,5 +806,5 @@ next_dim: ; // `continue` out of the inner loop
 error:
     lut->error = true;
     pl_free(tmp);
-    return NULL;
+    return NULL_IDENT;
 }
