@@ -81,7 +81,7 @@ int main()
     REQUIRE(pl_primaries_superset(bt2020, bt709));
     REQUIRE(!pl_primaries_superset(bt709, bt2020));
 
-    struct pl_matrix3x3 rgb2xyz, rgb2xyz_;
+    pl_matrix3x3 rgb2xyz, rgb2xyz_;
     rgb2xyz = rgb2xyz_ = pl_get_rgb2xyz_matrix(bt709);
     pl_matrix3x3_invert(&rgb2xyz_);
     pl_matrix3x3_invert(&rgb2xyz_);
@@ -105,7 +105,7 @@ int main()
     REQUIRE_FEQ(rgb2xyz.m[2][2], pl_cie_Z(bt709->blue) * Y, 1e-7);
 
     // Make sure the gamut mapping round-trips
-    struct pl_matrix3x3 bt709_bt2020, bt2020_bt709;
+    pl_matrix3x3 bt709_bt2020, bt2020_bt709;
     bt709_bt2020 = pl_get_color_mapping_matrix(bt709, bt2020, PL_INTENT_RELATIVE_COLORIMETRIC);
     bt2020_bt709 = pl_get_color_mapping_matrix(bt2020, bt709, PL_INTENT_RELATIVE_COLORIMETRIC);
     for (int n = 0; n < 10; n++) {
@@ -134,7 +134,7 @@ int main()
         };
 
         float scale = pl_color_repr_normalize(&repr);
-        struct pl_transform3x3 yuv2rgb = pl_color_repr_decode(&repr, NULL);
+        pl_transform3x3 yuv2rgb = pl_color_repr_decode(&repr, NULL);
         pl_matrix3x3_scale(&yuv2rgb.mat, scale);
 
         static const float white_ycbcr[3] = { 235/1023., 128/1023., 128/1023. };
@@ -169,7 +169,7 @@ int main()
     bt709_d50 = *pl_raw_primaries_get(PL_COLOR_PRIM_BT_709);
     bt709_d50.white = (struct pl_cie_xy) { 0.34567, 0.35850 };
 
-    struct pl_matrix3x3 d50_d65;
+    pl_matrix3x3 d50_d65;
     d50_d65 = pl_get_color_mapping_matrix(&bt709_d50, bt709, PL_INTENT_RELATIVE_COLORIMETRIC);
 
     float white[3] = { 1.0, 1.0, 1.0 };
@@ -181,7 +181,7 @@ int main()
     // Simulate a typical 10-bit YCbCr -> 16 bit texture conversion
     tv_repr.bits.color_depth  = 10;
     tv_repr.bits.sample_depth = 16;
-    struct pl_transform3x3 yuv2rgb;
+    pl_transform3x3 yuv2rgb;
     yuv2rgb = pl_color_repr_decode(&tv_repr, NULL);
     float test[3] = { 575/65535., 336/65535., 640/65535. };
     pl_transform3x3_apply(&yuv2rgb, test);
@@ -229,12 +229,6 @@ int main()
     REQUIRE(pl_color_primaries_is_wide_gamut(PL_COLOR_PRIM_V_GAMUT));
     REQUIRE(pl_color_primaries_is_wide_gamut(PL_COLOR_PRIM_S_GAMUT));
 
-    REQUIRE(!pl_color_light_is_scene_referred(PL_COLOR_LIGHT_UNKNOWN));
-    REQUIRE(!pl_color_light_is_scene_referred(PL_COLOR_LIGHT_DISPLAY));
-    REQUIRE(pl_color_light_is_scene_referred(PL_COLOR_LIGHT_SCENE_HLG));
-    REQUIRE(pl_color_light_is_scene_referred(PL_COLOR_LIGHT_SCENE_709_1886));
-    REQUIRE(pl_color_light_is_scene_referred(PL_COLOR_LIGHT_SCENE_1_2));
-
     struct pl_color_space space = pl_color_space_unknown;
     pl_color_space_merge(&space, &pl_color_space_bt709);
     REQUIRE(pl_color_space_equal(&space, &pl_color_space_bt709));
@@ -246,7 +240,7 @@ int main()
     };
 
     pl_color_space_infer(&hlg);
-    REQUIRE_CMP(hlg.hdr.max_luma, ==, 1000.0f, "f");
+    REQUIRE_CMP(hlg.hdr.max_luma, ==, PL_COLOR_HLG_PEAK, "f");
 
     struct pl_color_space unknown = {0};
     struct pl_color_space display = {
@@ -283,11 +277,11 @@ int main()
 #define TEST_CONE(model, color)                                                 \
     do {                                                                        \
         float tmp[3] = { (color)[0], (color)[1], (color)[2] };                  \
-        struct pl_matrix3x3 mat = pl_get_cone_matrix(&(model), bt709);          \
+        pl_matrix3x3 mat = pl_get_cone_matrix(&(model), bt709);                 \
         pl_matrix3x3_apply(&mat, tmp);                                          \
         printf("%s + %s = %f %f %f\n", #model, #color, tmp[0], tmp[1], tmp[2]); \
         for (int i = 0; i < 3; i++)                                             \
-            REQUIRE_FEQ((color)[i], tmp[i], 1e-6);                              \
+            REQUIRE_FEQ((color)[i], tmp[i], 1e-5f);                             \
     } while(0)
 
     struct pl_cone_params red_only = { .cones = PL_CONE_MS };
@@ -325,4 +319,150 @@ int main()
     struct pl_cie_xy d55 = pl_white_from_temp(5503);
     REQUIRE_FEQ(d55.x, 0.33242, 1e-3);
     REQUIRE_FEQ(d55.y, 0.34743, 1e-3);
+
+    // Make sure we infer the correct set of metadata parameters
+#define TEST_METADATA(CSP, TYPE, MIN, MAX, AVG)                             \
+    do {                                                                    \
+        float _min, _max, _avg;                                             \
+        pl_color_space_nominal_luma_ex(pl_nominal_luma_params(              \
+            .color    = &(CSP),                                             \
+            .metadata = TYPE,                                               \
+            .scaling  = PL_HDR_PQ,                                          \
+            .out_min  = &_min,                                              \
+            .out_max  = &_max,                                              \
+            .out_avg  = &_avg,                                              \
+        ));                                                                 \
+        const float _min_ref = pl_hdr_rescale(PL_HDR_NITS, PL_HDR_PQ, MIN); \
+        const float _max_ref = pl_hdr_rescale(PL_HDR_NITS, PL_HDR_PQ, MAX); \
+        const float _avg_ref = pl_hdr_rescale(PL_HDR_NITS, PL_HDR_PQ, AVG); \
+        REQUIRE_FEQ(_min, _min_ref, 1e-5);                                  \
+        REQUIRE_FEQ(_max, _max_ref, 1e-5);                                  \
+        REQUIRE_FEQ(_avg, _avg_ref, 1e-5);                                  \
+    } while (0)
+
+    const struct pl_color_space hdr10plus = {
+        .primaries = PL_COLOR_PRIM_BT_2020,
+        .transfer  = PL_COLOR_TRC_PQ,
+        .hdr = {
+            .min_luma  = 0.005,
+            .max_luma  = 4000,
+            .scene_max = {596.69, 1200, 500},
+            .scene_avg = 300,
+        },
+    };
+
+    REQUIRE(pl_hdr_metadata_contains(&hdr10plus.hdr, PL_HDR_METADATA_ANY));
+    REQUIRE(pl_hdr_metadata_contains(&hdr10plus.hdr, PL_HDR_METADATA_NONE));
+    REQUIRE(pl_hdr_metadata_contains(&hdr10plus.hdr, PL_HDR_METADATA_HDR10));
+    REQUIRE(pl_hdr_metadata_contains(&hdr10plus.hdr, PL_HDR_METADATA_HDR10PLUS));
+    REQUIRE(!pl_hdr_metadata_contains(&hdr10plus.hdr, PL_HDR_METADATA_CIE_Y));
+
+    TEST_METADATA(hdr10plus, PL_HDR_METADATA_NONE,      PL_COLOR_HDR_BLACK, 10000, 0);
+    TEST_METADATA(hdr10plus, PL_HDR_METADATA_CIE_Y,     PL_COLOR_HDR_BLACK, 4000, 0);
+    TEST_METADATA(hdr10plus, PL_HDR_METADATA_HDR10,     PL_COLOR_HDR_BLACK, 4000, 0);
+    TEST_METADATA(hdr10plus, PL_HDR_METADATA_HDR10PLUS, PL_COLOR_HDR_BLACK, 1000, 250);
+    TEST_METADATA(hdr10plus, PL_HDR_METADATA_ANY,       PL_COLOR_HDR_BLACK, 1000, 250);
+
+    const struct pl_color_space dovi = {
+        .primaries = PL_COLOR_PRIM_BT_2020,
+        .transfer  = PL_COLOR_TRC_PQ,
+        .hdr = {
+            .min_luma = 0.005,
+            .max_luma = 4000,
+            .max_pq_y = pl_hdr_rescale(PL_HDR_NITS, PL_HDR_PQ, 1000),
+            .avg_pq_y = pl_hdr_rescale(PL_HDR_NITS, PL_HDR_PQ, 250),
+        },
+    };
+
+    REQUIRE(pl_hdr_metadata_contains(&dovi.hdr, PL_HDR_METADATA_ANY));
+    REQUIRE(pl_hdr_metadata_contains(&dovi.hdr, PL_HDR_METADATA_NONE));
+    REQUIRE(pl_hdr_metadata_contains(&dovi.hdr, PL_HDR_METADATA_HDR10));
+    REQUIRE(pl_hdr_metadata_contains(&dovi.hdr, PL_HDR_METADATA_CIE_Y));
+    REQUIRE(!pl_hdr_metadata_contains(&dovi.hdr, PL_HDR_METADATA_HDR10PLUS));
+
+    TEST_METADATA(dovi, PL_HDR_METADATA_NONE,      PL_COLOR_HDR_BLACK, 10000, 0);
+    TEST_METADATA(dovi, PL_HDR_METADATA_HDR10,     PL_COLOR_HDR_BLACK, 4000, 0);
+    TEST_METADATA(dovi, PL_HDR_METADATA_HDR10PLUS, PL_COLOR_HDR_BLACK, 4000, 0);
+    TEST_METADATA(dovi, PL_HDR_METADATA_CIE_Y,     PL_COLOR_HDR_BLACK, 1000, 250);
+    TEST_METADATA(dovi, PL_HDR_METADATA_ANY,       PL_COLOR_HDR_BLACK, 1000, 250);
+
+    const struct pl_color_space hlg4000 = {
+        .primaries    = PL_COLOR_PRIM_BT_2020,
+        .transfer     = PL_COLOR_TRC_HLG,
+        .hdr.max_luma = 4000,
+        .hdr.min_luma = 0.005,
+    };
+
+    TEST_METADATA(hlg4000, PL_HDR_METADATA_NONE,  PL_COLOR_HDR_BLACK, PL_COLOR_HLG_PEAK, 0);
+    TEST_METADATA(hlg4000, PL_HDR_METADATA_HDR10, 0.005, 4000, 0);
+    TEST_METADATA(hlg4000, PL_HDR_METADATA_ANY,   0.005, 4000, 0);
+
+    const struct pl_color_space untagged = {
+        .primaries = PL_COLOR_PRIM_BT_709,
+        .transfer  = PL_COLOR_TRC_BT_1886,
+    };
+
+    REQUIRE(pl_hdr_metadata_contains(&untagged.hdr, PL_HDR_METADATA_NONE));
+    REQUIRE(!pl_hdr_metadata_contains(&untagged.hdr, PL_HDR_METADATA_ANY));
+    REQUIRE(!pl_hdr_metadata_contains(&untagged.hdr, PL_HDR_METADATA_HDR10));
+    REQUIRE(!pl_hdr_metadata_contains(&untagged.hdr, PL_HDR_METADATA_CIE_Y));
+    REQUIRE(!pl_hdr_metadata_contains(&untagged.hdr, PL_HDR_METADATA_HDR10PLUS));
+
+    const float sdr_black = PL_COLOR_SDR_WHITE / PL_COLOR_SDR_CONTRAST;
+    TEST_METADATA(untagged, PL_HDR_METADATA_NONE, sdr_black, PL_COLOR_SDR_WHITE, 0);
+    TEST_METADATA(untagged, PL_HDR_METADATA_ANY,  sdr_black, PL_COLOR_SDR_WHITE, 0);
+
+    const struct pl_color_space sdr50 = {
+        .primaries    = PL_COLOR_PRIM_BT_709,
+        .transfer     = PL_COLOR_TRC_BT_1886,
+        .hdr.max_luma = 50,
+    };
+
+    REQUIRE(pl_hdr_metadata_contains(&sdr50.hdr, PL_HDR_METADATA_NONE));
+    REQUIRE(pl_hdr_metadata_contains(&sdr50.hdr, PL_HDR_METADATA_ANY));
+    REQUIRE(pl_hdr_metadata_contains(&sdr50.hdr, PL_HDR_METADATA_HDR10));
+    REQUIRE(!pl_hdr_metadata_contains(&sdr50.hdr, PL_HDR_METADATA_CIE_Y));
+    REQUIRE(!pl_hdr_metadata_contains(&sdr50.hdr, PL_HDR_METADATA_HDR10PLUS));
+
+    TEST_METADATA(sdr50, PL_HDR_METADATA_NONE,  sdr_black, PL_COLOR_SDR_WHITE, 0);
+    TEST_METADATA(sdr50, PL_HDR_METADATA_HDR10, 50 / PL_COLOR_SDR_CONTRAST, 50, 0);
+    TEST_METADATA(sdr50, PL_HDR_METADATA_ANY,   50 / PL_COLOR_SDR_CONTRAST, 50, 0);
+
+    const struct pl_color_space sdr10k = {
+        .primaries    = PL_COLOR_PRIM_BT_709,
+        .transfer     = PL_COLOR_TRC_BT_1886,
+        .hdr.min_luma = PL_COLOR_SDR_WHITE / 10000,
+    };
+
+    REQUIRE(pl_hdr_metadata_contains(&sdr10k.hdr, PL_HDR_METADATA_NONE));
+    REQUIRE(!pl_hdr_metadata_contains(&sdr10k.hdr, PL_HDR_METADATA_ANY));
+    REQUIRE(!pl_hdr_metadata_contains(&sdr10k.hdr, PL_HDR_METADATA_HDR10));
+    TEST_METADATA(sdr10k, PL_HDR_METADATA_NONE,  sdr_black, PL_COLOR_SDR_WHITE, 0);
+    TEST_METADATA(sdr10k, PL_HDR_METADATA_HDR10, PL_COLOR_SDR_WHITE / 10000, PL_COLOR_SDR_WHITE, 0);
+    TEST_METADATA(sdr10k, PL_HDR_METADATA_ANY,   PL_COLOR_SDR_WHITE / 10000, PL_COLOR_SDR_WHITE, 0);
+
+    const struct pl_color_space bogus_vals = {
+        .primaries = PL_COLOR_PRIM_BT_2020,
+        .transfer  = PL_COLOR_TRC_HLG,
+        .hdr.min_luma = 1e-9,
+        .hdr.max_luma = 1000000,
+    };
+
+    const struct pl_color_space bogus_flip = {
+        .primaries = PL_COLOR_PRIM_BT_2020,
+        .transfer  = PL_COLOR_TRC_PQ,
+        .hdr.min_luma = 4000,
+        .hdr.max_luma = 0.05,
+    };
+
+    const struct pl_color_space bogus_sign = {
+        .primaries = PL_COLOR_PRIM_BT_2020,
+        .transfer  = PL_COLOR_TRC_HLG,
+        .hdr.min_luma = -0.5,
+        .hdr.max_luma = -4000,
+    };
+
+    TEST_METADATA(bogus_vals, PL_HDR_METADATA_HDR10, PL_COLOR_HDR_BLACK, 10000, 0);
+    TEST_METADATA(bogus_flip, PL_HDR_METADATA_HDR10, PL_COLOR_HDR_BLACK, 10000, 0);
+    TEST_METADATA(bogus_sign, PL_HDR_METADATA_HDR10, PL_COLOR_HDR_BLACK, PL_COLOR_HLG_PEAK, 0);
 }
