@@ -1,5 +1,7 @@
 #include "tests.h"
 
+#include <libplacebo/filters.h>
+
 int main()
 {
     pl_log log = pl_test_logger();
@@ -18,19 +20,19 @@ int main()
 
         if (params.config.polar) {
             // Ensure the kernel seems sanely scaled
-            REQUIRE(feq(flt->weights[0], 1.0, 1e-7));
-            REQUIRE(feq(flt->weights[params.lut_entries - 1], 0.0, 1e-7));
+            REQUIRE_FEQ(flt->weights[0], 1.0, 1e-7);
+            REQUIRE_FEQ(flt->weights[params.lut_entries - 1], 0.0, 1e-7);
         } else {
             // Ensure the weights for each row add up to unity
             for (int i = 0; i < params.lut_entries; i++) {
                 float sum = 0.0;
                 REQUIRE(flt->row_size);
-                REQUIRE(flt->row_stride >= flt->row_size);
+                REQUIRE_CMP(flt->row_stride, >=, flt->row_size, "d");
                 for (int n = 0; n < flt->row_size; n++) {
                     float w = flt->weights[i * flt->row_stride + n];
                     sum += w;
                 }
-                REQUIRE(feq(sum, 1.0, 1e-6));
+                REQUIRE_FEQ(sum, 1.0, 1e-6);
             }
         }
 
