@@ -1,6 +1,12 @@
 #include "tests.h"
 #include "shaders.h"
 
+#include <libplacebo/renderer.h>
+#include <libplacebo/utils/frame_queue.h>
+#include <libplacebo/utils/upload.h>
+
+//#define PRINT_OUTPUT
+
 static void pl_buffer_tests(pl_gpu gpu)
 {
     const size_t buf_size = 1024;
@@ -12,69 +18,115 @@ static void pl_buffer_tests(pl_gpu gpu)
     assert(test_src && test_dst);
     memset(test_dst, 0, buf_size);
     for (int i = 0; i < buf_size; i++)
-        test_src[i] = (RANDOM * 256);
+        test_src[i] = RANDOM_U8;
 
     pl_buf buf = NULL, tbuf = NULL;
 
     printf("test buffer static creation and readback\n");
-    buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+    buf = pl_buf_create(gpu, pl_buf_params(
         .size = buf_size,
         .host_readable = true,
         .initial_data = test_src,
-    });
+    ));
 
     REQUIRE(buf);
     REQUIRE(pl_buf_read(gpu, buf, 0, test_dst, buf_size));
-    REQUIRE(memcmp(test_src, test_dst, buf_size) == 0);
+    REQUIRE_MEMEQ(test_src, test_dst, buf_size);
     pl_buf_destroy(gpu, &buf);
 
     printf("test buffer empty creation, update and readback\n");
     memset(test_dst, 0, buf_size);
-    buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+    buf = pl_buf_create(gpu, pl_buf_params(
         .size = buf_size,
         .host_writable = true,
         .host_readable = true,
-    });
+    ));
 
     REQUIRE(buf);
     pl_buf_write(gpu, buf, 0, test_src, buf_size);
     REQUIRE(pl_buf_read(gpu, buf, 0, test_dst, buf_size));
-    REQUIRE(memcmp(test_src, test_dst, buf_size) == 0);
+    REQUIRE_MEMEQ(test_src, test_dst, buf_size);
     pl_buf_destroy(gpu, &buf);
 
     printf("test buffer-buffer copy and readback\n");
     memset(test_dst, 0, buf_size);
-    buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+    buf = pl_buf_create(gpu, pl_buf_params(
         .size = buf_size,
         .initial_data = test_src,
-    });
+    ));
 
-    tbuf = pl_buf_create(gpu, &(struct pl_buf_params) {
+    tbuf = pl_buf_create(gpu, pl_buf_params(
         .size = buf_size,
         .host_readable = true,
-    });
+    ));
 
     REQUIRE(buf && tbuf);
     pl_buf_copy(gpu, tbuf, 0, buf, 0, buf_size);
     REQUIRE(pl_buf_read(gpu, tbuf, 0, test_dst, buf_size));
-    REQUIRE(memcmp(test_src, test_dst, buf_size) == 0);
+    REQUIRE_MEMEQ(test_src, test_dst, buf_size);
     pl_buf_destroy(gpu, &buf);
     pl_buf_destroy(gpu, &tbuf);
 
     if (buf_size <= gpu->limits.max_mapped_size) {
         printf("test host mapped buffer readback\n");
-        buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+        buf = pl_buf_create(gpu, pl_buf_params(
             .size = buf_size,
             .host_mapped = true,
             .initial_data = test_src,
-        });
+        ));
 
         REQUIRE(buf);
         REQUIRE(!pl_buf_poll(gpu, buf, 0));
-        REQUIRE(memcmp(test_src, buf->data, buf_size) == 0);
+        REQUIRE_MEMEQ(test_src, buf->data, buf_size);
         pl_buf_destroy(gpu, &buf);
     }
 
+    // `compute_queues` check is to exclude dummy GPUs here
+    if (buf_size <= gpu->limits.max_ssbo_size && gpu->limits.compute_queues)
+    {
+        printf("test endian swapping\n");
+        buf = pl_buf_create(gpu, pl_buf_params(
+            .size = buf_size,
+            .storable = true,
+            .initial_data = test_src,
+        ));
+
+        tbuf = pl_buf_create(gpu, pl_buf_params(
+            .size = buf_size,
+            .storable = true,
+            .host_readable = true,
+        ));
+
+        REQUIRE(buf && tbuf);
+        REQUIRE(pl_buf_copy_swap(gpu, &(struct pl_buf_copy_swap_params) {
+            .src = buf,
+            .dst = tbuf,
+            .size = buf_size,
+            .wordsize = 2,
+        }));
+        REQUIRE(pl_buf_read(gpu, tbuf, 0, test_dst, buf_size));
+        for (int i = 0; i < buf_size / 2; i++) {
+            REQUIRE_CMP(test_src[2 * i + 0], ==, test_dst[2 * i + 1], PRIu8);
+            REQUIRE_CMP(test_src[2 * i + 1], ==, test_dst[2 * i + 0], PRIu8);
+        }
+        // test endian swap in-place
+        REQUIRE(pl_buf_copy_swap(gpu, &(struct pl_buf_copy_swap_params) {
+            .src = tbuf,
+            .dst = tbuf,
+            .size = buf_size,
+            .wordsize = 4,
+        }));
+        REQUIRE(pl_buf_read(gpu, tbuf, 0, test_dst, buf_size));
+        for (int i = 0; i < buf_size / 4; i++) {
+            REQUIRE_CMP(test_src[4 * i + 0], ==, test_dst[4 * i + 2], PRIu8);
+            REQUIRE_CMP(test_src[4 * i + 1], ==, test_dst[4 * i + 3], PRIu8);
+            REQUIRE_CMP(test_src[4 * i + 2], ==, test_dst[4 * i + 0], PRIu8);
+            REQUIRE_CMP(test_src[4 * i + 3], ==, test_dst[4 * i + 1], PRIu8);
+        }
+        pl_buf_destroy(gpu, &buf);
+        pl_buf_destroy(gpu, &tbuf);
+    }
+
     free(test_src);
 }
 
@@ -102,7 +154,7 @@ static void pl_test_roundtrip(pl_gpu gpu, pl_tex tex[2],
     memset(dst, 0, bytes);
 
     for (size_t i = 0; i < bytes; i++)
-        src[i] = (RANDOM * 256);
+        src[i] = RANDOM_U8;
 
     pl_timer ul, dl;
     ul = pl_timer_create(gpu);
@@ -146,7 +198,7 @@ static void pl_test_roundtrip(pl_gpu gpu, pl_tex tex[2],
         // emulated 16/32 bit upload paths, figure out a better way to
         // generate data and verify the roundtrip!
     } else {
-        REQUIRE(memcmp(src, dst, bytes) == 0);
+        REQUIRE_MEMEQ(src, dst, bytes);
     }
 
     // Report timer results
@@ -223,6 +275,43 @@ static void pl_texture_tests(pl_gpu gpu)
     free(test_src);
 }
 
+static void pl_planar_tests(pl_gpu gpu)
+{
+    pl_fmt fmt = pl_find_named_fmt(gpu, "g8_b8_r8_420");
+    if (!fmt)
+        return;
+    REQUIRE_CMP(fmt->num_planes, ==, 3, "d");
+
+    const int width = 64, height = 32;
+    pl_tex tex = pl_tex_create(gpu, pl_tex_params(
+        .w              = width,
+        .h              = height,
+        .format         = fmt,
+        .blit_dst       = true,
+        .host_readable  = true,
+    ));
+    if (!tex)
+        return;
+    for (int i = 0; i < fmt->num_planes; i++)
+        REQUIRE(tex->planes[i]);
+
+    pl_tex plane = tex->planes[1];
+    uint8_t data[(width * height) >> 2];
+    REQUIRE_CMP(plane->params.w * plane->params.h, ==, PL_ARRAY_SIZE(data), "d");
+
+    pl_tex_clear(gpu, plane, (float[]){ (float) 0x80 / 0xFF, 0.0, 0.0, 1.0 });
+    REQUIRE(pl_tex_download(gpu, pl_tex_transfer_params(
+        .tex = plane,
+        .ptr = data,
+    )));
+
+    uint8_t ref[PL_ARRAY_SIZE(data)];
+    memset(ref, 0x80, sizeof(ref));
+    REQUIRE_MEMEQ(data, ref, PL_ARRAY_SIZE(data));
+
+    pl_tex_destroy(gpu, &tex);
+}
+
 static void pl_shader_tests(pl_gpu gpu)
 {
     if (gpu->glsl.version < 410)
@@ -339,10 +428,10 @@ static void pl_shader_tests(pl_gpu gpu)
         for (int y = 0; y < FBO_H; y++) {                                   \
             for (int x = 0; x < FBO_W; x++) {                               \
                 float *color = &data[(y * FBO_W + x) * 4];                  \
-                REQUIRE(feq(color[0], (x + 0.5) / FBO_W, eps));             \
-                REQUIRE(feq(color[1], (y + 0.5) / FBO_H, eps));             \
-                REQUIRE(feq(color[2], 0.0, eps));                           \
-                REQUIRE(feq(color[3], 1.0, eps));                           \
+                REQUIRE_FEQ(color[0], (x + 0.5) / FBO_W, eps);              \
+                REQUIRE_FEQ(color[1], (y + 0.5) / FBO_H, eps);              \
+                REQUIRE_FEQ(color[2], 0.0, eps);                            \
+                REQUIRE_FEQ(color[3], 1.0, eps);                            \
             }                                                               \
         }                                                                   \
     } while (0)
@@ -428,7 +517,7 @@ static void pl_shader_tests(pl_gpu gpu)
 
     if (fbo->params.storable) {
         // Test 1x1 blit, to make sure the scaling code runs
-        REQUIRE(pl_tex_blit_compute(gpu, dp, &(struct pl_tex_blit_params) {
+        REQUIRE(pl_tex_blit_compute(gpu, &(struct pl_tex_blit_params) {
             .src = src,
             .dst = fbo,
             .src_rc = {0, 0, 0, 1, 1, 1},
@@ -437,7 +526,7 @@ static void pl_shader_tests(pl_gpu gpu)
         }));
 
         // Test non-resizing blit, which uses the efficient imageLoad path
-        REQUIRE(pl_tex_blit_compute(gpu, dp, &(struct pl_tex_blit_params) {
+        REQUIRE(pl_tex_blit_compute(gpu, &(struct pl_tex_blit_params) {
             .src = src,
             .dst = fbo,
             .src_rc = {0, 0, 0, FBO_W, FBO_H, 1},
@@ -450,10 +539,14 @@ static void pl_shader_tests(pl_gpu gpu)
 
     // Test encoding/decoding of all gamma functions, color spaces, etc.
     for (enum pl_color_transfer trc = 0; trc < PL_COLOR_TRC_COUNT; trc++) {
+        struct pl_color_space test_csp = {
+            .transfer = trc,
+            .hdr.min_luma = PL_COLOR_HDR_BLACK,
+        };
         sh = pl_dispatch_begin(dp);
         pl_shader_sample_nearest(sh, pl_sample_src( .tex = src ));
-        pl_shader_delinearize(sh, pl_color_space( .transfer = trc ));
-        pl_shader_linearize(sh, pl_color_space( .transfer = trc ));
+        pl_shader_delinearize(sh, &test_csp);
+        pl_shader_linearize(sh, &test_csp);
         REQUIRE(pl_dispatch_finish(dp, pl_dispatch_params(
             .shader = &sh,
             .target = fbo,
@@ -478,6 +571,7 @@ static void pl_shader_tests(pl_gpu gpu)
         float epsilon;
         switch (sys) {
         case PL_COLOR_SYSTEM_BT_2020_C:
+        case PL_COLOR_SYSTEM_XYZ:
             epsilon = 1e-5;
             break;
 
@@ -499,10 +593,10 @@ static void pl_shader_tests(pl_gpu gpu)
         if (i == 5) {
             printf("Recreating pl_dispatch to test the caching\n");
             size_t size = pl_dispatch_save(dp, NULL);
-            REQUIRE(size > 0);
+            REQUIRE(size);
             uint8_t *cache = malloc(size);
             REQUIRE(cache);
-            REQUIRE(pl_dispatch_save(dp, cache) == size);
+            REQUIRE_CMP(pl_dispatch_save(dp, cache), ==, size, "zu");
 
             pl_dispatch_destroy(&dp);
             dp = pl_dispatch_create(gpu->log, gpu);
@@ -513,9 +607,9 @@ static void pl_shader_tests(pl_gpu gpu)
             // this on MSAN because it doesn't like it when we read from
             // program cache data generated by the non-instrumented GPU driver
             uint64_t hash = pl_str_hash((pl_str) { cache, size });
-            REQUIRE(pl_dispatch_save(dp, NULL) == size);
-            REQUIRE(pl_dispatch_save(dp, cache) == size);
-            REQUIRE(pl_str_hash((pl_str) { cache, size }) == hash);
+            REQUIRE_CMP(pl_dispatch_save(dp, NULL), ==, size, "zu");
+            REQUIRE_CMP(pl_dispatch_save(dp, cache), ==, size, "zu");
+            REQUIRE_CMP(pl_str_hash((pl_str) { cache, size }), ==, hash, PRIu64);
 #endif
             free(cache);
         }
@@ -525,8 +619,8 @@ static void pl_shader_tests(pl_gpu gpu)
         // For testing, force the use of CS if possible
         if (gpu->glsl.compute) {
             sh->type = SH_COMPUTE;
-            sh->res.compute_group_size[0] = 8;
-            sh->res.compute_group_size[1] = 8;
+            sh->group_size[0] = 8;
+            sh->group_size[1] = 8;
         }
 
         pl_shader_deband(sh, pl_sample_src( .tex = src ), pl_deband_params(
@@ -557,24 +651,25 @@ static void pl_shader_tests(pl_gpu gpu)
 
         float peak, avg;
         REQUIRE(pl_get_detected_peak(peak_state, &peak, &avg));
-        printf("detected peak: %f, average: %f\n", peak, avg);
 
         float real_peak = 0, real_avg = 0;
         for (int y = 0; y < FBO_H; y++) {
             for (int x = 0; x < FBO_W; x++) {
                 float *color = &data[(y * FBO_W + x) * 4];
-                float smax = powf(PL_MAX(color[0], color[1]), 2.2);
-                smax = (1 - 1e-3f) * smax + 1e-3f;
-                float slog = logf(PL_MAX(smax, 0.001));
-                real_peak = PL_MAX(smax, real_peak);
-                real_avg += slog;
+                float luma = 0.212639f * powf(color[0], 2.2f) +
+                             0.715169f * powf(color[1], 2.2f) +
+                             0.072192f * powf(color[2], 2.2f);
+                luma = pl_hdr_rescale(PL_HDR_NORM, PL_HDR_PQ, luma);
+                real_peak = PL_MAX(real_peak, luma);
+                real_avg += luma;
             }
         }
+        real_avg = real_avg / (FBO_W * FBO_H);
 
-        real_avg = expf(real_avg / (FBO_W * FBO_H));
-        printf("real peak: %f, real average: %f\n", real_peak, real_avg);
-        REQUIRE(feq(peak, real_peak, 1e-4));
-        REQUIRE(feq(avg, real_avg, 1e-3));
+        real_avg  = pl_hdr_rescale(PL_HDR_PQ, PL_HDR_NORM, real_avg);
+        real_peak = pl_hdr_rescale(PL_HDR_PQ, PL_HDR_NORM, real_peak);
+        REQUIRE_FEQ(peak, real_peak, 1e-3);
+        REQUIRE_FEQ(avg, real_avg, 1e-2);
     }
 
     pl_dispatch_abort(dp, &sh);
@@ -651,20 +746,18 @@ static void pl_shader_tests(pl_gpu gpu)
     }));
 
     // Test dolbyvision
-    if (gpu->glsl.version >= 130) {
-        struct pl_color_repr repr = {
-            .sys = PL_COLOR_SYSTEM_DOLBYVISION,
-            .dovi = &dovi_meta,
-        };
+    struct pl_color_repr repr = {
+        .sys = PL_COLOR_SYSTEM_DOLBYVISION,
+        .dovi = &dovi_meta,
+    };
 
-        sh = pl_dispatch_begin(dp);
-        pl_shader_sample_direct(sh, pl_sample_src( .tex = src ));
-        pl_shader_decode_color(sh, &repr, NULL);
-        REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
-            .shader = &sh,
-            .target = fbo,
-        }));
-    }
+    sh = pl_dispatch_begin(dp);
+    pl_shader_sample_direct(sh, pl_sample_src( .tex = src ));
+    pl_shader_decode_color(sh, &repr, NULL);
+    REQUIRE(pl_dispatch_finish(dp, &(struct pl_dispatch_params) {
+        .shader = &sh,
+        .target = fbo,
+    }));
 
     // Test deinterlacing
     sh = pl_dispatch_begin(dp);
@@ -769,6 +862,7 @@ static void pl_scaler_tests(pl_gpu gpu)
             .ptr            = fbo_data,
         }));
 
+#ifdef PRINT_OUTPUT
         int max = 255;
         printf("P2\n%d %d\n%d\n", fbo->params.w, fbo->params.h, max);
         for (int y = 0; y < fbo->params.h; y++) {
@@ -778,6 +872,7 @@ static void pl_scaler_tests(pl_gpu gpu)
             }
             printf("\n");
         }
+#endif
     }
 
 error:
@@ -864,8 +959,28 @@ static const char *user_shader_tests[] = {
     "//!BUFFER buf_storage                                                  \n"
     "//!VAR vec2 bat                                                        \n"
     "//!VAR int big[32];                                                    \n"
-    "//!STORAGE                                                             \n"
+    "//!STORAGE                                                             \n",
 
+    // Test custom parameters
+    "//!PARAM test                                                          \n"
+    "//!DESC test parameter                                                 \n"
+    "//!TYPE DYNAMIC float                                                  \n"
+    "//!MINIMUM 0.0                                                         \n"
+    "//!MAXIMUM 100.0                                                       \n"
+    "1.0                                                                    \n"
+    "                                                                       \n"
+    "//!PARAM testconst                                                     \n"
+    "//!TYPE CONSTANT uint                                                  \n"
+    "//!MAXIMUM 16                                                          \n"
+    "3                                                                      \n"
+    "                                                                       \n"
+    "//!PARAM testdefine                                                    \n"
+    "//!TYPE DEFINE                                                         \n"
+    "100                                                                    \n"
+    "                                                                       \n"
+    "//!HOOK MAIN                                                           \n"
+    "//!WHEN testconst 30 >                                                 \n"
+    "#error should not be run                                               \n",
 };
 
 static const char *test_luts[] = {
@@ -915,33 +1030,31 @@ static void render_info_cb(void *priv, const struct pl_render_info *info)
 
 static void pl_render_tests(pl_gpu gpu)
 {
-    pl_tex img5x5_tex = NULL, fbo = NULL;
+    pl_tex img_tex = NULL, fbo = NULL;
     pl_renderer rr = NULL;
 
-    static float data_5x5[5][5] = {
-        { 0.0, 0.0, 0.0, 0.0, 0.0 },
-        { 0.0, 0.0, 0.0, 0.0, 0.0 },
-        { 1.0, 0.0, 0.5, 0.0, 0.0 },
-        { 0.0, 0.0, 0.0, 1.0, 0.0 },
-        { 0.0, 0.3, 0.0, 0.0, 0.0 },
-    };
+    enum { width = 50, height = 50 };
+    static float data[width][height];
+    for (int y = 0; y < height; y++) {
+        for (int x = 0; x < width; x++)
+            data[y][x] = RANDOM;
+    }
 
-    const int width = 5, height = 5;
-    struct pl_plane img5x5 = {0};
-    struct pl_plane_data img5x5_data = {
+    struct pl_plane img_plane = {0};
+    struct pl_plane_data plane_data = {
         .type = PL_FMT_FLOAT,
         .width = width,
         .height = height,
         .component_size = { 8 * sizeof(float) },
         .component_map  = { 0 },
         .pixel_stride = sizeof(float),
-        .pixels = &data_5x5,
+        .pixels = data,
     };
 
-    if (!pl_recreate_plane(gpu, NULL, &fbo, &img5x5_data))
+    if (!pl_recreate_plane(gpu, NULL, &fbo, &plane_data))
         return;
 
-    if (!pl_upload_plane(gpu, &img5x5, &img5x5_tex, &img5x5_data))
+    if (!pl_upload_plane(gpu, &img_plane, &img_tex, &plane_data))
         goto error;
 
     rr = pl_renderer_create(gpu->log, gpu);
@@ -949,13 +1062,12 @@ static void pl_render_tests(pl_gpu gpu)
 
     struct pl_frame image = {
         .num_planes     = 1,
-        .planes         = { img5x5 },
+        .planes         = { img_plane },
         .repr = {
             .sys        = PL_COLOR_SYSTEM_BT_709,
             .levels     = PL_COLOR_LEVELS_FULL,
         },
         .color          = pl_color_space_srgb,
-        .crop           = {-1.0, 0.0, width - 1.0, height},
     };
 
     struct pl_frame target = {
@@ -965,7 +1077,6 @@ static void pl_render_tests(pl_gpu gpu)
             .components         = 3,
             .component_mapping  = {0, 1, 2},
         }},
-        .crop           = {2, 2, fbo->params.w - 2, fbo->params.h - 2},
         .repr = {
             .sys        = PL_COLOR_SYSTEM_RGB,
             .levels     = PL_COLOR_LEVELS_FULL,
@@ -975,6 +1086,7 @@ static void pl_render_tests(pl_gpu gpu)
     };
 
     REQUIRE(pl_render_image(rr, &image, &target, NULL));
+    REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
 
     // TODO: embed a reference texture and ensure it matches
 
@@ -990,25 +1102,45 @@ static void pl_render_tests(pl_gpu gpu)
             params.SNAME = &tmp;                                        \
             REQUIRE(pl_render_image(rr, &image, &target, &params));     \
             pl_gpu_flush(gpu);                                          \
+            REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE); \
         }                                                               \
     } while (0)
 
 #define TEST_PARAMS(NAME, FIELD, LIMIT) \
     TEST(NAME##_params, pl_##NAME##_params, pl_##NAME##_default_params, FIELD, LIMIT)
 
+    image.crop.x1 = width / 2.0;
+    image.crop.y1 = height / 2.0;
     for (int i = 0; i < pl_num_scale_filters; i++) {
         struct pl_render_params params = pl_render_default_params;
         params.upscaler = pl_scale_filters[i].filter;
         printf("testing `params.upscaler = /* %s */`\n", pl_scale_filters[i].name);
         REQUIRE(pl_render_image(rr, &image, &target, &params));
         pl_gpu_flush(gpu);
+        REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
     }
+    image.crop.x1 = image.crop.y1 = 0;
+
+    target.crop.x1 = width / 2.0;
+    target.crop.y1 = height / 2.0;
+    for (int i = 0; i < pl_num_scale_filters; i++) {
+        struct pl_render_params params = pl_render_default_params;
+        params.downscaler = pl_scale_filters[i].filter;
+        printf("testing `params.downscaler = /* %s */`\n", pl_scale_filters[i].name);
+        REQUIRE(pl_render_image(rr, &image, &target, &params));
+        pl_gpu_flush(gpu);
+        REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
+    }
+    target.crop.x1 = target.crop.y1 = 0;
 
     TEST_PARAMS(deband, iterations, 3);
     TEST_PARAMS(sigmoid, center, 1);
     TEST_PARAMS(color_map, intent, PL_INTENT_ABSOLUTE_COLORIMETRIC);
     TEST_PARAMS(dither, method, PL_DITHER_WHITE_NOISE);
     TEST_PARAMS(dither, temporal, true);
+    TEST_PARAMS(distort, alpha_mode, PL_ALPHA_INDEPENDENT);
+    TEST_PARAMS(distort, constrain, true);
+    TEST_PARAMS(distort, bicubic, true);
     TEST(cone_params, pl_cone_params, pl_vision_deuteranomaly, strength, 0);
 
     // Test gamma-correct dithering
@@ -1018,9 +1150,18 @@ static void pl_render_tests(pl_gpu gpu)
 
     // Test HDR tone mapping
     image.color = pl_color_space_hdr10;
-    TEST_PARAMS(color_map, tone_mapping_mode, PL_TONE_MAP_MODE_COUNT - 1);
-    TEST_PARAMS(color_map, gamut_mode, PL_GAMUT_MODE_COUNT - 1);
+    TEST_PARAMS(color_map, visualize_lut, true);
+    if (gpu->limits.max_ssbo_size)
+        TEST_PARAMS(peak_detect, allow_delayed, true);
+
+    // Test inverse tone-mapping and pure BPC
+    image.color.hdr.max_luma = 1000;
+    target.color.hdr.max_luma = 4000;
+    target.color.hdr.min_luma = 0.02;
+    TEST_PARAMS(color_map, inverse_tone_mapping, true);
+
     image.color = pl_color_space_srgb;
+    target.color = pl_color_space_srgb;
 
     // Test some misc stuff
     struct pl_render_params params = pl_render_default_params;
@@ -1032,19 +1173,33 @@ static void pl_render_tests(pl_gpu gpu)
         .temperature = 0.3,
     };
     REQUIRE(pl_render_image(rr, &image, &target, &params));
+    REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
     params = pl_render_default_params;
 
-    params.force_icc_lut = true;
+    // Test background blending and alpha transparency
+    params.blend_against_tiles = true;
+    params.corner_rounding = 0.25f;
     REQUIRE(pl_render_image(rr, &image, &target, &params));
+    REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
     params = pl_render_default_params;
 
     // Test film grain synthesis
     image.film_grain.type = PL_FILM_GRAIN_AV1;
-    image.film_grain.params.av1 = av1_grain_data,
+    image.film_grain.params.av1 = av1_grain_data;
     REQUIRE(pl_render_image(rr, &image, &target, &params));
+    REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
+
     image.film_grain.type = PL_FILM_GRAIN_H274;
-    image.film_grain.params.h274 = h274_grain_data,
+    image.film_grain.params.h274 = h274_grain_data;
     REQUIRE(pl_render_image(rr, &image, &target, &params));
+    // H.274 film grain synthesis requires compute shaders
+    if (gpu->glsl.compute) {
+        REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
+    } else {
+        const struct pl_render_errors rr_err = pl_renderer_get_errors(rr);
+        REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_FILM_GRAIN);
+        pl_renderer_reset_errors(rr, &rr_err);
+    }
     image.film_grain = (struct pl_film_grain_data) {0};
 
     // Test mpv-style custom shaders
@@ -1054,7 +1209,7 @@ static void pl_render_tests(pl_gpu gpu)
         hook = pl_mpv_user_shader_parse(gpu, user_shader_tests[i],
                                         strlen(user_shader_tests[i]));
 
-        if (gpu->glsl.compute) {
+        if (gpu->glsl.compute && gpu->limits.max_ssbo_size) {
             REQUIRE(hook);
         } else {
             // Not all shaders compile without compute shader support
@@ -1065,6 +1220,7 @@ static void pl_render_tests(pl_gpu gpu)
         params.hooks = &hook;
         params.num_hooks = 1;
         REQUIRE(pl_render_image(rr, &image, &target, &params));
+        REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
 
         pl_mpv_user_shader_destroy(&hook);
     }
@@ -1090,6 +1246,7 @@ static void pl_render_tests(pl_gpu gpu)
             printf("testing LUT method %d\n", t);
             image.lut_type = target.lut_type = params.lut_type = t;
             REQUIRE(pl_render_image(rr, &image, &target, &params));
+            REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
         }
 
         image.lut = target.lut = params.lut = NULL;
@@ -1097,27 +1254,34 @@ static void pl_render_tests(pl_gpu gpu)
     }
 
 #ifdef PL_HAVE_LCMS
-    // Test ICC profiles
-    image.profile = TEST_PROFILE(sRGB_v2_nano_icc);
-    REQUIRE(pl_render_image(rr, &image, &target, &params));
-    image.profile = (struct pl_icc_profile) {0};
 
-    target.profile = TEST_PROFILE(sRGB_v2_nano_icc);
-    REQUIRE(pl_render_image(rr, &image, &target, &params));
-    target.profile = (struct pl_icc_profile) {0};
+    // It doesn't fit without use of 3D textures on GLES2
+    if (gpu->glsl.version > 100) {
+        // Test ICC profiles
+        image.profile = TEST_PROFILE(sRGB_v2_nano_icc);
+        REQUIRE(pl_render_image(rr, &image, &target, &params));
+        REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
+        image.profile = (struct pl_icc_profile) {0};
 
-    image.profile = TEST_PROFILE(sRGB_v2_nano_icc);
-    target.profile = image.profile;
-    REQUIRE(pl_render_image(rr, &image, &target, &params));
-    image.profile = (struct pl_icc_profile) {0};
-    target.profile = (struct pl_icc_profile) {0};
+        target.profile = TEST_PROFILE(sRGB_v2_nano_icc);
+        REQUIRE(pl_render_image(rr, &image, &target, &params));
+        REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
+        target.profile = (struct pl_icc_profile) {0};
+
+        image.profile = TEST_PROFILE(sRGB_v2_nano_icc);
+        target.profile = image.profile;
+        REQUIRE(pl_render_image(rr, &image, &target, &params));
+        REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
+        image.profile = (struct pl_icc_profile) {0};
+        target.profile = (struct pl_icc_profile) {0};
+    }
 
 #endif
 
     // Test overlays
     image.num_overlays = 1;
     image.overlays = &(struct pl_overlay) {
-        .tex = img5x5.texture,
+        .tex = img_plane.texture,
         .mode = PL_OVERLAY_NORMAL,
         .num_parts = 2,
         .parts = (struct pl_overlay_part[]) {{
@@ -1129,14 +1293,16 @@ static void pl_render_tests(pl_gpu gpu)
         }},
     };
     REQUIRE(pl_render_image(rr, &image, &target, &params));
+    REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
     params.disable_fbos = true;
     REQUIRE(pl_render_image(rr, &image, &target, &params));
+    REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
     image.num_overlays = 0;
     params = pl_render_default_params;
 
     target.num_overlays = 1;
     target.overlays = &(struct pl_overlay) {
-        .tex = img5x5.texture,
+        .tex = img_plane.texture,
         .mode = PL_OVERLAY_MONOCHROME,
         .num_parts = 1,
         .parts = &(struct pl_overlay_part) {
@@ -1146,13 +1312,16 @@ static void pl_render_tests(pl_gpu gpu)
         },
     };
     REQUIRE(pl_render_image(rr, &image, &target, &params));
+    REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
     REQUIRE(pl_render_image(rr, NULL, &target, &params));
+    REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
     target.num_overlays = 0;
 
     // Test rotation
     for (pl_rotation rot = 0; rot < PL_ROTATION_360; rot += PL_ROTATION_90) {
         image.rotation = rot;
         REQUIRE(pl_render_image(rr, &image, &target, &params));
+        REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
     }
 
     // Attempt frame mixing, using the mixer queue helper
@@ -1205,12 +1374,12 @@ static void pl_render_tests(pl_gpu gpu)
 
     while ((ret = pl_queue_update(queue, &mix, &qparams)) != PL_QUEUE_EOF) {
         if (ret == PL_QUEUE_MORE) {
-            REQUIRE(qparams.pts > 0.0);
+            REQUIRE_CMP(qparams.pts, >, 0.0f, "f");
             pl_queue_push(queue, NULL); // push delayed EOF
             continue;
         }
 
-        REQUIRE(ret == PL_QUEUE_OK);
+        REQUIRE_CMP(ret, ==, PL_QUEUE_OK, "u");
         REQUIRE(pl_render_image_mix(rr, &mix, &target, &mix_params));
 
         // Simulate advancing vsync
@@ -1230,8 +1399,8 @@ static void pl_render_tests(pl_gpu gpu)
 
     pl_queue_reset(queue);
     while ((ret = pl_queue_update(queue, &mix, &qparams)) != PL_QUEUE_EOF) {
-        REQUIRE(ret == PL_QUEUE_OK);
-        REQUIRE(mix.num_frames <= 2);
+        REQUIRE_CMP(ret, ==, PL_QUEUE_OK, "u");
+        REQUIRE_CMP(mix.num_frames, <=, 2, "d");
         REQUIRE(pl_render_image_mix(rr, &mix, &target, &mix_params));
         qparams.pts += qparams.vsync_duration;
     }
@@ -1240,33 +1409,31 @@ static void pl_render_tests(pl_gpu gpu)
     pl_queue_reset(queue);
     REQUIRE(pl_queue_update(queue, &mix, &qparams) == PL_QUEUE_EOF);
 
-    if (gpu->glsl.version >= 130) {
-        // Test deinterlacing
-        pl_queue_reset(queue);
-        printf("testing deinterlacing \n");
-        for (int i = 0; i < NUM_MIX_FRAMES; i++) {
-            struct pl_source_frame *src = &srcframes[i];
-            if (i > 10)
-                src = &srcframes[NUM_MIX_FRAMES + 10 - i];
-            src->first_field = PL_FIELD_EVEN;
-            pl_queue_push(queue, src);
-        }
-        pl_queue_push(queue, NULL);
+    // Test deinterlacing
+    pl_queue_reset(queue);
+    printf("testing deinterlacing \n");
+    for (int i = 0; i < NUM_MIX_FRAMES; i++) {
+        struct pl_source_frame *src = &srcframes[i];
+        if (i > 10)
+            src = &srcframes[NUM_MIX_FRAMES + 10 - i];
+        src->first_field = PL_FIELD_EVEN;
+        pl_queue_push(queue, src);
+    }
+    pl_queue_push(queue, NULL);
 
-        qparams.pts = 0;
-        qparams.get_frame = NULL;
-        while ((ret = pl_queue_update(queue, &mix, &qparams)) != PL_QUEUE_EOF) {
-            REQUIRE(ret == PL_QUEUE_OK);
-            REQUIRE(pl_render_image_mix(rr, &mix, &target, &mix_params));
-            qparams.pts += qparams.vsync_duration;
-        }
+    qparams.pts = 0;
+    qparams.get_frame = NULL;
+    while ((ret = pl_queue_update(queue, &mix, &qparams)) != PL_QUEUE_EOF) {
+        REQUIRE_CMP(ret, ==, PL_QUEUE_OK, "u");
+        REQUIRE(pl_render_image_mix(rr, &mix, &target, &mix_params));
+        qparams.pts += qparams.vsync_duration;
     }
 
     pl_queue_destroy(&queue);
 
 error:
     pl_renderer_destroy(&rr);
-    pl_tex_destroy(gpu, &img5x5_tex);
+    pl_tex_destroy(gpu, &img_tex);
     pl_tex_destroy(gpu, &fbo);
 }
 
@@ -1277,10 +1444,6 @@ static struct pl_hook_res noop_hook(void *priv, const struct pl_hook_params *par
 
 static void pl_ycbcr_tests(pl_gpu gpu)
 {
-    pl_renderer rr = pl_renderer_create(gpu->log, gpu);
-    if (!rr)
-        return;
-
     struct pl_plane_data data[3];
     for (int i = 0; i < 3; i++) {
         const int sub = i > 0 ? 1 : 0;
@@ -1304,6 +1467,10 @@ static void pl_ycbcr_tests(pl_gpu gpu)
     if (!fmt || (fmt->caps & caps) != caps)
         return;
 
+    pl_renderer rr = pl_renderer_create(gpu->log, gpu);
+    if (!rr)
+        return;
+
     pl_tex src_tex[3] = {0};
     pl_tex dst_tex[3] = {0};
     struct pl_frame img = {
@@ -1370,6 +1537,7 @@ static void pl_ycbcr_tests(pl_gpu gpu)
             .hook = noop_hook,
         }},
     }));
+    REQUIRE(pl_renderer_get_errors(rr).errors == PL_RENDER_ERR_NONE);
 
     size_t buf_size = data[0].height * data[0].row_stride;
     dst_buffer = malloc(buf_size);
@@ -1390,7 +1558,7 @@ static void pl_ycbcr_tests(pl_gpu gpu)
                 uint16_t *src_pixel = (uint16_t *) &src_buffer[i][off];
                 uint16_t *dst_pixel = (uint16_t *) &dst_buffer[off];
                 int diff = abs((int) *src_pixel - (int) *dst_pixel);
-                REQUIRE(diff <= 50); // a little under 0.1%
+                REQUIRE_CMP(diff, <=, 50, "d"); // a little under 0.1%
             }
         }
     }
@@ -1499,7 +1667,7 @@ static void pl_test_host_ptr(pl_gpu gpu)
     });
 
     REQUIRE(buf);
-    REQUIRE(memcmp(data + offset, buf->data, slice) == 0);
+    REQUIRE_MEMEQ(data + offset, buf->data, slice);
 
     pl_buf_destroy(gpu, &buf);
     free(data);
@@ -1511,6 +1679,7 @@ static void gpu_shader_tests(pl_gpu gpu)
 {
     pl_buffer_tests(gpu);
     pl_texture_tests(gpu);
+    pl_planar_tests(gpu);
     pl_shader_tests(gpu);
     pl_scaler_tests(gpu);
     pl_render_tests(gpu);
