@@ -1,6 +1,8 @@
 #include "gpu_tests.h"
 #include "opengl/utils.h"
 
+#include <libplacebo/opengl.h>
+
 static void opengl_interop_tests(pl_gpu gpu)
 {
     pl_fmt fmt = pl_find_fmt(gpu, PL_FMT_UNORM, 1, 0, 0,
@@ -31,7 +33,7 @@ static void opengl_interop_tests(pl_gpu gpu)
     pl_tex import = pl_opengl_wrap(gpu, &wrap);
     REQUIRE(import);
     REQUIRE(import->params.renderable);
-    REQUIRE(import->params.blit_dst == export->params.blit_dst);
+    REQUIRE_CMP(import->params.blit_dst, ==, export->params.blit_dst, "d");
 
     pl_tex_destroy(gpu, &import);
     pl_tex_destroy(gpu, &export);
@@ -159,17 +161,9 @@ int main()
         EGLenum profile;
     } egl_vers[] = {
         { EGL_OPENGL_API,       EGL_OPENGL_BIT,     4, 6, 460, EGL_CONTEXT_OPENGL_CORE_PROFILE_BIT },
-        { EGL_OPENGL_API,       EGL_OPENGL_BIT,     4, 5, 450, EGL_CONTEXT_OPENGL_CORE_PROFILE_BIT },
-        { EGL_OPENGL_API,       EGL_OPENGL_BIT,     4, 4, 440, EGL_CONTEXT_OPENGL_CORE_PROFILE_BIT },
-        { EGL_OPENGL_API,       EGL_OPENGL_BIT,     4, 0, 400, EGL_CONTEXT_OPENGL_CORE_PROFILE_BIT },
         { EGL_OPENGL_API,       EGL_OPENGL_BIT,     3, 3, 330, EGL_CONTEXT_OPENGL_CORE_PROFILE_BIT },
-        { EGL_OPENGL_API,       EGL_OPENGL_BIT,     3, 2, 150, EGL_CONTEXT_OPENGL_CORE_PROFILE_BIT },
-        { EGL_OPENGL_API,       EGL_OPENGL_BIT,     3, 1, 140, EGL_CONTEXT_OPENGL_COMPATIBILITY_PROFILE_BIT, },
         { EGL_OPENGL_API,       EGL_OPENGL_BIT,     3, 0, 130, EGL_CONTEXT_OPENGL_COMPATIBILITY_PROFILE_BIT, },
-        { EGL_OPENGL_API,       EGL_OPENGL_BIT,     2, 1, 120, EGL_CONTEXT_OPENGL_COMPATIBILITY_PROFILE_BIT, },
-        { EGL_OPENGL_API,       EGL_OPENGL_BIT,     2, 0, 110, EGL_CONTEXT_OPENGL_COMPATIBILITY_PROFILE_BIT, },
         { EGL_OPENGL_ES_API,    EGL_OPENGL_ES3_BIT, 3, 0, 300, },
-        { EGL_OPENGL_ES_API,    EGL_OPENGL_ES2_BIT, 2, 0, 100, },
     };
 
     struct pl_glsl_version last_glsl = {0};
@@ -256,6 +250,11 @@ int main()
             goto next;
         }
 
+#ifdef CI_MAXGL
+        if (last_glsl.version && last_glsl.gles == gpu->glsl.gles)
+            goto next;
+#endif
+
         last_glsl = gpu->glsl;
         last_limits = gpu->limits;
 
