@@ -22,6 +22,8 @@
 #include "log.h"
 #include "pl_thread.h"
 
+#include <libplacebo/utils/frame_queue.h>
+
 struct cache_entry {
     pl_tex tex[4];
 };
@@ -115,7 +117,11 @@ pl_queue pl_queue_create(pl_gpu gpu)
 
     pl_mutex_init(&p->lock_strong);
     pl_mutex_init(&p->lock_weak);
-    PL_CHECK_ERR(pl_cond_init(&p->wakeup));
+    int ret = pl_cond_init(&p->wakeup);
+    if (ret) {
+        PL_ERR(p, "Failed to init conditional variable: %d", ret);
+        return NULL;
+    }
     return p;
 }
 
@@ -233,7 +239,7 @@ void pl_queue_reset(pl_queue p)
 
 static inline float delta(float old, float new)
 {
-    return fabs((new - old) / PL_MIN(new, old));
+    return fabsf((new - old) / PL_MIN(new, old));
 }
 
 static inline void default_estimate(struct pool *pool, float val)
@@ -291,8 +297,8 @@ static void queue_push(pl_queue p, const struct pl_source_frame *src)
     if (p->queue.num) {
         float last_pts = p->queue.elem[p->queue.num - 1]->pts;
         float delta = src->pts - last_pts;
-        if (delta < 0.0) {
-            PL_DEBUG(p, "Backwards source PTS jump %f -> %f", last_pts, src->pts);
+        if (delta <= 0.0) {
+            PL_DEBUG(p, "Non monotonically increasing PTS %f -> %f", last_pts, src->pts);
         } else if (p->fps.estimate && delta > 10.0 * p->fps.estimate) {
             PL_DEBUG(p, "Discontinuous source PTS jump %f -> %f", last_pts, src->pts);
         } else {
@@ -325,6 +331,7 @@ static void queue_push(pl_queue p, const struct pl_source_frame *src)
                 struct entry *prev = i > 0 ? p->queue.elem[i - 1] : NULL;
                 struct entry *next = i < p->queue.num ? p->queue.elem[i] : NULL;
                 struct entry *entry2 = pl_zalloc_ptr(NULL, entry2);
+                pl_rc_init(&entry2->rc);
                 if (next) {
                     entry2->pts = (entry->pts + next->pts) / 2;
                 } else if (src->duration) {
@@ -616,7 +623,9 @@ done:
 
         // Last frame is held for an extra `p->fps.estimate` duration,
         // afterwards this function just returns EOF.
-        if (p->queue.elem[0]->pts + p->fps.estimate < pts) {
+        if (pts < p->queue.elem[0]->pts + p->fps.estimate) {
+            ret = PL_QUEUE_OK;
+        } else {
             entry_cull(p, p->queue.elem[0], true);
             p->queue.num = 0;
             return PL_QUEUE_EOF;
@@ -633,6 +642,11 @@ static inline enum pl_queue_status point(pl_queue p, struct pl_frame_mix *mix,
     // Find closest frame (nearest neighbour semantics)
     pl_assert(p->queue.num);
     struct entry *entry = p->queue.elem[0];
+    if (entry->pts > params->pts) { // first frame not visible yet
+        *mix = (struct pl_frame_mix) {0};
+        return PL_QUEUE_OK;
+    }
+
     double best = fabs(entry->pts - params->pts);
     for (int i = 1; i < p->queue.num; i++) {
         double dist = fabs(p->queue.elem[i]->pts - params->pts);
@@ -815,7 +829,13 @@ static enum pl_queue_status interpolate(pl_queue p, struct pl_frame_mix *mix,
         case PL_QUEUE_ERR:
             return ret;
         case PL_QUEUE_MORE:
+            goto done;
         case PL_QUEUE_EOF:
+            // Don't forward EOF until we've held the last frame for the
+            // desired ZOH hold duration
+            if (params->pts >= p->queue.elem[p->queue.num - 1]->pts + p->fps.estimate)
+                return ret;
+            ret = PL_QUEUE_OK;
             goto done;
         case PL_QUEUE_OK:
             continue;
@@ -823,9 +843,14 @@ static enum pl_queue_status interpolate(pl_queue p, struct pl_frame_mix *mix,
     }
 
     if (!entry_complete(p->queue.elem[p->queue.num - 1])) {
-        ret = get_frame(p, params);
-        if (ret == PL_QUEUE_ERR)
+        switch ((ret = get_frame(p, params))) {
+        case PL_QUEUE_MORE:
+        case PL_QUEUE_OK:
+            break;
+        case PL_QUEUE_ERR:
+        case PL_QUEUE_EOF:
             return ret;
+        }
     }
 
 done: ;
@@ -900,7 +925,6 @@ enum pl_queue_status pl_queue_update(pl_queue p, struct pl_frame_mix *out_mix,
 {
     pl_mutex_lock(&p->lock_strong);
     pl_mutex_lock(&p->lock_weak);
-    default_estimate(&p->fps, params->frame_duration);
     default_estimate(&p->vps, params->vsync_duration);
 
     float delta = params->pts - p->prev_pts;
@@ -948,9 +972,10 @@ enum pl_queue_status pl_queue_update(pl_queue p, struct pl_frame_mix *out_mix,
     // Ignore unrealistically high or low FPS, common near start of playback
     static const float max_vsync = 1.0 / MIN_FPS;
     static const float min_vsync = 1.0 / MAX_FPS;
+    bool estimation_ok = p->vps.estimate > min_vsync && p->vps.estimate < max_vsync;
     enum pl_queue_status ret;
 
-    if (p->vps.estimate > min_vsync && p->vps.estimate < max_vsync) {
+    if (estimation_ok || params->vsync_duration > 0) {
         // We know the vsync duration, so construct an interpolation mix
         ret = interpolate(p, out_mix, params);
     } else {
