@@ -22,7 +22,12 @@
 static VkResult vk_cmd_poll(struct vk_ctx *vk, struct vk_cmd *cmd,
                             uint64_t timeout)
 {
-    return vk->WaitForFences(vk->dev, 1, &cmd->fence, false, timeout);
+    return vk->WaitSemaphores(vk->dev, &(VkSemaphoreWaitInfo) {
+        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_WAIT_INFO,
+        .semaphoreCount = 1,
+        .pSemaphores = &cmd->sync.sem,
+        .pValues = &cmd->sync.value,
+    }, timeout);
 }
 
 static void flush_callbacks(struct vk_ctx *vk)
@@ -46,10 +51,7 @@ static void vk_cmd_reset(struct vk_ctx *vk, struct vk_cmd *cmd)
 
     cmd->callbacks.num = 0;
     cmd->deps.num = 0;
-    cmd->depstages.num = 0;
-    cmd->depvalues.num = 0;
     cmd->sigs.num = 0;
-    cmd->sigvalues.num = 0;
 }
 
 static void vk_cmd_destroy(struct vk_ctx *vk, struct vk_cmd *cmd)
@@ -59,7 +61,7 @@ static void vk_cmd_destroy(struct vk_ctx *vk, struct vk_cmd *cmd)
 
     vk_cmd_poll(vk, cmd, UINT64_MAX);
     vk_cmd_reset(vk, cmd);
-    vk->DestroyFence(vk->dev, cmd->fence, PL_VK_ALLOC);
+    vk->DestroySemaphore(vk->dev, cmd->sync.sem, PL_VK_ALLOC);
     vk->FreeCommandBuffers(vk->dev, cmd->pool->pool, 1, &cmd->buf);
 
     pl_free(cmd);
@@ -79,13 +81,19 @@ static struct vk_cmd *vk_cmd_create(struct vk_ctx *vk, struct vk_cmdpool *pool)
 
     VK(vk->AllocateCommandBuffers(vk->dev, &ainfo, &cmd->buf));
 
-    VkFenceCreateInfo finfo = {
-        .sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO,
-        .flags = VK_FENCE_CREATE_SIGNALED_BIT,
+    static const VkSemaphoreTypeCreateInfo stinfo = {
+        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO,
+        .semaphoreType  = VK_SEMAPHORE_TYPE_TIMELINE,
+        .initialValue   = 0,
     };
 
-    VK(vk->CreateFence(vk->dev, &finfo, PL_VK_ALLOC, &cmd->fence));
-    PL_VK_NAME(FENCE, cmd->fence, "cmd");
+    static const VkSemaphoreCreateInfo sinfo = {
+        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
+        .pNext = &stinfo,
+    };
+
+    VK(vk->CreateSemaphore(vk->dev, &sinfo, PL_VK_ALLOC, &cmd->sync.sem));
+    PL_VK_NAME(SEMAPHORE, cmd->sync.sem, "cmd");
 
     return cmd;
 
@@ -119,59 +127,40 @@ void vk_cmd_callback(struct vk_cmd *cmd, vk_cb callback,
     });
 }
 
-void vk_cmd_dep(struct vk_cmd *cmd, VkPipelineStageFlags stage, pl_vulkan_sem dep)
+void vk_cmd_dep(struct vk_cmd *cmd, VkPipelineStageFlags2 stage, pl_vulkan_sem dep)
 {
-    assert(cmd->deps.num == cmd->depstages.num);
-    assert(cmd->deps.num == cmd->depvalues.num);
-    PL_ARRAY_APPEND(cmd, cmd->deps, dep.sem);
-    PL_ARRAY_APPEND(cmd, cmd->depvalues, dep.value);
-    PL_ARRAY_APPEND(cmd, cmd->depstages, stage);
+    PL_ARRAY_APPEND(cmd, cmd->deps, (VkSemaphoreSubmitInfo) {
+        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO,
+        .semaphore  = dep.sem,
+        .value      = dep.value,
+        .stageMask  = stage,
+    });
 }
 
-void vk_cmd_sig(struct vk_cmd *cmd, pl_vulkan_sem sig)
+void vk_cmd_sig(struct vk_cmd *cmd, VkPipelineStageFlags2 stage, pl_vulkan_sem sig)
 {
-    assert(cmd->sigs.num == cmd->sigvalues.num);
-    PL_ARRAY_APPEND(cmd, cmd->sigs, sig.sem);
-    PL_ARRAY_APPEND(cmd, cmd->sigvalues, sig.value);
-}
-
-void vk_sem_uninit(struct vk_ctx *vk, struct vk_sem *sem)
-{
-    vk->DestroySemaphore(vk->dev, sem->semaphore, PL_VK_ALLOC);
-    *sem = (struct vk_sem) {0};
-}
-
-bool vk_sem_init(struct vk_ctx *vk, struct vk_sem *sem, pl_debug_tag debug_tag)
-{
-    *sem = (struct vk_sem) {
-        .write.stage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT,
-        .read.stage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT,
+    VkSemaphoreSubmitInfo sinfo = {
+        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO,
+        .semaphore  = sig.sem,
+        .value      = sig.value,
+        .stageMask  = stage,
     };
 
-    static const VkSemaphoreTypeCreateInfo stinfo = {
-        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO,
-        .semaphoreType  = VK_SEMAPHORE_TYPE_TIMELINE,
-        .initialValue   = 0,
-    };
+    // Try updating existing semaphore signal operations in-place
+    for (int i = 0; i < cmd->sigs.num; i++) {
+        if (cmd->sigs.elem[i].semaphore == sig.sem) {
+            pl_assert(sig.value > cmd->sigs.elem[i].value);
+            cmd->sigs.elem[i] = sinfo;
+            return;
+        }
+    }
 
-    static const VkSemaphoreCreateInfo sinfo = {
-        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
-        .pNext = &stinfo,
-    };
-
-    // We always create a semaphore, so we can perform host waits on it
-    VK(vk->CreateSemaphore(vk->dev, &sinfo, PL_VK_ALLOC, &sem->semaphore));
-    PL_VK_NAME(SEMAPHORE, sem->semaphore, PL_DEF(debug_tag, "vk_sem"));
-    return true;
-
-error:
-    vk->failed = true;
-    return false;
+    PL_ARRAY_APPEND(cmd, cmd->sigs, sinfo);
 }
 
 struct vk_sync_scope vk_sem_barrier(struct vk_ctx *vk, struct vk_cmd *cmd,
-                                    struct vk_sem *sem, VkPipelineStageFlags stage,
-                                    VkAccessFlags access, bool is_trans)
+                                    struct vk_sem *sem, VkPipelineStageFlags2 stage,
+                                    VkAccessFlags2 access, bool is_trans)
 {
     bool is_write = (access & vk_access_write) || is_trans;
 
@@ -187,13 +176,14 @@ struct vk_sync_scope vk_sem_barrier(struct vk_ctx *vk, struct vk_cmd *cmd,
             // No semaphore needed in this case because the implicit submission
             // order execution dependencies already transitively imply a wait
             // for the previous write
-        } else if (last.queue) {
-            vk_cmd_dep(cmd, stage, (pl_vulkan_sem) {
-                .sem = sem->semaphore,
-                .value = last.value,
-            });
+        } else if (last.sync.sem) {
+            // Image barrier still needs to depend on this stage for implicit
+            // ordering guarantees to apply properly
+            vk_cmd_dep(cmd, stage, last.sync);
+            last.stage = stage;
         }
-        last.stage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT;
+
+        // Last access is on different queue, so no pipeline barrier needed
         last.access = 0;
     }
 
@@ -203,71 +193,63 @@ struct vk_sync_scope vk_sem_barrier(struct vk_ctx *vk, struct vk_cmd *cmd,
     {
         // A past pipeline barrier already covers this access transitively, so
         // we don't need to emit another pipeline barrier at all
-        last.stage = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT;
         last.access = 0;
     }
 
-    pl_assert(sem->read.value >= sem->write.value);
-    uint64_t next_value = sem->read.value + 1;
-    vk_cmd_sig(cmd, (pl_vulkan_sem) {
-        .sem = sem->semaphore,
-        .value = next_value,
-    });
-
     if (is_write) {
         sem->write = (struct vk_sync_scope) {
-            .value = next_value,
+            .sync = cmd->sync,
             .queue = cmd->queue,
             .stage = stage,
             .access = access,
         };
 
         sem->read = (struct vk_sync_scope) {
-            .value = next_value,
+            .sync = cmd->sync,
             .queue = cmd->queue,
             // no stage or access scope, because no reads happened yet
         };
     } else if (sem->read.queue == cmd->queue) {
         // Coalesce multiple same-queue reads into a single access scope
-        sem->read.value = next_value;
+        sem->read.sync = cmd->sync;
         sem->read.stage |= stage;
         sem->read.access |= access;
     } else {
         sem->read = (struct vk_sync_scope) {
-            .value = next_value,
+            .sync = cmd->sync,
             .queue = cmd->queue,
             .stage = stage,
             .access = access,
         };
     }
 
+    // We never need to include pipeline barriers for reads, only writes
+    last.access &= vk_access_write;
     return last;
 }
 
-struct vk_cmdpool *vk_cmdpool_create(struct vk_ctx *vk,
-                                     VkDeviceQueueCreateInfo qinfo,
+struct vk_cmdpool *vk_cmdpool_create(struct vk_ctx *vk, int qf, int qnum,
                                      VkQueueFamilyProperties props)
 {
     struct vk_cmdpool *pool = pl_alloc_ptr(NULL, pool);
     *pool = (struct vk_cmdpool) {
-        .props = props,
-        .qf = qinfo.queueFamilyIndex,
-        .queues = pl_calloc(pool, qinfo.queueCount, sizeof(VkQueue)),
-        .num_queues = qinfo.queueCount,
+        .props      = props,
+        .qf         = qf,
+        .queues     = pl_calloc(pool, qnum, sizeof(VkQueue)),
+        .num_queues = qnum,
     };
 
-    for (int n = 0; n < pool->num_queues; n++)
-        vk->GetDeviceQueue(vk->dev, pool->qf, n, &pool->queues[n]);
+    for (int n = 0; n < qnum; n++)
+        vk->GetDeviceQueue(vk->dev, qf, n, &pool->queues[n]);
 
     VkCommandPoolCreateInfo cinfo = {
         .sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO,
         .flags = VK_COMMAND_POOL_CREATE_TRANSIENT_BIT |
                  VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT,
-        .queueFamilyIndex = pool->qf,
+        .queueFamilyIndex = qf,
     };
 
     VK(vk->CreateCommandPool(vk->dev, &cinfo, PL_VK_ALLOC, &pool->pool));
-
     return pool;
 
 error:
@@ -317,8 +299,11 @@ struct vk_cmd *vk_cmd_begin(struct vk_ctx *vk, struct vk_cmdpool *pool,
     VK(vk->BeginCommandBuffer(cmd->buf, &binfo));
 
     debug_tag = PL_DEF(debug_tag, "vk_cmd");
-    PL_VK_NAME(COMMAND_BUFFER, cmd->buf, debug_tag);
-    PL_VK_NAME(FENCE, cmd->fence, debug_tag);
+    PL_VK_NAME_HANDLE(COMMAND_BUFFER, cmd->buf, debug_tag);
+    PL_VK_NAME(SEMAPHORE, cmd->sync.sem, debug_tag);
+
+    cmd->sync.value++;
+    vk_cmd_sig(cmd, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, cmd->sync);
     return cmd;
 
 error:
@@ -338,48 +323,39 @@ bool vk_cmd_submit(struct vk_ctx *vk, struct vk_cmd **pcmd)
     struct vk_cmdpool *pool = cmd->pool;
 
     VK(vk->EndCommandBuffer(cmd->buf));
-    VK(vk->ResetFences(vk->dev, 1, &cmd->fence));
 
-    VkTimelineSemaphoreSubmitInfo tinfo = {
-        .sType = VK_STRUCTURE_TYPE_TIMELINE_SEMAPHORE_SUBMIT_INFO,
-        .waitSemaphoreValueCount = cmd->depvalues.num,
-        .pWaitSemaphoreValues = cmd->depvalues.elem,
-        .signalSemaphoreValueCount = cmd->sigvalues.num,
-        .pSignalSemaphoreValues = cmd->sigvalues.elem,
-    };
-
-    VkSubmitInfo sinfo = {
-        .sType = VK_STRUCTURE_TYPE_SUBMIT_INFO,
-        .pNext = &tinfo,
-        .commandBufferCount = 1,
-        .pCommandBuffers = &cmd->buf,
-        .waitSemaphoreCount = cmd->deps.num,
-        .pWaitSemaphores = cmd->deps.elem,
-        .pWaitDstStageMask = cmd->depstages.elem,
-        .signalSemaphoreCount = cmd->sigs.num,
-        .pSignalSemaphores = cmd->sigs.elem,
+    VkSubmitInfo2 sinfo = {
+        .sType = VK_STRUCTURE_TYPE_SUBMIT_INFO_2,
+        .waitSemaphoreInfoCount = cmd->deps.num,
+        .pWaitSemaphoreInfos = cmd->deps.elem,
+        .signalSemaphoreInfoCount = cmd->sigs.num,
+        .pSignalSemaphoreInfos = cmd->sigs.elem,
+        .commandBufferInfoCount = 1,
+        .pCommandBufferInfos = &(VkCommandBufferSubmitInfo) {
+            .sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_SUBMIT_INFO,
+            .commandBuffer = cmd->buf,
+        },
     };
 
     if (pl_msg_test(vk->log, PL_LOG_TRACE)) {
         PL_TRACE(vk, "Submitting command %p on queue %p (QF %d):",
                  (void *) cmd->buf, (void *) cmd->queue, pool->qf);
         for (int n = 0; n < cmd->deps.num; n++) {
-            PL_TRACE(vk, "    waits on semaphore %p = %"PRIu64,
-                     (void *) cmd->deps.elem[n], cmd->depvalues.elem[n]);
+            PL_TRACE(vk, "    waits on semaphore 0x%"PRIx64" = %"PRIu64,
+                     (uint64_t) cmd->deps.elem[n].semaphore, cmd->deps.elem[n].value);
         }
         for (int n = 0; n < cmd->sigs.num; n++) {
-            PL_TRACE(vk, "    signals semaphore %p = %"PRIu64,
-                    (void *) cmd->sigs.elem[n], cmd->sigvalues.elem[n]);
+            PL_TRACE(vk, "    signals semaphore 0x%"PRIx64" = %"PRIu64,
+                    (uint64_t) cmd->sigs.elem[n].semaphore, cmd->sigs.elem[n].value);
         }
-        PL_TRACE(vk, "    signals fence %p", (void *) cmd->fence);
         if (cmd->callbacks.num)
             PL_TRACE(vk, "    signals %d callbacks", cmd->callbacks.num);
     }
 
     vk->lock_queue(vk->queue_ctx, pool->qf, cmd->qindex);
-    VkResult res = vk->QueueSubmit(cmd->queue, 1, &sinfo, cmd->fence);
+    VkResult res = vk->QueueSubmit2KHR(cmd->queue, 1, &sinfo, VK_NULL_HANDLE);
     vk->unlock_queue(vk->queue_ctx, pool->qf, cmd->qindex);
-    PL_VK_ASSERT(res, "vkQueueSubmit");
+    PL_VK_ASSERT(res, "vkQueueSubmit2");
 
     pl_mutex_lock(&vk->lock);
     PL_ARRAY_APPEND(vk->alloc, vk->cmds_pending, cmd);
@@ -410,7 +386,8 @@ bool vk_poll_commands(struct vk_ctx *vk, uint64_t timeout)
         if (!vk->cmds_pending.num || vk->cmds_pending.elem[0] != cmd)
             continue; // another thread modified this state while blocking
 
-        PL_TRACE(vk, "VkFence signalled: %p", (void *) cmd->fence);
+        PL_TRACE(vk, "VkSemaphore signalled: 0x%"PRIx64" = %"PRIu64,
+                 (uint64_t) cmd->sync.sem, cmd->sync.value);
         PL_ARRAY_REMOVE_AT(vk->cmds_pending, 0); // remove before callbacks
         vk_cmd_reset(vk, cmd);
         PL_ARRAY_APPEND(pool, pool->cmds, cmd);
