@@ -19,11 +19,14 @@
 
 #define VK_NO_PROTOTYPES
 #define VK_ENABLE_BETA_EXTENSIONS // for VK_KHR_portability_subset
+#define VK_USE_PLATFORM_METAL_EXT
 
 #include "../common.h"
 #include "../log.h"
 #include "../pl_thread.h"
 
+#include <libplacebo/vulkan.h>
+
 #ifdef PL_HAVE_WIN32
 #include <windows.h>
 #include <vulkan/vulkan_win32.h>
@@ -41,12 +44,8 @@
 #define PL_VK_LOAD_FUN(inst, name, get_addr) \
     PL_VK_FUN(name) = (PFN_vk##name) get_addr(inst, "vk" #name);
 
-// Hard-coded limit on the number of pending commands, to avoid OOM loops
-#define PL_VK_MAX_PENDING_CMDS 1024
-
-// Shitty compatibility alias for very old vulkan.h versions
-#ifndef VK_API_VERSION_1_2
-#define VK_API_VERSION_1_2 VK_MAKE_VERSION(1, 2, 0)
+#ifndef VK_VENDOR_ID_NVIDIA
+#define VK_VENDOR_ID_NVIDIA 0x10DE
 #endif
 
 // Shared struct used to hold vulkan context information
@@ -59,7 +58,7 @@ struct vk_ctx {
     pl_log log;
     VkInstance inst;
     VkPhysicalDevice physd;
-    VkPhysicalDeviceLimits limits;
+    VkPhysicalDeviceProperties props;
     VkPhysicalDeviceFeatures2 features;
     uint32_t api_ver; // device API version
     VkDevice dev;
@@ -80,9 +79,9 @@ struct vk_ctx {
     struct vk_cmdpool *pool_transfer;
 
     // Queue locking functions
-    PL_ARRAY(pl_mutex *) queue_locks;
-    void (*lock_queue)(void *queue_ctx, int qf, int idx);
-    void (*unlock_queue)(void *queue_ctx, int qf, int idx);
+    PL_ARRAY(PL_ARRAY(pl_mutex)) queue_locks;
+    void (*lock_queue)(void *queue_ctx, uint32_t qf, uint32_t idx);
+    void (*unlock_queue)(void *queue_ctx, uint32_t qf, uint32_t idx);
     void *queue_ctx;
 
     // Pending commands. These are shared for the entire mpvk_ctx to ensure
@@ -124,7 +123,7 @@ struct vk_ctx {
     PL_VK_FUN(BindBufferMemory);
     PL_VK_FUN(BindImageMemory);
     PL_VK_FUN(CmdBeginDebugUtilsLabelEXT);
-    PL_VK_FUN(CmdBeginRenderPass);
+    PL_VK_FUN(CmdBeginRenderingKHR);
     PL_VK_FUN(CmdBindDescriptorSets);
     PL_VK_FUN(CmdBindIndexBuffer);
     PL_VK_FUN(CmdBindPipeline);
@@ -139,17 +138,15 @@ struct vk_ctx {
     PL_VK_FUN(CmdDraw);
     PL_VK_FUN(CmdDrawIndexed);
     PL_VK_FUN(CmdEndDebugUtilsLabelEXT);
-    PL_VK_FUN(CmdEndRenderPass);
-    PL_VK_FUN(CmdPipelineBarrier);
+    PL_VK_FUN(CmdEndRenderingKHR);
+    PL_VK_FUN(CmdPipelineBarrier2KHR);
     PL_VK_FUN(CmdPushConstants);
     PL_VK_FUN(CmdPushDescriptorSetKHR);
     PL_VK_FUN(CmdResetQueryPool);
-    PL_VK_FUN(CmdSetEvent);
     PL_VK_FUN(CmdSetScissor);
     PL_VK_FUN(CmdSetViewport);
     PL_VK_FUN(CmdUpdateBuffer);
-    PL_VK_FUN(CmdWaitEvents);
-    PL_VK_FUN(CmdWriteTimestamp);
+    PL_VK_FUN(CmdWriteTimestamp2KHR);
     PL_VK_FUN(CreateBuffer);
     PL_VK_FUN(CreateBufferView);
     PL_VK_FUN(CreateCommandPool);
@@ -157,16 +154,13 @@ struct vk_ctx {
     PL_VK_FUN(CreateDebugReportCallbackEXT);
     PL_VK_FUN(CreateDescriptorPool);
     PL_VK_FUN(CreateDescriptorSetLayout);
-    PL_VK_FUN(CreateEvent);
     PL_VK_FUN(CreateFence);
-    PL_VK_FUN(CreateFramebuffer);
     PL_VK_FUN(CreateGraphicsPipelines);
     PL_VK_FUN(CreateImage);
     PL_VK_FUN(CreateImageView);
     PL_VK_FUN(CreatePipelineCache);
     PL_VK_FUN(CreatePipelineLayout);
     PL_VK_FUN(CreateQueryPool);
-    PL_VK_FUN(CreateRenderPass);
     PL_VK_FUN(CreateSampler);
     PL_VK_FUN(CreateSemaphore);
     PL_VK_FUN(CreateShaderModule);
@@ -178,9 +172,7 @@ struct vk_ctx {
     PL_VK_FUN(DestroyDescriptorPool);
     PL_VK_FUN(DestroyDescriptorSetLayout);
     PL_VK_FUN(DestroyDevice);
-    PL_VK_FUN(DestroyEvent);
     PL_VK_FUN(DestroyFence);
-    PL_VK_FUN(DestroyFramebuffer);
     PL_VK_FUN(DestroyImage);
     PL_VK_FUN(DestroyImageView);
     PL_VK_FUN(DestroyInstance);
@@ -188,11 +180,11 @@ struct vk_ctx {
     PL_VK_FUN(DestroyPipelineCache);
     PL_VK_FUN(DestroyPipelineLayout);
     PL_VK_FUN(DestroyQueryPool);
-    PL_VK_FUN(DestroyRenderPass);
     PL_VK_FUN(DestroySampler);
     PL_VK_FUN(DestroySemaphore);
     PL_VK_FUN(DestroyShaderModule);
     PL_VK_FUN(DestroySwapchainKHR);
+    PL_VK_FUN(DeviceWaitIdle);
     PL_VK_FUN(EndCommandBuffer);
     PL_VK_FUN(FlushMappedMemoryRanges);
     PL_VK_FUN(FreeCommandBuffers);
@@ -212,18 +204,25 @@ struct vk_ctx {
     PL_VK_FUN(InvalidateMappedMemoryRanges);
     PL_VK_FUN(MapMemory);
     PL_VK_FUN(QueuePresentKHR);
-    PL_VK_FUN(QueueSubmit);
-    PL_VK_FUN(ResetEvent);
+    PL_VK_FUN(QueueSubmit2KHR);
+    PL_VK_FUN(QueueWaitIdle);
     PL_VK_FUN(ResetFences);
-    PL_VK_FUN(ResetQueryPoolEXT);
+    PL_VK_FUN(ResetQueryPool);
     PL_VK_FUN(SetDebugUtilsObjectNameEXT);
     PL_VK_FUN(SetHdrMetadataEXT);
     PL_VK_FUN(UpdateDescriptorSets);
     PL_VK_FUN(WaitForFences);
-    PL_VK_FUN(WaitSemaphoresKHR);
+    PL_VK_FUN(WaitSemaphores);
 
 #ifdef PL_HAVE_WIN32
     PL_VK_FUN(GetMemoryWin32HandleKHR);
     PL_VK_FUN(GetSemaphoreWin32HandleKHR);
 #endif
+
+#ifdef VK_EXT_metal_objects
+    PL_VK_FUN(ExportMetalObjectsEXT);
+#endif
+#ifdef VK_EXT_full_screen_exclusive
+    PL_VK_FUN(AcquireFullScreenExclusiveModeEXT);
+#endif
 };
