@@ -36,7 +36,6 @@ struct vk_fun {
 
 struct vk_ext {
     const char *name;
-    uint32_t core_ver;
     const struct vk_fun *funs;
 };
 
@@ -155,29 +154,12 @@ static const struct vk_ext vk_device_extensions[] = {
             PL_VK_DEV_FUN(SetHdrMetadataEXT),
             {0}
         },
-    }, {
-        .name = VK_EXT_HOST_QUERY_RESET_EXTENSION_NAME,
-        .core_ver = VK_API_VERSION_1_2,
-        .funs = (const struct vk_fun[]) {
-            PL_VK_DEV_FUN(ResetQueryPoolEXT),
-            {0}
-        },
-    }, {
-        .name = VK_KHR_IMAGE_FORMAT_LIST_EXTENSION_NAME,
-        .core_ver = VK_API_VERSION_1_2,
     }, {
         .name = VK_EXT_IMAGE_DRM_FORMAT_MODIFIER_EXTENSION_NAME,
         .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(GetImageDrmFormatModifierPropertiesEXT),
             {0}
         },
-    }, {
-        .name = VK_KHR_TIMELINE_SEMAPHORE_EXTENSION_NAME,
-        .core_ver = VK_API_VERSION_1_2,
-        .funs = (const struct vk_fun[]) {
-            PL_VK_DEV_FUN(WaitSemaphoresKHR),
-            {0}
-        },
 #ifdef VK_KHR_portability_subset
     }, {
         .name = VK_KHR_PORTABILITY_SUBSET_EXTENSION_NAME,
@@ -198,6 +180,20 @@ static const struct vk_ext vk_device_extensions[] = {
             {0}
         },
 #endif
+    }, {
+        .name = VK_KHR_SYNCHRONIZATION_2_EXTENSION_NAME,
+        .funs = (const struct vk_fun[]) {
+            PL_VK_DEV_FUN(CmdPipelineBarrier2KHR),
+            PL_VK_DEV_FUN(QueueSubmit2KHR),
+            {0}
+        },
+    }, {
+        .name = VK_KHR_DYNAMIC_RENDERING_EXTENSION_NAME,
+        .funs = (const struct vk_fun[]) {
+            PL_VK_DEV_FUN(CmdBeginRenderingKHR),
+            PL_VK_DEV_FUN(CmdEndRenderingKHR),
+            {0}
+        },
     },
 };
 
@@ -214,10 +210,7 @@ const char * const pl_vulkan_recommended_extensions[] = {
 #endif
     VK_EXT_PCI_BUS_INFO_EXTENSION_NAME,
     VK_EXT_HDR_METADATA_EXTENSION_NAME,
-    VK_EXT_HOST_QUERY_RESET_EXTENSION_NAME,
-    VK_KHR_IMAGE_FORMAT_LIST_EXTENSION_NAME,
     VK_EXT_IMAGE_DRM_FORMAT_MODIFIER_EXTENSION_NAME,
-    VK_KHR_TIMELINE_SEMAPHORE_EXTENSION_NAME,
 #ifdef VK_KHR_portability_subset
     VK_KHR_PORTABILITY_SUBSET_EXTENSION_NAME,
 #endif
@@ -227,6 +220,8 @@ const char * const pl_vulkan_recommended_extensions[] = {
 #ifdef VK_EXT_full_screen_exclusive
     VK_EXT_FULL_SCREEN_EXCLUSIVE_EXTENSION_NAME,
 #endif
+    VK_KHR_SYNCHRONIZATION_2_EXTENSION_NAME,
+    VK_KHR_DYNAMIC_RENDERING_EXTENSION_NAME,
 };
 
 const int pl_vulkan_num_recommended_extensions =
@@ -238,29 +233,38 @@ static_assert(PL_ARRAY_SIZE(pl_vulkan_recommended_extensions) + 1 ==
               "pl_vulkan_recommended_extensions out of sync with "
               "vk_device_extensions?");
 
-#ifdef VK_VERSION_1_3
-// Needed for LocalSizeId execution mode
-static const VkPhysicalDeviceMaintenance4Features device_maintenance4_features = {
-    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_FEATURES,
-    .maintenance4 = true
-};
-#endif
-
-// pNext chain of features we want enabled
-static const VkPhysicalDeviceTimelineSemaphoreFeatures timeline_semaphores = {
-    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES,
-    .timelineSemaphore = true,
+// Recommended features; keep in sync with libavutil vulkan hwcontext
+static const VkPhysicalDeviceVulkan13Features recommended_vk13 = {
+    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES,
+    .computeFullSubgroups = true,
+    .maintenance4 = true,
+    .shaderZeroInitializeWorkgroupMemory = true,
 };
 
-static const VkPhysicalDeviceHostQueryResetFeatures host_query_reset = {
-    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES,
-    .pNext = (void *) &timeline_semaphores,
-    .hostQueryReset = true,
+static const VkPhysicalDeviceVulkan12Features recommended_vk12 = {
+    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES,
+    .pNext = (void *) &recommended_vk13,
+    .bufferDeviceAddress = true,
+    .storagePushConstant8 = true,
+    .shaderInt8 = true,
+    .shaderFloat16 = true,
+    .shaderSharedInt64Atomics = true,
+    .storageBuffer8BitAccess = true,
+    .uniformAndStorageBuffer8BitAccess = true,
+    .vulkanMemoryModel = true,
+    .vulkanMemoryModelDeviceScope = true,
+};
+
+static const VkPhysicalDeviceVulkan11Features recommended_vk11 = {
+    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_FEATURES,
+    .pNext = (void *) &recommended_vk12,
+    .samplerYcbcrConversion = true,
+    .storagePushConstant16 = true,
 };
 
 const VkPhysicalDeviceFeatures2 pl_vulkan_recommended_features = {
     .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2,
-    .pNext = (void *) &host_query_reset,
+    .pNext = (void *) &recommended_vk11,
     .features = {
         .shaderImageGatherExtended = true,
         .shaderStorageImageReadWithoutFormat = true,
@@ -273,6 +277,52 @@ const VkPhysicalDeviceFeatures2 pl_vulkan_recommended_features = {
     }
 };
 
+// Required features
+static const VkPhysicalDeviceVulkan13Features required_vk13 = {
+    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES,
+    .dynamicRendering = true,
+    .synchronization2 = true,
+};
+
+static const VkPhysicalDeviceVulkan12Features required_vk12 = {
+    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES,
+    .pNext = (void *) &required_vk13,
+    .hostQueryReset = true,
+    .timelineSemaphore = true,
+};
+
+static const VkPhysicalDeviceVulkan11Features required_vk11 = {
+    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_FEATURES,
+    .pNext = (void *) &required_vk12,
+};
+
+const VkPhysicalDeviceFeatures2 pl_vulkan_required_features = {
+    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2,
+    .pNext = (void *) &required_vk11,
+};
+
+static bool check_required_features(struct vk_ctx *vk)
+{
+    #define CHECK_FEATURE(maj, min, feat) do {                                  \
+        const VkPhysicalDeviceVulkan##maj##min##Features *f;                    \
+        f = vk_find_struct(&vk->features,                                       \
+            VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_##maj##_##min##_FEATURES); \
+        if (!f || !f->feat) {                                                   \
+            PL_ERR(vk, "Missing device feature: " #feat);                       \
+            return false;                                                       \
+        }                                                                       \
+    } while (0)
+
+    CHECK_FEATURE(1, 3, dynamicRendering);
+    CHECK_FEATURE(1, 3, synchronization2);
+    CHECK_FEATURE(1, 2, hostQueryReset);
+    CHECK_FEATURE(1, 2, timelineSemaphore);
+
+    #undef CHECK_FEATURE
+    return true;
+}
+
+
 // List of mandatory device-level functions
 //
 // Note: Also includes VK_EXT_debug_utils functions, even though they aren't
@@ -285,7 +335,6 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(BindBufferMemory),
     PL_VK_DEV_FUN(BindImageMemory),
     PL_VK_DEV_FUN(CmdBeginDebugUtilsLabelEXT),
-    PL_VK_DEV_FUN(CmdBeginRenderPass),
     PL_VK_DEV_FUN(CmdBindDescriptorSets),
     PL_VK_DEV_FUN(CmdBindIndexBuffer),
     PL_VK_DEV_FUN(CmdBindPipeline),
@@ -300,32 +349,25 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(CmdDraw),
     PL_VK_DEV_FUN(CmdDrawIndexed),
     PL_VK_DEV_FUN(CmdEndDebugUtilsLabelEXT),
-    PL_VK_DEV_FUN(CmdEndRenderPass),
-    PL_VK_DEV_FUN(CmdPipelineBarrier),
     PL_VK_DEV_FUN(CmdPushConstants),
     PL_VK_DEV_FUN(CmdResetQueryPool),
-    PL_VK_DEV_FUN(CmdSetEvent),
     PL_VK_DEV_FUN(CmdSetScissor),
     PL_VK_DEV_FUN(CmdSetViewport),
     PL_VK_DEV_FUN(CmdUpdateBuffer),
-    PL_VK_DEV_FUN(CmdWaitEvents),
-    PL_VK_DEV_FUN(CmdWriteTimestamp),
+    PL_VK_DEV_FUN(CmdWriteTimestamp2KHR),
     PL_VK_DEV_FUN(CreateBuffer),
     PL_VK_DEV_FUN(CreateBufferView),
     PL_VK_DEV_FUN(CreateCommandPool),
     PL_VK_DEV_FUN(CreateComputePipelines),
     PL_VK_DEV_FUN(CreateDescriptorPool),
     PL_VK_DEV_FUN(CreateDescriptorSetLayout),
-    PL_VK_DEV_FUN(CreateEvent),
     PL_VK_DEV_FUN(CreateFence),
-    PL_VK_DEV_FUN(CreateFramebuffer),
     PL_VK_DEV_FUN(CreateGraphicsPipelines),
     PL_VK_DEV_FUN(CreateImage),
     PL_VK_DEV_FUN(CreateImageView),
     PL_VK_DEV_FUN(CreatePipelineCache),
     PL_VK_DEV_FUN(CreatePipelineLayout),
     PL_VK_DEV_FUN(CreateQueryPool),
-    PL_VK_DEV_FUN(CreateRenderPass),
     PL_VK_DEV_FUN(CreateSampler),
     PL_VK_DEV_FUN(CreateSemaphore),
     PL_VK_DEV_FUN(CreateShaderModule),
@@ -335,9 +377,7 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(DestroyDescriptorPool),
     PL_VK_DEV_FUN(DestroyDescriptorSetLayout),
     PL_VK_DEV_FUN(DestroyDevice),
-    PL_VK_DEV_FUN(DestroyEvent),
     PL_VK_DEV_FUN(DestroyFence),
-    PL_VK_DEV_FUN(DestroyFramebuffer),
     PL_VK_DEV_FUN(DestroyImage),
     PL_VK_DEV_FUN(DestroyImageView),
     PL_VK_DEV_FUN(DestroyInstance),
@@ -345,10 +385,10 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(DestroyPipelineCache),
     PL_VK_DEV_FUN(DestroyPipelineLayout),
     PL_VK_DEV_FUN(DestroyQueryPool),
-    PL_VK_DEV_FUN(DestroyRenderPass),
     PL_VK_DEV_FUN(DestroySampler),
     PL_VK_DEV_FUN(DestroySemaphore),
     PL_VK_DEV_FUN(DestroyShaderModule),
+    PL_VK_DEV_FUN(DeviceWaitIdle),
     PL_VK_DEV_FUN(EndCommandBuffer),
     PL_VK_DEV_FUN(FlushMappedMemoryRanges),
     PL_VK_DEV_FUN(FreeCommandBuffers),
@@ -361,13 +401,13 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(GetQueryPoolResults),
     PL_VK_DEV_FUN(InvalidateMappedMemoryRanges),
     PL_VK_DEV_FUN(MapMemory),
-    PL_VK_DEV_FUN(QueueSubmit),
     PL_VK_DEV_FUN(QueueWaitIdle),
-    PL_VK_DEV_FUN(ResetEvent),
     PL_VK_DEV_FUN(ResetFences),
+    PL_VK_DEV_FUN(ResetQueryPool),
     PL_VK_DEV_FUN(SetDebugUtilsObjectNameEXT),
     PL_VK_DEV_FUN(UpdateDescriptorSets),
     PL_VK_DEV_FUN(WaitForFences),
+    PL_VK_DEV_FUN(WaitSemaphores),
 };
 
 static void load_vk_fun(struct vk_ctx *vk, const struct vk_fun *fun)
@@ -426,6 +466,9 @@ void pl_vk_inst_destroy(pl_vk_inst *inst_ptr)
     pl_free_ptr((void **) inst_ptr);
 }
 
+#ifdef MSAN
+__attribute__((no_sanitize("memory")))
+#endif
 static VkBool32 VKAPI_PTR vk_dbg_utils_cb(VkDebugUtilsMessageSeverityFlagBitsEXT sev,
                                           VkDebugUtilsMessageTypeFlagsEXT msgType,
                                           const VkDebugUtilsMessengerCallbackDataEXT *data,
@@ -436,7 +479,6 @@ static VkBool32 VKAPI_PTR vk_dbg_utils_cb(VkDebugUtilsMessageSeverityFlagBitsEXT
     // MSAN really doesn't like reading from the stack-allocated memory
     // allocated by the non-instrumented vulkan library, so just comment it out
     // when building with MSAN as a cheap hack-around.
-#ifndef MSAN
 
     // Ignore errors for messages that we consider false positives
     switch (data->messageIdNumber) {
@@ -445,12 +487,18 @@ static VkBool32 VKAPI_PTR vk_dbg_utils_cb(VkDebugUtilsMessageSeverityFlagBitsEXT
     case 0xdc18ad6b: // UNASSIGNED-BestPractices-vkAllocateMemory-small-allocation
     case 0xb3d4346b: // UNASSIGNED-BestPractices-vkBindMemory-small-dedicated-allocation
     case 0x6cfe18a5: // UNASSIGNED-BestPractices-SemaphoreCount
+    // profile chain expectations
+    case 0x30f4ac70: // VUID-VkImageCreateInfo-pNext-06811
         return false;
 
     case 0x5f379b89: // UNASSIGNED-BestPractices-Error-Result
         if (strstr(data->pMessage, "VK_ERROR_FORMAT_NOT_SUPPORTED"))
             return false;
         break;
+
+    case 0xf6a37cfa: // VUID-vkGetImageSubresourceLayout-format-04461
+        // Work around https://github.com/KhronosGroup/Vulkan-Docs/issues/2109
+        return false;
     }
 
     enum pl_log_level lev;
@@ -489,8 +537,6 @@ static VkBool32 VKAPI_PTR vk_dbg_utils_cb(VkDebugUtilsMessageSeverityFlagBitsEXT
         return true;
     }
 
-#endif // MSAN
-
     return false;
 }
 
@@ -510,9 +556,9 @@ static PFN_vkGetInstanceProcAddr get_proc_addr_fallback(pl_log log,
 }
 
 #define PRINTF_VER(ver) \
-    (int) VK_VERSION_MAJOR(ver), \
-    (int) VK_VERSION_MINOR(ver), \
-    (int) VK_VERSION_PATCH(ver)
+    (int) VK_API_VERSION_MAJOR(ver), \
+    (int) VK_API_VERSION_MINOR(ver), \
+    (int) VK_API_VERSION_PATCH(ver)
 
 pl_vk_inst pl_vk_inst_create(pl_log log, const struct pl_vk_inst_params *params)
 {
@@ -540,10 +586,10 @@ pl_vk_inst pl_vk_inst_create(pl_log log, const struct pl_vk_inst_params *params)
                 PRINTF_VER(params->max_api_version), PRINTF_VER(api_ver));
     }
 
-    if (api_ver < VK_API_VERSION_1_1) {
+    if (api_ver < PL_VK_MIN_VERSION) {
         pl_fatal(log, "Instance API version %d.%d.%d is lower than the minimum "
                  "required version of %d.%d.%d, cannot proceed!",
-                 PRINTF_VER(api_ver), PRINTF_VER(VK_API_VERSION_1_1));
+                 PRINTF_VER(api_ver), PRINTF_VER(PL_VK_MIN_VERSION));
         goto error;
     }
 
@@ -613,6 +659,19 @@ debug_layers_done: ;
         }
     }
 
+    static const char *extra_layers[] = {
+        "VK_LAYER_KHRONOS_synchronization2",
+    };
+
+    for (int i = 0; i < PL_ARRAY_SIZE(extra_layers); i++) {
+        for (int n = 0; n < num_layers_avail; n++) {
+            if (strcmp(extra_layers[i], layers_avail[n].layerName) == 0) {
+                PL_ARRAY_APPEND(tmp, layers, extra_layers[i]);
+                break;
+            }
+        }
+    }
+
     // Enumerate all supported extensions
     PL_VK_LOAD_FUN(NULL, EnumerateInstanceExtensionProperties, get_addr);
     uint32_t num_exts_avail = 0;
@@ -881,6 +940,11 @@ void pl_vulkan_destroy(pl_vulkan *pl_vk)
             vk->DestroyDevice(vk->dev, PL_VK_ALLOC);
     }
 
+    for (int i = 0; i < vk->queue_locks.num; i++) {
+        for (int n = 0; n < vk->queue_locks.elem[i].num; n++)
+            pl_mutex_destroy(&vk->queue_locks.elem[i].elem[n]);
+    }
+
     pl_vk_inst_destroy(&vk->internal_instance);
     pl_mutex_destroy(&vk->lock);
     pl_free_ptr((void **) pl_vk);
@@ -959,7 +1023,8 @@ VkPhysicalDevice pl_vulkan_choose_device(pl_log log,
         GetPhysicalDeviceProperties2(devices[i], &prop);
         VkPhysicalDeviceType t = prop.properties.deviceType;
         const char *dtype = t < PL_ARRAY_SIZE(types) ? types[t].name : "unknown?";
-        PL_INFO(vk, "    GPU %d: %s (%s)", i, prop.properties.deviceName, dtype);
+        PL_INFO(vk, "    GPU %d: %s v%d.%d.%d (%s)", i, prop.properties.deviceName,
+                PRINTF_VER(prop.properties.apiVersion), dtype);
         PL_INFO(vk, "           uuid: %s", PRINT_UUID(id_props.deviceUUID));
 
         if (params->surface) {
@@ -992,7 +1057,7 @@ VkPhysicalDevice pl_vulkan_choose_device(pl_log log,
             continue;
         }
 
-        if (prop.properties.apiVersion < VK_API_VERSION_1_1) {
+        if (prop.properties.apiVersion < PL_VK_MIN_VERSION) {
             PL_DEBUG(vk, "      -> excluding due to too low API version");
             continue;
         }
@@ -1009,6 +1074,36 @@ error:
     return dev;
 }
 
+static void lock_queue_internal(void *priv, uint32_t qf, uint32_t qidx)
+{
+    struct vk_ctx *vk = priv;
+    pl_mutex_lock(&vk->queue_locks.elem[qf].elem[qidx]);
+}
+
+static void unlock_queue_internal(void *priv, uint32_t qf, uint32_t qidx)
+{
+    struct vk_ctx *vk = priv;
+    pl_mutex_unlock(&vk->queue_locks.elem[qf].elem[qidx]);
+}
+
+static void init_queue_locks(struct vk_ctx *vk, uint32_t qfnum,
+                             const VkQueueFamilyProperties *qfs)
+{
+    vk->queue_locks.elem = pl_calloc_ptr(vk->alloc, qfnum, vk->queue_locks.elem);
+    vk->queue_locks.num = qfnum;
+    for (int i = 0; i < qfnum; i++) {
+        const uint32_t qnum = qfs[i].queueCount;
+        vk->queue_locks.elem[i].elem = pl_calloc(vk->alloc, qnum, sizeof(pl_mutex));
+        vk->queue_locks.elem[i].num = qnum;
+        for (int n = 0; n < qnum; n++)
+            pl_mutex_init(&vk->queue_locks.elem[i].elem[n]);
+    }
+
+    vk->lock_queue = lock_queue_internal;
+    vk->unlock_queue = unlock_queue_internal;
+    vk->queue_ctx = vk;
+}
+
 // Find the most specialized queue supported a combination of flags. In cases
 // where there are multiple queue families at the same specialization level,
 // this finds the one with the most queues. Returns -1 if no queue was found.
@@ -1037,34 +1132,6 @@ static int find_qf(VkQueueFamilyProperties *qfs, int qfnum, VkQueueFlags flags)
     return idx;
 }
 
-typedef PL_ARRAY(VkDeviceQueueCreateInfo) qinfo_arr_t;
-
-static void add_qinfo(void *alloc, qinfo_arr_t *qinfos,
-                      VkQueueFamilyProperties *qfs, int idx, int qcount)
-{
-    if (idx < 0)
-        return;
-
-    // Check to see if we've already added this queue family
-    for (int i = 0; i < qinfos->num; i++) {
-        if (qinfos->elem[i].queueFamilyIndex == idx)
-            return;
-    }
-
-    if (!qcount)
-        qcount = qfs[idx].queueCount;
-
-    float *priorities = pl_calloc_ptr(alloc, qcount, priorities);
-    VkDeviceQueueCreateInfo qinfo = {
-        .sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO,
-        .queueFamilyIndex = idx,
-        .queueCount = PL_MIN(qcount, qfs[idx].queueCount),
-        .pQueuePriorities = priorities,
-    };
-
-    PL_ARRAY_APPEND(alloc, *qinfos, qinfo);
-}
-
 static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params)
 {
     pl_assert(vk->physd);
@@ -1075,12 +1142,12 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     vk->GetPhysicalDeviceQueueFamilyProperties(vk->physd, &qfnum, NULL);
     VkQueueFamilyProperties *qfs = pl_calloc_ptr(tmp, qfnum, qfs);
     vk->GetPhysicalDeviceQueueFamilyProperties(vk->physd, &qfnum, qfs);
+    init_queue_locks(vk, qfnum, qfs);
 
     PL_DEBUG(vk, "Queue families supported by device:");
-
     for (int i = 0; i < qfnum; i++) {
-        PL_DEBUG(vk, "    %d: flags 0x%x num %d", i,
-                 (unsigned) qfs[i].queueFlags, (int) qfs[i].queueCount);
+        PL_DEBUG(vk, "    %d: flags 0x%"PRIx32" num %"PRIu32, i,
+                 qfs[i].queueFlags, qfs[i].queueCount);
     }
 
     VkQueueFlagBits gfx_flags = VK_QUEUE_GRAPHICS_BIT;
@@ -1119,12 +1186,6 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
         }
     }
 
-    // Now that we know which QFs we want, we can create the logical device
-    qinfo_arr_t qinfos = {0};
-    add_qinfo(tmp, &qinfos, qfs, idx_gfx, params->queue_count);
-    add_qinfo(tmp, &qinfos, qfs, idx_comp, params->queue_count);
-    add_qinfo(tmp, &qinfos, qfs, idx_tf, params->queue_count);
-
     // Enumerate all supported extensions
     uint32_t num_exts_avail = 0;
     VK(vk->EnumerateDeviceExtensionProperties(vk->physd, NULL, &num_exts_avail, NULL));
@@ -1138,8 +1199,6 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     // Add all extensions we need
     if (params->surface)
         PL_ARRAY_APPEND(vk->alloc, vk->exts, VK_KHR_SWAPCHAIN_EXTENSION_NAME);
-    if (vk->api_ver < VK_API_VERSION_1_2)
-        PL_ARRAY_APPEND(vk->alloc, vk->exts, VK_KHR_TIMELINE_SEMAPHORE_EXTENSION_NAME);
 
     // Keep track of all optional function pointers associated with extensions
     PL_ARRAY(const struct vk_fun *) ext_funs = {0};
@@ -1147,7 +1206,8 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     // Add all optional device-level extensions extensions
     for (int i = 0; i < PL_ARRAY_SIZE(vk_device_extensions); i++) {
         const struct vk_ext *ext = &vk_device_extensions[i];
-        if (ext->core_ver && vk->api_ver >= ext->core_ver) {
+        uint32_t core_ver = vk_ext_promoted_ver(ext->name);
+        if (core_ver && vk->api_ver >= core_ver) {
             // Layer is already implicitly enabled by the API version
             for (const struct vk_fun *f = ext->funs; f && f->name; f++)
                 PL_ARRAY_APPEND(tmp, ext_funs,  f);
@@ -1179,72 +1239,62 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
         }
     }
 
-    VkPhysicalDeviceFeatures2 *device_features;
-    device_features = vk_chain_memdup(tmp, &pl_vulkan_recommended_features);
+    VkPhysicalDeviceFeatures2 features = {
+        .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2_KHR
+    };
 
-#ifdef VK_VERSION_1_3
-    if (vk->api_ver >= VK_API_VERSION_1_3) {
-        vk_link_struct(device_features,
-                       vk_struct_memdup(tmp, &device_maintenance4_features));
-    }
-#endif
+    vk_features_normalize(tmp, &pl_vulkan_required_features, vk->api_ver, &features);
+    vk_features_normalize(tmp, &pl_vulkan_recommended_features, vk->api_ver, &features);
+    vk_features_normalize(tmp, params->features, vk->api_ver, &features);
 
-    // Query all supported device features by constructing a pNext chain
-    // starting with the features we care about and ending with whatever
-    // features were requested by the user
-    vk->features.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2_KHR;
-    for (const VkBaseInStructure *in = device_features->pNext;
-            in; in = in->pNext)
-        vk_link_struct(&vk->features, vk_struct_memdup(vk->alloc, in));
-
-    for (const VkBaseInStructure *in = (const VkBaseInStructure *) params->features;
-            in; in = in->pNext)
-    {
-        if (vk_find_struct(&vk->features, in->sType))
-            continue; // skip structs already present
-
-        void *copy = vk_struct_memdup(vk->alloc, in);
-        if (!copy) {
-            PL_ERR(vk, "Unknown struct type %"PRIu64"?", (uint64_t) in->sType);
-            continue;
-        }
-
-        vk_link_struct(&vk->features, copy);
+    // Explicitly clear the features struct before querying feature support
+    // from the driver. This way, we don't mistakenly mark as supported
+    // features coming from structs the driver doesn't have support for.
+    VkPhysicalDeviceFeatures2 *features_sup = vk_chain_memdup(tmp, &features);;
+    for (VkBaseOutStructure *out = (void *) features_sup; out; out = out->pNext) {
+        const size_t size = vk_struct_size(out->sType);
+        memset(&out[1], 0, size - sizeof(out[0]));
     }
 
-    vk->GetPhysicalDeviceFeatures2KHR(vk->physd, &vk->features);
+    vk->GetPhysicalDeviceFeatures2KHR(vk->physd, features_sup);
 
-    const VkPhysicalDeviceTimelineSemaphoreFeatures *timeline_sem;
-    timeline_sem = vk_find_struct(&vk->features,
-        VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES);
-    if (!timeline_sem || !timeline_sem->timelineSemaphore) {
-        PL_ERR(vk, "Selected vulkan device does not support timeline semaphores!");
+    // Filter out unsupported features
+    for (VkBaseOutStructure *f = (VkBaseOutStructure *) &features; f; f = f->pNext) {
+        const VkBaseInStructure *sup = vk_find_struct(features_sup, f->sType);
+        VkBool32 *flags = (VkBool32 *) &f[1];
+        const VkBool32 *flags_sup = (const VkBool32 *) &sup[1];
+        const size_t size = vk_struct_size(f->sType) - sizeof(VkBaseOutStructure);
+        for (int i = 0; i < size / sizeof(VkBool32); i++)
+            flags[i] &= flags_sup[i];
+    }
+
+    // Construct normalized output chain
+    vk->features.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2;
+    vk_features_normalize(vk->alloc, &features, 0, &vk->features);
+    if (!check_required_features(vk)) {
+        PL_FATAL(vk, "Vulkan device does not support all required features!");
         goto error;
     }
 
-    // Go through the features chain a second time and mask every option
-    // that wasn't whitelisted by either libplacebo or the user
-    for (VkBaseOutStructure *chain = (VkBaseOutStructure *) &vk->features;
-            chain; chain = chain->pNext)
-    {
-        const VkBaseInStructure *in_a, *in_b;
-        in_a = vk_find_struct(device_features, chain->sType);
-        in_b = vk_find_struct(params->features, chain->sType);
-        in_a = PL_DEF(in_a, in_b);
-        in_b = PL_DEF(in_b, in_a);
-        pl_assert(in_a && in_b);
-
-        VkBool32 *req = (VkBool32 *) &chain[1];
-        const VkBool32 *wl_a = (const VkBool32 *) &in_a[1];
-        const VkBool32 *wl_b = (const VkBool32 *) &in_b[1];
-        size_t size = vk_struct_size(chain->sType) - sizeof(chain[0]);
-        for (int i = 0; i < size / sizeof(VkBool32); i++)
-            req[i] &= wl_a[i] || wl_b[i];
+    // Enable all queues at device creation time, to maximize compatibility
+    // with other API users (e.g. FFmpeg)
+    PL_ARRAY(VkDeviceQueueCreateInfo) qinfos = {0};
+    for (int i = 0; i < qfnum; i++) {
+        bool use_qf = i == idx_gfx || i == idx_comp || i == idx_tf;
+        use_qf |= qfs[i].queueFlags & params->extra_queues;
+        if (!use_qf)
+            continue;
+        PL_ARRAY_APPEND(tmp, qinfos, (VkDeviceQueueCreateInfo) {
+            .sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO,
+            .queueFamilyIndex = i,
+            .queueCount = qfs[i].queueCount,
+            .pQueuePriorities = pl_calloc(tmp, qfs[i].queueCount, sizeof(float)),
+        });
     }
 
     VkDeviceCreateInfo dinfo = {
         .sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO,
-        .pNext = &vk->features,
+        .pNext = &features,
         .pQueueCreateInfos = qinfos.elem,
         .queueCreateInfoCount = qinfos.num,
         .ppEnabledExtensionNames = vk->exts.elem,
@@ -1265,25 +1315,34 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
     for (int i = 0; i < ext_funs.num; i++)
         load_vk_fun(vk, ext_funs.elem[i]);
 
-    // Create the command pools
-    for (int i = 0; i < qinfos.num; i++) {
-        int qf = qinfos.elem[i].queueFamilyIndex;
-        struct vk_cmdpool *pool = vk_cmdpool_create(vk, qinfos.elem[i], qfs[qf]);
+    // Create the command pools for the queues we care about
+    const uint32_t qmax = PL_DEF(params->queue_count, UINT32_MAX);
+    for (int i = 0; i < qfnum; i++) {
+        if (i != idx_gfx && i != idx_tf && i != idx_comp)
+            continue; // ignore QFs not used internally
+
+        int qnum = qfs[i].queueCount;
+        if (qmax < qnum) {
+            PL_DEBUG(vk, "Restricting QF %d from %d queues to %d", i, qnum, qmax);
+            qnum = qmax;
+        }
+
+        struct vk_cmdpool *pool = vk_cmdpool_create(vk, i, qnum, qfs[i]);
         if (!pool)
             goto error;
         PL_ARRAY_APPEND(vk->alloc, vk->pools, pool);
 
         // Update the pool_* pointers based on the corresponding index
         const char *qf_name = NULL;
-        if (qf == idx_tf) {
+        if (i == idx_tf) {
             vk->pool_transfer = pool;
             qf_name = "transfer";
         }
-        if (qf == idx_comp) {
+        if (i == idx_comp) {
             vk->pool_compute = pool;
             qf_name = "compute";
         }
-        if (qf == idx_gfx) {
+        if (i == idx_gfx) {
             vk->pool_graphics = pool;
             qf_name = "graphics";
         }
@@ -1302,25 +1361,13 @@ error:
     return false;
 }
 
-static void lock_queue_internal(void *priv, int qf, int qidx)
-{
-    struct vk_ctx *vk = priv;
-    pl_mutex_lock(&vk->queue_locks.elem[qf][qidx]);
-}
-
-static void unlock_queue_internal(void *priv, int qf, int qidx)
-{
-    struct vk_ctx *vk = priv;
-    pl_mutex_unlock(&vk->queue_locks.elem[qf][qidx]);
-}
-
-static void lock_queue(pl_vulkan pl_vk, int qf, int qidx)
+static void lock_queue(pl_vulkan pl_vk, uint32_t qf, uint32_t qidx)
 {
     struct vk_ctx *vk = PL_PRIV(pl_vk);
     vk->lock_queue(vk->queue_ctx, qf, qidx);
 }
 
-static void unlock_queue(pl_vulkan pl_vk, int qf, int qidx)
+static void unlock_queue(pl_vulkan pl_vk, uint32_t qf, uint32_t qidx)
 {
     struct vk_ctx *vk = PL_PRIV(pl_vk);
     vk->unlock_queue(vk->queue_ctx, qf, qidx);
@@ -1365,7 +1412,6 @@ static bool finalize_context(struct pl_vulkan_t *pl_vk, int max_glsl_version)
     pl_vk->lock_queue = lock_queue;
     pl_vk->unlock_queue = unlock_queue;
 
-    uint32_t max_qf = 0;
     for (int i = 0; i < vk->pools.num; i++) {
         struct pl_vulkan_queue *queues = (struct pl_vulkan_queue *) pl_vk->queues;
         queues[i] = (struct pl_vulkan_queue) {
@@ -1379,25 +1425,6 @@ static bool finalize_context(struct pl_vulkan_t *pl_vk, int max_glsl_version)
             pl_vk->queue_compute = queues[i];
         if (vk->pools.elem[i] == vk->pool_transfer)
             pl_vk->queue_transfer = queues[i];
-
-        max_qf = PL_MAX(max_qf, queues[i].index);
-    }
-
-    if (!vk->lock_queue) {
-        vk->queue_locks.num = max_qf + 1;
-        PL_ARRAY_RESIZE(vk->alloc, vk->queue_locks, vk->queue_locks.num);
-
-        for (int i = 0; i < vk->pools.num; i++) {
-            struct vk_cmdpool *pool = vk->pools.elem[i];
-            pl_mutex **locks = vk->queue_locks.elem;
-            locks[pool->qf] = pl_alloc(vk->alloc, pool->num_queues * sizeof(pl_mutex));
-            for (int n = 0; n < pool->num_queues; n++)
-                pl_mutex_init(&locks[pool->qf][n]);
-        }
-
-        vk->lock_queue = lock_queue_internal;
-        vk->unlock_queue = unlock_queue_internal;
-        vk->queue_ctx = vk;
     }
 
     pl_assert(vk->lock_queue);
@@ -1473,14 +1500,14 @@ pl_vulkan pl_vulkan_create(pl_log log, const struct pl_vulkan_params *params)
     };
 
     vk->GetPhysicalDeviceProperties2(vk->physd, &prop);
-    vk->limits = prop.properties.limits;
+    vk->props = prop.properties;
 
     PL_INFO(vk, "Vulkan device properties:");
     PL_INFO(vk, "    Device Name: %s", prop.properties.deviceName);
-    PL_INFO(vk, "    Device ID: %x:%x", (unsigned) prop.properties.vendorID,
-            (unsigned) prop.properties.deviceID);
+    PL_INFO(vk, "    Device ID: %"PRIx32":%"PRIx32, prop.properties.vendorID,
+            prop.properties.deviceID);
     PL_INFO(vk, "    Device UUID: %s", PRINT_UUID(id_props.deviceUUID));
-    PL_INFO(vk, "    Driver version: %x", (unsigned) prop.properties.driverVersion);
+    PL_INFO(vk, "    Driver version: %"PRIx32, prop.properties.driverVersion);
     PL_INFO(vk, "    API version: %d.%d.%d", PRINTF_VER(prop.properties.apiVersion));
 
     // Needed by device_init
@@ -1491,10 +1518,10 @@ pl_vulkan pl_vulkan_create(pl_log log, const struct pl_vulkan_params *params)
                 PRINTF_VER(params->max_api_version), PRINTF_VER(vk->api_ver));
     }
 
-    if (vk->api_ver < VK_API_VERSION_1_1) {
+    if (vk->api_ver < PL_VK_MIN_VERSION) {
         PL_FATAL(vk, "Device API version %d.%d.%d is lower than the minimum "
                  "required version of %d.%d.%d, cannot proceed!",
-                 PRINTF_VER(vk->api_ver), PRINTF_VER(VK_API_VERSION_1_1));
+                 PRINTF_VER(vk->api_ver), PRINTF_VER(PL_VK_MIN_VERSION));
         goto error;
     }
 
@@ -1551,14 +1578,14 @@ pl_vulkan pl_vulkan_import(pl_log log, const struct pl_vulkan_import_params *par
 
     pl_assert(vk->GetPhysicalDeviceProperties2);
     vk->GetPhysicalDeviceProperties2(vk->physd, &prop);
-    vk->limits = prop.properties.limits;
+    vk->props = prop.properties;
 
     PL_INFO(vk, "Imported vulkan device properties:");
     PL_INFO(vk, "    Device Name: %s", prop.properties.deviceName);
-    PL_INFO(vk, "    Device ID: %x:%x", (unsigned) prop.properties.vendorID,
-            (unsigned) prop.properties.deviceID);
+    PL_INFO(vk, "    Device ID: %"PRIx32":%"PRIx32, prop.properties.vendorID,
+            prop.properties.deviceID);
     PL_INFO(vk, "    Device UUID: %s", PRINT_UUID(id_props.deviceUUID));
-    PL_INFO(vk, "    Driver version: %d", (int) prop.properties.driverVersion);
+    PL_INFO(vk, "    Driver version: %"PRIx32, prop.properties.driverVersion);
     PL_INFO(vk, "    API version: %d.%d.%d", PRINTF_VER(prop.properties.apiVersion));
 
     vk->api_ver = prop.properties.apiVersion;
@@ -1568,46 +1595,18 @@ pl_vulkan pl_vulkan_import(pl_log log, const struct pl_vulkan_import_params *par
                 PRINTF_VER(params->max_api_version), PRINTF_VER(vk->api_ver));
     }
 
-    if (vk->api_ver < VK_API_VERSION_1_1) {
+    if (vk->api_ver < PL_VK_MIN_VERSION) {
         PL_FATAL(vk, "Device API version %d.%d.%d is lower than the minimum "
                  "required version of %d.%d.%d, cannot proceed!",
-                 PRINTF_VER(vk->api_ver), PRINTF_VER(VK_API_VERSION_1_1));
+                 PRINTF_VER(vk->api_ver), PRINTF_VER(PL_VK_MIN_VERSION));
         goto error;
     }
 
-    VkPhysicalDeviceFeatures2 *features;
-    features = vk_chain_memdup(vk->alloc, params->features);
-    if (features) {
-        // Go through and replace all meta-features structs by their individual
-        // extension variants, since that's what we check for in our code
-        const VkPhysicalDeviceVulkan12Features *vk12 = vk_find_struct(features,
-            VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES);
-
-        if (vk12 && vk12->hostQueryReset) {
-            const VkPhysicalDeviceHostQueryResetFeatures hqr = {
-                .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES,
-                .hostQueryReset = true,
-            };
-            vk_link_struct(features, vk_struct_memdup(vk->alloc, &hqr));
-        }
-
-        if (vk12 && vk12->timelineSemaphore) {
-            const VkPhysicalDeviceTimelineSemaphoreFeatures ts = {
-                .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES,
-                .timelineSemaphore = true,
-            };
-            vk_link_struct(features, vk_struct_memdup(vk->alloc, &ts));
-        }
-
-        vk->features = *features;
-    }
-
-    const VkPhysicalDeviceTimelineSemaphoreFeatures *timeline_sem;
-    timeline_sem = vk_find_struct(&vk->features,
-        VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES);
-    if (!timeline_sem || !timeline_sem->timelineSemaphore) {
-        PL_ERR(vk, "Imported Vulkan device does not support timeline "
-               "semaphores. Please enable this device feature.");
+    vk->features.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2;
+    vk_features_normalize(vk->alloc, params->features, 0, &vk->features);
+    if (!check_required_features(vk)) {
+        PL_FATAL(vk, "Imported Vulkan device was not created with all required "
+                 "features!");
         goto error;
     }
 
@@ -1618,11 +1617,14 @@ pl_vulkan pl_vulkan_import(pl_log log, const struct pl_vulkan_import_params *par
     // Load all of the optional functions from the extensions enabled
     for (int i = 0; i < PL_ARRAY_SIZE(vk_device_extensions); i++) {
         const struct vk_ext *ext = &vk_device_extensions[i];
+        uint32_t core_ver = vk_ext_promoted_ver(ext->name);
+        if (core_ver && vk->api_ver >= core_ver) {
+            for (const struct vk_fun *f = ext->funs; f && f->name; f++)
+                load_vk_fun(vk, f);
+            continue;
+        }
         for (int n = 0; n < params->num_extensions; n++) {
-            if (strcmp(ext->name, params->extensions[n]) == 0 ||
-                (ext->core_ver && ext->core_ver >= vk->api_ver))
-            {
-                // Extension is available, directly load it
+            if (strcmp(ext->name, params->extensions[n]) == 0) {
                 for (const struct vk_fun *f = ext->funs; f && f->name; f++)
                     load_vk_fun(vk, f);
                 break;
@@ -1634,6 +1636,8 @@ pl_vulkan pl_vulkan_import(pl_log log, const struct pl_vulkan_import_params *par
     vk->GetPhysicalDeviceQueueFamilyProperties(vk->physd, &qfnum, NULL);
     VkQueueFamilyProperties *qfs = pl_calloc_ptr(tmp, qfnum, qfs);
     vk->GetPhysicalDeviceQueueFamilyProperties(vk->physd, &qfnum, qfs);
+    if (!params->lock_queue)
+        init_queue_locks(vk, qfnum, qfs);
 
     // Create the command pools for each unique qf that exists
     struct {
@@ -1675,12 +1679,7 @@ pl_vulkan pl_vulkan_import(pl_log log, const struct pl_vulkan_import_params *par
             }
         }
 
-        struct VkDeviceQueueCreateInfo qinfo = {
-            .queueFamilyIndex = qf,
-            .queueCount = qinfos[i].info->count,
-        };
-
-        *pool = vk_cmdpool_create(vk, qinfo, qfs[qf]);
+        *pool = vk_cmdpool_create(vk, qf, qinfos[i].info->count, qfs[qf]);
         if (!*pool)
             goto error;
         PL_ARRAY_APPEND(vk->alloc, vk->pools, *pool);
