@@ -18,7 +18,7 @@
 #include "formats.h"
 
 #define FMT(_name, num, size, ftype, bits, idx) \
-    (struct pl_fmt) {                           \
+    (struct pl_fmt_t) {                         \
         .name = _name,                          \
         .type = PL_FMT_##ftype,                 \
         .num_components  = num,                 \
@@ -40,7 +40,7 @@
         IDX(0, 1, 2, 3))
 
 #define EMUFMT(_name, in, en, ib, eb, ftype)    \
-    (struct pl_fmt) {                           \
+    (struct pl_fmt_t) {                         \
         .name = _name,                          \
         .type = PL_FMT_##ftype,                 \
         .num_components  = en,                  \
@@ -54,6 +54,30 @@
         .sample_order    = IDX(0, 1, 2, 3),     \
     }
 
+#define PACKED16FMT(_name, num, b)              \
+    (struct pl_fmt_t) {                         \
+        .name            = _name,               \
+        .type            = PL_FMT_UNORM,        \
+        .num_components  = num,                 \
+        .component_depth = BITS(b, b, b, b),    \
+        .internal_size   = (num) * 2,           \
+        .texel_size      = (num) * 2,           \
+        .texel_align     = (num) * 2,           \
+        .host_bits       = BITS(16, 16, 16, 16),\
+        .sample_order    = IDX(0, 1, 2, 3),     \
+    }
+
+#define PLANARFMT(_name, planes, size, bits)    \
+    (struct pl_fmt_t) {                         \
+        .name            = _name,               \
+        .type            = PL_FMT_UNORM,        \
+        .num_planes      = planes,              \
+        .num_components  = 3,                   \
+        .component_depth = {bits, bits, bits},  \
+        .internal_size   = size,                \
+        .opaque          = true,                \
+    }
+
 static const struct vk_format rgb8e = {
     .tfmt   = VK_FORMAT_R8G8B8A8_UNORM,
     .bfmt   = VK_FORMAT_R8G8B8_UNORM,
@@ -68,7 +92,7 @@ static const struct vk_format rgb16e = {
     .fmt    = EMUFMT("rgb16", 4, 3, 16, 16, UNORM),
 };
 
-const struct vk_format vk_formats[] = {
+static const struct vk_format vk_formats[] = {
     // Regular, byte-aligned integer formats
     {VK_FORMAT_R8_UNORM,              REGFMT("r8",       1,  8, UNORM)},
     {VK_FORMAT_R8G8_UNORM,            REGFMT("rg8",      2,  8, UNORM)},
@@ -167,6 +191,184 @@ const struct vk_format vk_formats[] = {
     {VK_FORMAT_A2B10G10R10_SINT_PACK32,  FMT("rgb10a2i", 4,  4, SINT,  BITS(10, 10, 10, 2), IDX(0, 1, 2, 3))},
     {VK_FORMAT_A2R10G10B10_SINT_PACK32,  FMT("bgr10a2i", 4,  4, SINT,  BITS(10, 10, 10, 2), IDX(2, 1, 0, 3))},
 
+
+    // Packed 16 bit formats
+    {VK_FORMAT_R10X6_UNORM_PACK16,                  PACKED16FMT("rx10",         1, 10)},
+    {VK_FORMAT_R10X6G10X6_UNORM_2PACK16,            PACKED16FMT("rxgx10",       2, 10)},
+    {VK_FORMAT_R10X6G10X6B10X6A10X6_UNORM_4PACK16,  PACKED16FMT("rxgxbxax10",   4, 10)},
+    {VK_FORMAT_R12X4_UNORM_PACK16,                  PACKED16FMT("rx12",         1, 12)},
+    {VK_FORMAT_R12X4G12X4_UNORM_2PACK16,            PACKED16FMT("rxgx12",       2, 12)},
+    {VK_FORMAT_R12X4G12X4B12X4A12X4_UNORM_4PACK16,  PACKED16FMT("rxgxbxax12",   4, 12)},
+
+    // Planar formats
+    {VK_FORMAT_G8_B8_R8_3PLANE_420_UNORM, PLANARFMT("g8_b8_r8_420", 3, 12, 8),
+        .pfmt = {
+            {VK_FORMAT_R8_UNORM},
+            {VK_FORMAT_R8_UNORM, .sx = 1, .sy = 1},
+            {VK_FORMAT_R8_UNORM, .sx = 1, .sy = 1},
+        },
+    },
+    {VK_FORMAT_G8_B8_R8_3PLANE_422_UNORM, PLANARFMT("g8_b8_r8_422", 3, 16, 8),
+        .pfmt = {
+            {VK_FORMAT_R8_UNORM},
+            {VK_FORMAT_R8_UNORM, .sx = 1},
+            {VK_FORMAT_R8_UNORM, .sx = 1},
+        },
+    },
+    {VK_FORMAT_G8_B8_R8_3PLANE_444_UNORM, PLANARFMT("g8_b8_r8_444", 3, 24, 8),
+        .pfmt = {
+            {VK_FORMAT_R8_UNORM},
+            {VK_FORMAT_R8_UNORM},
+            {VK_FORMAT_R8_UNORM},
+        },
+    },
+
+    {VK_FORMAT_G16_B16_R16_3PLANE_420_UNORM, PLANARFMT("g16_b16_r16_420", 3, 24, 16),
+        .pfmt = {
+            {VK_FORMAT_R16_UNORM},
+            {VK_FORMAT_R16_UNORM, .sx = 1, .sy = 1},
+            {VK_FORMAT_R16_UNORM, .sx = 1, .sy = 1},
+        },
+    },
+    {VK_FORMAT_G16_B16_R16_3PLANE_422_UNORM, PLANARFMT("g16_b16_r16_422", 3, 32, 16),
+        .pfmt = {
+            {VK_FORMAT_R16_UNORM},
+            {VK_FORMAT_R16_UNORM, .sx = 1},
+            {VK_FORMAT_R16_UNORM, .sx = 1},
+        },
+    },
+    {VK_FORMAT_G16_B16_R16_3PLANE_444_UNORM, PLANARFMT("g16_b16_r16_444", 3, 48, 16),
+        .pfmt = {
+            {VK_FORMAT_R16_UNORM},
+            {VK_FORMAT_R16_UNORM},
+            {VK_FORMAT_R16_UNORM},
+        },
+    },
+
+    {VK_FORMAT_G10X6_B10X6_R10X6_3PLANE_420_UNORM_3PACK16, PLANARFMT("gx10_bx10_rx10_420", 3, 24, 10),
+        .pfmt = {
+            {VK_FORMAT_R10X6_UNORM_PACK16},
+            {VK_FORMAT_R10X6_UNORM_PACK16, .sx = 1, .sy = 1},
+            {VK_FORMAT_R10X6_UNORM_PACK16, .sx = 1, .sy = 1},
+        },
+    },
+    {VK_FORMAT_G10X6_B10X6_R10X6_3PLANE_422_UNORM_3PACK16, PLANARFMT("gx10_bx10_rx10_422", 3, 32, 10),
+        .pfmt = {
+            {VK_FORMAT_R10X6_UNORM_PACK16},
+            {VK_FORMAT_R10X6_UNORM_PACK16, .sx = 1},
+            {VK_FORMAT_R10X6_UNORM_PACK16, .sx = 1},
+        },
+    },
+    {VK_FORMAT_G10X6_B10X6_R10X6_3PLANE_444_UNORM_3PACK16, PLANARFMT("gx10_bx10_rx10_444", 3, 48, 10),
+        .pfmt = {
+            {VK_FORMAT_R10X6_UNORM_PACK16},
+            {VK_FORMAT_R10X6_UNORM_PACK16},
+            {VK_FORMAT_R10X6_UNORM_PACK16},
+        },
+    },
+
+    {VK_FORMAT_G12X4_B12X4_R12X4_3PLANE_420_UNORM_3PACK16, PLANARFMT("gx12_bx12_rx12_420", 3, 24, 12),
+        .pfmt = {
+            {VK_FORMAT_R12X4_UNORM_PACK16},
+            {VK_FORMAT_R12X4_UNORM_PACK16, .sx = 1, .sy = 1},
+            {VK_FORMAT_R12X4_UNORM_PACK16, .sx = 1, .sy = 1},
+        },
+    },
+    {VK_FORMAT_G12X4_B12X4_R12X4_3PLANE_422_UNORM_3PACK16, PLANARFMT("gx12_bx12_rx12_422", 3, 32, 12),
+        .pfmt = {
+            {VK_FORMAT_R12X4_UNORM_PACK16},
+            {VK_FORMAT_R12X4_UNORM_PACK16, .sx = 1},
+            {VK_FORMAT_R12X4_UNORM_PACK16, .sx = 1},
+        },
+    },
+    {VK_FORMAT_G12X4_B12X4_R12X4_3PLANE_444_UNORM_3PACK16, PLANARFMT("gx12_bx12_rx12_444", 3, 48, 12),
+        .pfmt = {
+            {VK_FORMAT_R12X4_UNORM_PACK16},
+            {VK_FORMAT_R12X4_UNORM_PACK16},
+            {VK_FORMAT_R12X4_UNORM_PACK16},
+        },
+    },
+
+    {VK_FORMAT_G8_B8R8_2PLANE_420_UNORM, PLANARFMT("g8_br8_420", 2, 12, 8),
+        .pfmt = {
+            {VK_FORMAT_R8_UNORM},
+            {VK_FORMAT_R8G8_UNORM, .sx = 1, .sy = 1},
+        },
+    },
+    {VK_FORMAT_G8_B8R8_2PLANE_422_UNORM, PLANARFMT("g8_br8_422", 2, 16, 8),
+        .pfmt = {
+            {VK_FORMAT_R8_UNORM},
+            {VK_FORMAT_R8G8_UNORM, .sx = 1},
+        },
+    },
+    {VK_FORMAT_G8_B8R8_2PLANE_444_UNORM, PLANARFMT("g8_br8_444", 2, 24, 8),
+        .min_ver = VK_API_VERSION_1_3,
+        .pfmt = {
+            {VK_FORMAT_R8_UNORM},
+            {VK_FORMAT_R8G8_UNORM},
+        },
+    },
+
+    {VK_FORMAT_G16_B16R16_2PLANE_420_UNORM, PLANARFMT("g16_br16_420", 2, 24, 16),
+        .pfmt = {
+            {VK_FORMAT_R16_UNORM},
+            {VK_FORMAT_R16G16_UNORM, .sx = 1, .sy = 1},
+        },
+    },
+    {VK_FORMAT_G16_B16R16_2PLANE_422_UNORM, PLANARFMT("g16_br16_422", 2, 32, 16),
+        .pfmt = {
+            {VK_FORMAT_R16_UNORM},
+            {VK_FORMAT_R16G16_UNORM, .sx = 1},
+        },
+    },
+    {VK_FORMAT_G16_B16R16_2PLANE_444_UNORM, PLANARFMT("g16_br16_444", 2, 48, 16),
+        .min_ver = VK_API_VERSION_1_3,
+        .pfmt = {
+            {VK_FORMAT_R16_UNORM},
+            {VK_FORMAT_R16G16_UNORM},
+        },
+    },
+
+    {VK_FORMAT_G10X6_B10X6R10X6_2PLANE_420_UNORM_3PACK16, PLANARFMT("gx10_bxrx10_420", 2, 24, 10),
+        .pfmt = {
+            {VK_FORMAT_R10X6_UNORM_PACK16},
+            {VK_FORMAT_R10X6G10X6_UNORM_2PACK16, .sx = 1, .sy = 1},
+        },
+    },
+    {VK_FORMAT_G10X6_B10X6R10X6_2PLANE_422_UNORM_3PACK16, PLANARFMT("gx10_bxrx10_422", 2, 32, 10),
+        .pfmt = {
+            {VK_FORMAT_R10X6_UNORM_PACK16},
+            {VK_FORMAT_R10X6G10X6_UNORM_2PACK16, .sx = 1},
+        },
+    },
+    {VK_FORMAT_G10X6_B10X6R10X6_2PLANE_444_UNORM_3PACK16, PLANARFMT("gx10_bxrx10_444", 2, 48, 10),
+        .min_ver = VK_API_VERSION_1_3,
+        .pfmt = {
+            {VK_FORMAT_R10X6_UNORM_PACK16},
+            {VK_FORMAT_R10X6G10X6_UNORM_2PACK16},
+        },
+    },
+
+    {VK_FORMAT_G12X4_B12X4R12X4_2PLANE_420_UNORM_3PACK16, PLANARFMT("gx12_bxrx12_420", 2, 24, 12),
+        .pfmt = {
+            {VK_FORMAT_R12X4_UNORM_PACK16},
+            {VK_FORMAT_R12X4G12X4_UNORM_2PACK16, .sx = 1, .sy = 1},
+        },
+    },
+    {VK_FORMAT_G12X4_B12X4R12X4_2PLANE_422_UNORM_3PACK16, PLANARFMT("gx12_bxrx12_422", 2, 32, 12),
+        .pfmt = {
+            {VK_FORMAT_R12X4_UNORM_PACK16},
+            {VK_FORMAT_R12X4G12X4_UNORM_2PACK16, .sx = 1},
+        },
+    },
+    {VK_FORMAT_G12X4_B12X4R12X4_2PLANE_444_UNORM_3PACK16, PLANARFMT("gx12_bxrx12_444", 2, 48, 12),
+        .min_ver = VK_API_VERSION_1_3,
+        .pfmt = {
+            {VK_FORMAT_R12X4_UNORM_PACK16},
+            {VK_FORMAT_R12X4G12X4_UNORM_2PACK16},
+        },
+    },
+
     {0}
 };
 
@@ -175,7 +377,7 @@ const struct vk_format vk_formats[] = {
 #undef REGFMT
 #undef FMT
 
-void vk_setup_formats(struct pl_gpu *gpu)
+void vk_setup_formats(struct pl_gpu_t *gpu)
 {
     struct pl_vk *p = PL_PRIV(gpu);
     struct vk_ctx *vk = p->vk;
@@ -187,6 +389,10 @@ void vk_setup_formats(struct pl_gpu *gpu)
     for (const struct vk_format *pvk_fmt = vk_formats; pvk_fmt->tfmt; pvk_fmt++) {
         const struct vk_format *vk_fmt = pvk_fmt;
 
+        // Skip formats that require a too new version of Vulkan
+        if (vk_fmt->min_ver > vk->api_ver)
+            continue;
+
         // Skip formats with innately emulated representation if unsupported
         if (vk_fmt->fmt.emulated && !has_emu)
             continue;
@@ -229,11 +435,14 @@ void vk_setup_formats(struct pl_gpu *gpu)
             } else {
                 bufflags = 0;
             }
+        } else if (vk_fmt->fmt.num_planes) {
+            // Planar textures cannot be used directly
+            texflags = bufflags = 0;
         }
 
         pl_log_level_cap(vk->log, PL_LOG_NONE);
 
-        struct pl_fmt *fmt = pl_alloc_obj(gpu, fmt, struct pl_fmt_vk);
+        struct pl_fmt_t *fmt = pl_alloc_obj(gpu, fmt, struct pl_fmt_vk);
         struct pl_fmt_vk *fmtp = PL_PRIV(fmt);
         *fmt = vk_fmt->fmt;
         *fmtp = (struct pl_fmt_vk) {
@@ -373,9 +582,9 @@ void vk_setup_formats(struct pl_gpu *gpu)
             int real_comps = PL_DEF(vk_fmt->icomps, fmt->num_components);
             fmt->glsl_format = pl_fmt_glsl_format(fmt, real_comps);
             if (!fmt->glsl_format && !has_nofmt) {
-                PL_WARN(gpu, "Storable format '%s' has no matching GLSL "
-                        "format qualifier but read/write without format "
-                        "is not supported.. disabling", fmt->name);
+                PL_DEBUG(gpu, "Storable format '%s' has no matching GLSL "
+                         "format qualifier but read/write without format "
+                         "is not supported.. disabling", fmt->name);
                 fmt->caps &= ~storable;
             }
         }
@@ -383,6 +592,20 @@ void vk_setup_formats(struct pl_gpu *gpu)
         if (fmt->caps & storable)
             fmt->caps |= PL_FMT_CAP_READWRITE;
 
+        // Pick sub-plane formats for planar formats
+        for (int n = 0; n < fmt->num_planes; n++) {
+            for (int i = 0; i < formats.num; i++) {
+                if (formats.elem[i]->signature == vk_fmt->pfmt[n].fmt) {
+                    fmt->planes[n].format = formats.elem[i];
+                    fmt->planes[n].shift_x = vk_fmt->pfmt[n].sx;
+                    fmt->planes[n].shift_y = vk_fmt->pfmt[n].sy;
+                    break;
+                }
+            }
+
+            pl_assert(fmt->planes[n].format);
+        }
+
         PL_ARRAY_APPEND(gpu, formats, fmt);
     }
 
