@@ -26,7 +26,7 @@
 // Gives us enough queries for 8 results
 #define QUERY_POOL_SIZE 16
 
-struct pl_timer {
+struct pl_timer_t {
     VkQueryPool qpool; // even=start, odd=stop
     int index_write; // next index to write to
     int index_read; // next index to read from
@@ -54,7 +54,7 @@ static pl_timer vk_timer_create(pl_gpu gpu)
     struct vk_ctx *vk = p->vk;
 
     pl_timer timer = pl_alloc_ptr(NULL, timer);
-    *timer = (struct pl_timer) {0};
+    *timer = (struct pl_timer_t) {0};
 
     struct VkQueryPoolCreateInfo qinfo = {
         .sType = VK_STRUCTURE_TYPE_QUERY_POOL_CREATE_INFO,
@@ -72,15 +72,7 @@ error:
 
 static void vk_timer_destroy(pl_gpu gpu, pl_timer timer)
 {
-    struct pl_vk *p = PL_PRIV(gpu);
-    struct vk_ctx *vk = p->vk;
-    pl_mutex_lock(&p->recording);
-    if (p->cmd) {
-        vk_cmd_callback(p->cmd, (vk_cb) timer_destroy_cb, gpu, timer);
-    } else {
-        vk_dev_callback(vk, (vk_cb) timer_destroy_cb, gpu, timer);
-    }
-    pl_mutex_unlock(&p->recording);
+    vk_gpu_idle_callback(gpu, (vk_cb) timer_destroy_cb, gpu, timer);
 }
 
 static uint64_t vk_timer_query(pl_gpu gpu, pl_timer timer)
@@ -104,7 +96,7 @@ static uint64_t vk_timer_query(pl_gpu gpu, pl_timer timer)
     switch (res) {
     case VK_SUCCESS:
         timer->index_read = (timer->index_read + 2) % QUERY_POOL_SIZE;
-        return (ts[1] - ts[0]) * vk->limits.timestampPeriod;
+        return (ts[1] - ts[0]) * vk->props.limits.timestampPeriod;
     case VK_NOT_READY:
         return 0;
     default:
@@ -136,17 +128,13 @@ static void timer_begin(pl_gpu gpu, struct vk_cmd *cmd, pl_timer timer)
     if (cmd->pool->props.queueFlags & reset_flags) {
         // Use direct command buffer resets
         vk->CmdResetQueryPool(cmd->buf, timer->qpool, timer->index_write, 2);
-    } else if (p->host_query_reset) {
-        // Use host query resets
-        vk->ResetQueryPoolEXT(vk->dev, timer->qpool, timer->index_write, 2);
     } else {
-        PL_TRACE(gpu, "QF %d supports no mechanism for resetting queries",
-                 cmd->pool->qf);
-        return;
+        // Use host query reset
+        vk->ResetQueryPool(vk->dev, timer->qpool, timer->index_write, 2);
     }
 
-    vk->CmdWriteTimestamp(cmd->buf, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
-                          timer->qpool, timer->index_write);
+    vk->CmdWriteTimestamp2KHR(cmd->buf, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT,
+                              timer->qpool, timer->index_write);
 
     p->cmd_timer = timer;
 }
@@ -219,8 +207,8 @@ bool _end_cmd(pl_gpu gpu, struct vk_cmd **pcmd, bool submit)
 
     if (p->cmd_timer) {
         pl_timer timer = p->cmd_timer;
-        vk->CmdWriteTimestamp(cmd->buf, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
-                              timer->qpool, timer->index_write + 1);
+        vk->CmdWriteTimestamp2KHR(cmd->buf, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT,
+                                  timer->qpool, timer->index_write + 1);
 
         timer->pending |= timer_bit(timer->index_write);
         vk_cmd_callback(cmd, (vk_cb) timer_end_cb, timer,
@@ -245,12 +233,25 @@ bool _end_cmd(pl_gpu gpu, struct vk_cmd **pcmd, bool submit)
     return ret;
 }
 
+void vk_gpu_idle_callback(pl_gpu gpu, vk_cb cb, const void *priv, const void *arg)
+{
+    struct pl_vk *p = PL_PRIV(gpu);
+    struct vk_ctx *vk = p->vk;
+
+    pl_mutex_lock(&p->recording);
+    if (p->cmd) {
+        vk_cmd_callback(p->cmd, cb, priv, arg);
+    } else {
+        vk_dev_callback(vk, cb, priv, arg);
+    }
+    pl_mutex_unlock(&p->recording);
+}
+
 static void vk_gpu_destroy(pl_gpu gpu)
 {
     struct pl_vk *p = PL_PRIV(gpu);
     struct vk_ctx *vk = p->vk;
 
-    pl_dispatch_destroy(&p->dp);
     vk_cmd_submit(vk, &p->cmd);
     vk_wait_idle(vk);
 
@@ -368,6 +369,11 @@ static pl_handle_caps vk_tex_handle_caps(struct vk_ctx *vk, bool import)
         }
     }
 
+#ifdef VK_EXT_metal_objects
+    if (vk->ExportMetalObjectsEXT && import)
+        caps |= PL_HANDLE_MTL_TEX | PL_HANDLE_IOSURFACE;
+#endif
+
     return caps;
 }
 
@@ -376,22 +382,44 @@ static const VkFilter filters[PL_TEX_SAMPLE_MODE_COUNT] = {
     [PL_TEX_SAMPLE_LINEAR]  = VK_FILTER_LINEAR,
 };
 
+static inline struct pl_spirv_version get_spirv_version(const struct vk_ctx *vk)
+{
+    if (vk->api_ver >= VK_API_VERSION_1_3) {
+        const VkPhysicalDeviceMaintenance4Features *device_maintenance4;
+        device_maintenance4 = vk_find_struct(&vk->features,
+            VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_FEATURES);
+
+        if (device_maintenance4 && device_maintenance4->maintenance4) {
+            return (struct pl_spirv_version) {
+                .env_version = VK_API_VERSION_1_3,
+                .spv_version = PL_SPV_VERSION(1, 6),
+            };
+        }
+    }
+
+    pl_assert(vk->api_ver >= VK_API_VERSION_1_2);
+    return (struct pl_spirv_version) {
+        .env_version = VK_API_VERSION_1_2,
+        .spv_version = PL_SPV_VERSION(1, 5),
+    };
+}
+
 static const struct pl_gpu_fns pl_fns_vk;
 
 pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
 {
     pl_assert(vk->dev);
 
-    struct pl_gpu *gpu = pl_zalloc_obj(NULL, gpu, struct pl_vk);
+    struct pl_gpu_t *gpu = pl_zalloc_obj(NULL, gpu, struct pl_vk);
     gpu->log = vk->log;
-    gpu->ctx = gpu->log;
 
     struct pl_vk *p = PL_PRIV(gpu);
     pl_mutex_init(&p->recording);
     p->impl = pl_fns_vk;
     p->vk = vk;
 
-    p->spirv = spirv_compiler_create(vk->log);
+    const struct pl_spirv_version spirv_ver = get_spirv_version(vk);
+    p->spirv = spirv_compiler_create(vk->log, &spirv_ver);
     if (!p->spirv)
         goto error;
 
@@ -425,28 +453,36 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
         .pNext = &host_props,
     };
 
+    bool is_portability = false;
+
 #ifdef VK_KHR_portability_subset
     VkPhysicalDevicePortabilitySubsetPropertiesKHR port_props = {
         .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_PORTABILITY_SUBSET_PROPERTIES_KHR,
         .minVertexInputBindingStrideAlignment = 1,
     };
-    vk_link_struct(&props, &port_props);
+
+    for (int i = 0; i < vk->exts.num; i++) {
+        if (!strcmp(vk->exts.elem[i], VK_KHR_PORTABILITY_SUBSET_EXTENSION_NAME)) {
+            vk_link_struct(&props, &port_props);
+            is_portability = true;
+        }
+    }
 #endif
 
-
     vk->GetPhysicalDeviceProperties2(vk->physd, &props);
+    VkPhysicalDeviceLimits limits = props.properties.limits;
 
     // Determine GLSL features and limits
     gpu->glsl = (struct pl_glsl_version) {
         .version = 450,
         .vulkan = true,
         .compute = true,
-        .max_shmem_size = vk->limits.maxComputeSharedMemorySize,
-        .max_group_threads = vk->limits.maxComputeWorkGroupInvocations,
+        .max_shmem_size = limits.maxComputeSharedMemorySize,
+        .max_group_threads = limits.maxComputeWorkGroupInvocations,
         .max_group_size = {
-            vk->limits.maxComputeWorkGroupSize[0],
-            vk->limits.maxComputeWorkGroupSize[1],
-            vk->limits.maxComputeWorkGroupSize[2],
+            limits.maxComputeWorkGroupSize[0],
+            limits.maxComputeWorkGroupSize[1],
+            limits.maxComputeWorkGroupSize[2],
         },
     };
 
@@ -465,8 +501,8 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
     }
 
     if (vk->features.features.shaderImageGatherExtended) {
-        gpu->glsl.min_gather_offset = vk->limits.minTexelGatherOffset;
-        gpu->glsl.max_gather_offset = vk->limits.maxTexelGatherOffset;
+        gpu->glsl.min_gather_offset = limits.minTexelGatherOffset;
+        gpu->glsl.max_gather_offset = limits.maxTexelGatherOffset;
     }
 
     gpu->limits = (struct pl_gpu_limits) {
@@ -474,34 +510,36 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
         .thread_safe        = true,
         .callbacks          = true,
         // pl_buf
-        .max_buf_size       = SIZE_MAX, // no limit imposed by vulkan
-        .max_ubo_size       = vk->limits.maxUniformBufferRange,
-        .max_ssbo_size      = vk->limits.maxStorageBufferRange,
-        .max_vbo_size       = SIZE_MAX,
-        .max_mapped_size    = SIZE_MAX,
-        .max_buffer_texels  = vk->limits.maxTexelBufferElements,
+        .max_buf_size       = vk_malloc_avail(vk->ma, 0),
+        .max_ubo_size       = limits.maxUniformBufferRange,
+        .max_ssbo_size      = limits.maxStorageBufferRange,
+        .max_vbo_size       = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT),
+        .max_mapped_size    = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT),
+        .max_buffer_texels  = limits.maxTexelBufferElements,
         .align_host_ptr     = host_props.minImportedHostPointerAlignment,
+        .host_cached        = vk_malloc_avail(vk->ma, VK_MEMORY_PROPERTY_HOST_CACHED_BIT),
         // pl_tex
-        .max_tex_1d_dim     = vk->limits.maxImageDimension1D,
-        .max_tex_2d_dim     = vk->limits.maxImageDimension2D,
-        .max_tex_3d_dim     = vk->limits.maxImageDimension3D,
+        .max_tex_1d_dim     = limits.maxImageDimension1D,
+        .max_tex_2d_dim     = limits.maxImageDimension2D,
+        .max_tex_3d_dim     = limits.maxImageDimension3D,
         .blittable_1d_3d    = true,
         .buf_transfer       = true,
-        .align_tex_xfer_pitch  = vk->limits.optimalBufferCopyRowPitchAlignment,
-        .align_tex_xfer_offset = pl_lcm(vk->limits.optimalBufferCopyOffsetAlignment, 4),
+        .align_tex_xfer_pitch  = limits.optimalBufferCopyRowPitchAlignment,
+        .align_tex_xfer_offset = pl_lcm(limits.optimalBufferCopyOffsetAlignment, 4),
         // pl_pass
         .max_variable_comps = 0, // vulkan doesn't support these at all
         .max_constants      = SIZE_MAX,
-        .max_pushc_size     = vk->limits.maxPushConstantsSize,
+        .array_size_constants = !is_portability,
+        .max_pushc_size     = limits.maxPushConstantsSize,
 #ifdef VK_KHR_portability_subset
         .align_vertex_stride = port_props.minVertexInputBindingStrideAlignment,
 #else
         .align_vertex_stride = 1,
 #endif
         .max_dispatch = {
-            vk->limits.maxComputeWorkGroupCount[0],
-            vk->limits.maxComputeWorkGroupCount[1],
-            vk->limits.maxComputeWorkGroupCount[2],
+            limits.maxComputeWorkGroupCount[0],
+            limits.maxComputeWorkGroupCount[1],
+            limits.maxComputeWorkGroupCount[2],
         },
         .fragment_queues    = vk->pool_graphics->num_queues,
         .compute_queues     = vk->pool_compute->num_queues,
@@ -527,20 +565,12 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
     if (vk->CmdPushDescriptorSetKHR)
         p->max_push_descriptors = pushd_props.maxPushDescriptors;
 
-    if (vk->ResetQueryPoolEXT) {
-        const VkPhysicalDeviceHostQueryResetFeatures *host_query_reset;
-        host_query_reset = vk_find_struct(&vk->features,
-            VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_HOST_QUERY_RESET_FEATURES);
-        if (host_query_reset)
-            p->host_query_reset = host_query_reset->hostQueryReset;
-    }
-
     vk_setup_formats(gpu);
 
     // Compute the correct minimum texture alignment
     p->min_texel_alignment = 1;
     for (int i = 0; i < gpu->num_formats; i++) {
-        if (gpu->formats[i]->emulated)
+        if (gpu->formats[i]->emulated || gpu->formats[i]->opaque)
             continue;
         size_t texel_size = gpu->formats[i]->texel_size;
         p->min_texel_alignment = pl_lcm(p->min_texel_alignment, texel_size);
@@ -570,8 +600,6 @@ pl_gpu pl_gpu_create_vk(struct vk_ctx *vk)
         }
     }
 
-    // Create the dispatch last, after any setup of `gpu` is done
-    p->dp = pl_dispatch_create(vk->log, gpu);
     return pl_gpu_finalize(gpu);
 
 error:
@@ -627,7 +655,7 @@ static pl_sync vk_sync_create(pl_gpu gpu, enum pl_handle_type handle_type)
     struct pl_vk *p = PL_PRIV(gpu);
     struct vk_ctx *vk = p->vk;
 
-    struct pl_sync *sync = pl_zalloc_obj(NULL, sync, struct pl_sync_vk);
+    struct pl_sync_t *sync = pl_zalloc_obj(NULL, sync, struct pl_sync_vk);
     sync->handle_type = handle_type;
 
     struct pl_sync_vk *sync_vk = PL_PRIV(sync);
@@ -650,6 +678,8 @@ static pl_sync vk_sync_create(pl_gpu gpu, enum pl_handle_type handle_type)
         break;
     case PL_HANDLE_DMA_BUF:
     case PL_HANDLE_HOST_PTR:
+    case PL_HANDLE_MTL_TEX:
+    case PL_HANDLE_IOSURFACE:
         pl_unreachable();
     }
 
@@ -704,6 +734,113 @@ error:
     return NULL;
 }
 
+void pl_vulkan_sem_destroy(pl_gpu gpu, VkSemaphore *semaphore)
+{
+    VkSemaphore sem = *semaphore;
+    if (!sem)
+        return;
+
+    struct pl_vk *p = PL_PRIV(gpu);
+    struct vk_ctx *vk = p->vk;
+    vk->DestroySemaphore(vk->dev, sem, PL_VK_ALLOC);
+    *semaphore = VK_NULL_HANDLE;
+}
+
+VkSemaphore pl_vulkan_sem_create(pl_gpu gpu, const struct pl_vulkan_sem_params *params)
+{
+    struct pl_vk *p = PL_PRIV(gpu);
+    struct vk_ctx *vk = p->vk;
+
+    pl_assert(PL_ISPOT(params->export_handle));
+    if ((params->export_handle & gpu->export_caps.sync) != params->export_handle) {
+        PL_ERR(gpu, "Invalid handle type 0x%"PRIx64" specified for "
+               "`pl_vulkan_sem_create`!", (uint64_t) params->export_handle);
+        return VK_NULL_HANDLE;
+    }
+
+    switch (params->export_handle) {
+    case PL_HANDLE_FD:
+        params->out_handle->fd = -1;
+        break;
+    case PL_HANDLE_WIN32:
+    case PL_HANDLE_WIN32_KMT:
+        params->out_handle->handle = NULL;
+        break;
+    case PL_HANDLE_DMA_BUF:
+    case PL_HANDLE_HOST_PTR:
+    case PL_HANDLE_MTL_TEX:
+    case PL_HANDLE_IOSURFACE:
+        pl_unreachable();
+    }
+
+    const VkExportSemaphoreCreateInfoKHR einfo = {
+        .sType = VK_STRUCTURE_TYPE_EXPORT_SEMAPHORE_CREATE_INFO_KHR,
+        .handleTypes = vk_sync_handle_type(params->export_handle),
+    };
+
+    const VkSemaphoreTypeCreateInfo stinfo = {
+        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO,
+        .pNext = params->export_handle ? &einfo : NULL,
+        .semaphoreType = params->type,
+        .initialValue = params->initial_value,
+    };
+
+    const VkSemaphoreCreateInfo sinfo = {
+        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
+        .pNext = &stinfo,
+    };
+
+    VkSemaphore sem = VK_NULL_HANDLE;
+    VK(vk->CreateSemaphore(vk->dev, &sinfo, PL_VK_ALLOC, &sem));
+    PL_VK_NAME(SEMAPHORE, sem, PL_DEF(params->debug_tag, "pl_vulkan_sem"));
+
+#ifdef PL_HAVE_UNIX
+    if (params->export_handle == PL_HANDLE_FD) {
+        VkSemaphoreGetFdInfoKHR finfo = {
+            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_GET_FD_INFO_KHR,
+            .handleType = einfo.handleTypes,
+            .semaphore = sem,
+        };
+
+        VK(vk->GetSemaphoreFdKHR(vk->dev, &finfo, &params->out_handle->fd));
+    }
+#endif
+
+#ifdef PL_HAVE_WIN32
+    if (params->export_handle == PL_HANDLE_WIN32 ||
+        params->export_handle == PL_HANDLE_WIN32_KMT)
+    {
+        VkSemaphoreGetWin32HandleInfoKHR handle_info = {
+            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_GET_WIN32_HANDLE_INFO_KHR,
+            .handleType = einfo.handleTypes,
+            .semaphore = sem,
+        };
+
+        VK(vk->GetSemaphoreWin32HandleKHR(vk->dev, &handle_info,
+                                          &params->out_handle->handle));
+    }
+#endif
+
+    return sem;
+
+error:
+#ifdef PL_HAVE_UNIX
+    if (params->export_handle == PL_HANDLE_FD) {
+        if (params->out_handle->fd > -1)
+            close(params->out_handle->fd);
+    }
+#endif
+#ifdef PL_HAVE_WIN32
+    if (params->export_handle == PL_HANDLE_WIN32) {
+        if (params->out_handle->handle != NULL)
+            CloseHandle(params->out_handle->handle);
+    }
+    // PL_HANDLE_WIN32_KMT is just an identifier. It doesn't get closed.
+#endif
+    vk->DestroySemaphore(vk->dev, sem, PL_VK_ALLOC);
+    return VK_NULL_HANDLE;
+}
+
 static void vk_gpu_flush(pl_gpu gpu)
 {
     struct pl_vk *p = PL_PRIV(gpu);
