@@ -24,7 +24,6 @@ struct pl_pass_vk {
     VkPipeline base;
     VkPipeline pipe;
     VkPipelineLayout pipeLayout;
-    VkRenderPass renderPass;
     // Descriptor set (bindings)
     bool use_pushd;
     VkDescriptorSetLayout dsLayout;
@@ -61,7 +60,6 @@ static void pass_destroy_cb(pl_gpu gpu, pl_pass pass)
 
     vk->DestroyPipeline(vk->dev, pass_vk->pipe, PL_VK_ALLOC);
     vk->DestroyPipeline(vk->dev, pass_vk->base, PL_VK_ALLOC);
-    vk->DestroyRenderPass(vk->dev, pass_vk->renderPass, PL_VK_ALLOC);
     vk->DestroyPipelineLayout(vk->dev, pass_vk->pipeLayout, PL_VK_ALLOC);
     vk->DestroyPipelineCache(vk->dev, pass_vk->cache, PL_VK_ALLOC);
     vk->DestroyDescriptorPool(vk->dev, pass_vk->dsPool, PL_VK_ALLOC);
@@ -155,9 +153,9 @@ static VkResult vk_compile_glsl(pl_gpu gpu, void *alloc,
 {
     struct pl_vk *p = PL_PRIV(gpu);
 
-    clock_t start = clock();
+    pl_clock_t start = pl_clock_now();
     *out_spirv = spirv_compile_glsl(p->spirv, alloc, &gpu->glsl, stage, shader);
-    pl_log_cpu_time(gpu->log, start, clock(), "translating SPIR-V");
+    pl_log_cpu_time(gpu->log, start, pl_clock_now(), "translating SPIR-V");
 
     return out_spirv->len ? VK_SUCCESS : VK_ERROR_INITIALIZATION_FAILED;
 }
@@ -231,8 +229,14 @@ static VkResult vk_recreate_pipelines(struct vk_ctx *vk, pl_pass pass,
             [PL_PRIM_TRIANGLE_STRIP] = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_STRIP,
         };
 
+        const VkFormat target_fmt = (VkFormat) params->target_format->signature;
         VkGraphicsPipelineCreateInfo cinfo = {
             .sType = VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO,
+            .pNext = &(VkPipelineRenderingCreateInfo) {
+                .sType = VK_STRUCTURE_TYPE_PIPELINE_RENDERING_CREATE_INFO,
+                .colorAttachmentCount = 1,
+                .pColorAttachmentFormats = &target_fmt,
+            },
             .flags = flags,
             .stageCount = 2,
             .pStages = (VkPipelineShaderStageCreateInfo[]) {
@@ -293,7 +297,6 @@ static VkResult vk_recreate_pipelines(struct vk_ctx *vk, pl_pass pass,
                 },
             },
             .layout = pass_vk->pipeLayout,
-            .renderPass = pass_vk->renderPass,
             .basePipelineHandle = base,
             .basePipelineIndex = -1,
         };
@@ -348,10 +351,10 @@ pl_pass vk_pass_create(pl_gpu gpu, const struct pl_pass_params *params)
     int num_desc = params->num_descriptors;
     if (!num_desc)
         goto no_descriptors;
-    if (num_desc > vk->limits.maxPerStageResources) {
+    if (num_desc > vk->props.limits.maxPerStageResources) {
         PL_ERR(gpu, "Pass with %d descriptors exceeds the maximum number of "
                "per-stage resources %" PRIu32"!",
-               num_desc, vk->limits.maxPerStageResources);
+               num_desc, vk->props.limits.maxPerStageResources);
         goto error;
     }
 
@@ -364,10 +367,10 @@ pl_pass vk_pass_create(pl_gpu gpu, const struct pl_pass_params *params)
     int dsSize[PL_DESC_TYPE_COUNT] = {0};
     VkDescriptorSetLayoutBinding *bindings = pl_calloc_ptr(tmp, num_desc, bindings);
 
-    uint32_t max_tex = vk->limits.maxPerStageDescriptorSampledImages,
-             max_img = vk->limits.maxPerStageDescriptorStorageImages,
-             max_ubo = vk->limits.maxPerStageDescriptorUniformBuffers,
-             max_ssbo = vk->limits.maxPerStageDescriptorStorageBuffers;
+    uint32_t max_tex = vk->props.limits.maxPerStageDescriptorSampledImages,
+             max_img = vk->props.limits.maxPerStageDescriptorStorageImages,
+             max_ubo = vk->props.limits.maxPerStageDescriptorUniformBuffers,
+             max_ssbo = vk->props.limits.maxPerStageDescriptorStorageBuffers;
 
     uint32_t *dsLimits[PL_DESC_TYPE_COUNT] = {
         [PL_DESC_SAMPLED_TEX] = &max_tex,
@@ -536,7 +539,7 @@ no_descriptors: ;
         .sType = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO,
     };
 
-    clock_t start = clock();
+    pl_clock_t start = pl_clock_now();
     switch (params->type) {
     case PL_PASS_RASTER: {
         sinfo.pCode = (uint32_t *) vert.buf;
@@ -561,32 +564,6 @@ no_descriptors: ;
                 .format   = PL_DEF((*pfmt_vk)->bfmt, (*pfmt_vk)->tfmt),
             };
         }
-
-        VkRenderPassCreateInfo rinfo = {
-            .sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO,
-            .attachmentCount = 1,
-            .pAttachments = &(VkAttachmentDescription) {
-                .format = (VkFormat) params->target_format->signature,
-                .samples = VK_SAMPLE_COUNT_1_BIT,
-                .loadOp = pass->params.load_target
-                            ? VK_ATTACHMENT_LOAD_OP_LOAD
-                            : VK_ATTACHMENT_LOAD_OP_DONT_CARE,
-                .storeOp = VK_ATTACHMENT_STORE_OP_STORE,
-                .initialLayout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL,
-                .finalLayout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL,
-            },
-            .subpassCount = 1,
-            .pSubpasses = &(VkSubpassDescription) {
-                .pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS,
-                .colorAttachmentCount = 1,
-                .pColorAttachments = &(VkAttachmentReference) {
-                    .attachment = 0,
-                    .layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL,
-                },
-            },
-        };
-
-        VK(vk->CreateRenderPass(vk->dev, &rinfo, PL_VK_ALLOC, &pass_vk->renderPass));
         break;
     }
     case PL_PASS_COMPUTE: {
@@ -601,13 +578,13 @@ no_descriptors: ;
         pl_unreachable();
     }
 
-    clock_t after_compilation = clock();
+    pl_clock_t after_compilation = pl_clock_now();
     pl_log_cpu_time(gpu->log, start, after_compilation, "compiling shader");
 
     // Create the graphics/compute pipeline
     VkPipeline *pipe = has_spec ? &pass_vk->base : &pass_vk->pipe;
     VK(vk_recreate_pipelines(vk, pass, has_spec, VK_NULL_HANDLE, pipe));
-    pl_log_cpu_time(gpu->log, after_compilation, clock(), "creating pipeline");
+    pl_log_cpu_time(gpu->log, after_compilation, pl_clock_now(), "creating pipeline");
 
     if (!has_spec) {
         // We can free these if we no longer need them for specialization
@@ -664,10 +641,9 @@ error:
     return pass;
 }
 
-static const VkPipelineStageFlags passStages[] = {
-    [PL_PASS_RASTER]  = VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT |
-                        VK_PIPELINE_STAGE_VERTEX_INPUT_BIT,
-    [PL_PASS_COMPUTE] = VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
+static const VkPipelineStageFlags2 shaderStages[] = {
+    [PL_PASS_RASTER]  = VK_PIPELINE_STAGE_2_FRAGMENT_SHADER_BIT,
+    [PL_PASS_COMPUTE] = VK_PIPELINE_STAGE_2_COMPUTE_SHADER_BIT,
 };
 
 static void vk_update_descriptor(pl_gpu gpu, struct vk_cmd *cmd, pl_pass pass,
@@ -687,11 +663,11 @@ static void vk_update_descriptor(pl_gpu gpu, struct vk_cmd *cmd, pl_pass pass,
         .descriptorType = dsType[desc->type],
     };
 
-    static const VkAccessFlags access[PL_DESC_ACCESS_COUNT] = {
-        [PL_DESC_ACCESS_READONLY]   = VK_ACCESS_SHADER_READ_BIT,
-        [PL_DESC_ACCESS_WRITEONLY]  = VK_ACCESS_SHADER_WRITE_BIT,
-        [PL_DESC_ACCESS_READWRITE]  = VK_ACCESS_SHADER_READ_BIT |
-                                      VK_ACCESS_SHADER_WRITE_BIT,
+    static const VkAccessFlags2 storageAccess[PL_DESC_ACCESS_COUNT] = {
+        [PL_DESC_ACCESS_READONLY]   = VK_ACCESS_2_SHADER_STORAGE_READ_BIT,
+        [PL_DESC_ACCESS_WRITEONLY]  = VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT,
+        [PL_DESC_ACCESS_READWRITE]  = VK_ACCESS_2_SHADER_STORAGE_READ_BIT |
+                                      VK_ACCESS_2_SHADER_STORAGE_WRITE_BIT,
     };
 
     switch (desc->type) {
@@ -699,8 +675,8 @@ static void vk_update_descriptor(pl_gpu gpu, struct vk_cmd *cmd, pl_pass pass,
         pl_tex tex = db.object;
         struct pl_tex_vk *tex_vk = PL_PRIV(tex);
 
-        vk_tex_barrier(gpu, cmd, tex, passStages[pass->params.type],
-                       VK_ACCESS_SHADER_READ_BIT,
+        vk_tex_barrier(gpu, cmd, tex, shaderStages[pass->params.type],
+                       VK_ACCESS_2_SHADER_SAMPLED_READ_BIT,
                        VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
                        VK_QUEUE_FAMILY_IGNORED);
 
@@ -718,8 +694,8 @@ static void vk_update_descriptor(pl_gpu gpu, struct vk_cmd *cmd, pl_pass pass,
         pl_tex tex = db.object;
         struct pl_tex_vk *tex_vk = PL_PRIV(tex);
 
-        vk_tex_barrier(gpu, cmd, tex, passStages[pass->params.type],
-                       access[desc->access], VK_IMAGE_LAYOUT_GENERAL,
+        vk_tex_barrier(gpu, cmd, tex, shaderStages[pass->params.type],
+                       storageAccess[desc->access], VK_IMAGE_LAYOUT_GENERAL,
                        VK_QUEUE_FAMILY_IGNORED);
 
         VkDescriptorImageInfo *iinfo = &pass_vk->dsiinfo[idx];
@@ -736,8 +712,12 @@ static void vk_update_descriptor(pl_gpu gpu, struct vk_cmd *cmd, pl_pass pass,
         pl_buf buf = db.object;
         struct pl_buf_vk *buf_vk = PL_PRIV(buf);
 
-        vk_buf_barrier(gpu, cmd, buf, passStages[pass->params.type],
-                       access[desc->access], 0, buf->params.size, false);
+        VkAccessFlags2 access = VK_ACCESS_2_UNIFORM_READ_BIT;
+        if (desc->type == PL_DESC_BUF_STORAGE)
+            access = storageAccess[desc->access];
+
+        vk_buf_barrier(gpu, cmd, buf, shaderStages[pass->params.type],
+                       access, 0, buf->params.size, false);
 
         VkDescriptorBufferInfo *binfo = &pass_vk->dsbinfo[idx];
         *binfo = (VkDescriptorBufferInfo) {
@@ -754,8 +734,12 @@ static void vk_update_descriptor(pl_gpu gpu, struct vk_cmd *cmd, pl_pass pass,
         pl_buf buf = db.object;
         struct pl_buf_vk *buf_vk = PL_PRIV(buf);
 
-        vk_buf_barrier(gpu, cmd, buf, passStages[pass->params.type],
-                       access[desc->access], 0, buf->params.size, false);
+        VkAccessFlags2 access = VK_ACCESS_2_SHADER_SAMPLED_READ_BIT;
+        if (desc->type == PL_DESC_BUF_TEXEL_STORAGE)
+            access = storageAccess[desc->access];
+
+        vk_buf_barrier(gpu, cmd, buf, shaderStages[pass->params.type],
+                       access, 0, buf->params.size, false);
 
         wds->pTexelBufferView = &buf_vk->view;
         return;
@@ -835,9 +819,9 @@ void vk_pass_run(pl_gpu gpu, const struct pl_pass_run_params *params)
 
     // Check if we need to re-specialize this pipeline
     if (need_respec(pass, params)) {
-        clock_t start = clock();
+        pl_clock_t start = pl_clock_now();
         VK(vk_recreate_pipelines(vk, pass, false, pass_vk->base, &pass_vk->pipe));
-        pl_log_cpu_time(gpu->log, start, clock(), "re-specializing shader");
+        pl_log_cpu_time(gpu->log, start, pl_clock_now(), "re-specializing shader");
     }
 
     if (!pass_vk->use_pushd) {
@@ -921,20 +905,22 @@ void vk_pass_run(pl_gpu gpu, const struct pl_pass_run_params *params)
 
         // In the edge case that vert = index buffer, we need to synchronize
         // for both flags simultaneously
-        VkAccessFlags vbo_flags = VK_ACCESS_VERTEX_ATTRIBUTE_READ_BIT;
-        if (index == vert)
-            vbo_flags |= VK_ACCESS_INDEX_READ_BIT;
+        VkPipelineStageFlags2 vbo_stage = VK_PIPELINE_STAGE_2_VERTEX_ATTRIBUTE_INPUT_BIT;
+        VkAccessFlags2 vbo_flags = VK_ACCESS_2_VERTEX_ATTRIBUTE_READ_BIT;
+        if (index == vert) {
+            vbo_stage |= VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT;
+            vbo_flags |= VK_ACCESS_2_INDEX_READ_BIT;
+        }
 
-        vk_buf_barrier(gpu, cmd, vert, VK_PIPELINE_STAGE_VERTEX_INPUT_BIT,
-                       vbo_flags, 0, vert->params.size, false);
+        vk_buf_barrier(gpu, cmd, vert, vbo_stage, vbo_flags, 0, vert->params.size, false);
 
         VkDeviceSize offset = vert_vk->mem.offset + params->buf_offset;
         vk->CmdBindVertexBuffers(cmd->buf, 0, 1, &vert_vk->mem.buf, &offset);
 
         if (index) {
             if (index != vert) {
-                vk_buf_barrier(gpu, cmd, index, VK_PIPELINE_STAGE_VERTEX_INPUT_BIT,
-                               VK_ACCESS_INDEX_READ_BIT, 0, index->params.size,
+                vk_buf_barrier(gpu, cmd, index, VK_PIPELINE_STAGE_2_INDEX_INPUT_BIT,
+                               VK_ACCESS_2_INDEX_READ_BIT, 0, index->params.size,
                                false);
             }
 
@@ -949,11 +935,11 @@ void vk_pass_run(pl_gpu gpu, const struct pl_pass_run_params *params)
         }
 
 
-        VkAccessFlags fbo_access = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
+        VkAccessFlags2 fbo_access = VK_ACCESS_2_COLOR_ATTACHMENT_WRITE_BIT;
         if (pass->params.load_target)
-            fbo_access |= VK_ACCESS_COLOR_ATTACHMENT_READ_BIT;
+            fbo_access |= VK_ACCESS_2_COLOR_ATTACHMENT_READ_BIT;
 
-        vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT,
+        vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_2_COLOR_ATTACHMENT_OUTPUT_BIT,
                        fbo_access, VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL,
                        VK_QUEUE_FAMILY_IGNORED);
 
@@ -971,15 +957,21 @@ void vk_pass_run(pl_gpu gpu, const struct pl_pass_run_params *params)
 
         vk->CmdSetViewport(cmd->buf, 0, 1, &viewport);
         vk->CmdSetScissor(cmd->buf, 0, 1, &scissor);
-
-        VkRenderPassBeginInfo binfo = {
-            .sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO,
-            .renderPass = pass_vk->renderPass,
-            .framebuffer = tex_vk->framebuffer,
-            .renderArea.extent = {tex->params.w, tex->params.h},
-        };
-
-        vk->CmdBeginRenderPass(cmd->buf, &binfo, VK_SUBPASS_CONTENTS_INLINE);
+        vk->CmdBeginRenderingKHR(cmd->buf, &(VkRenderingInfo) {
+            .sType = VK_STRUCTURE_TYPE_RENDERING_INFO,
+            .renderArea.extent = { tex->params.w, tex->params.h },
+            .layerCount = 1,
+            .colorAttachmentCount = 1,
+            .pColorAttachments = &(VkRenderingAttachmentInfo) {
+                .sType = VK_STRUCTURE_TYPE_RENDERING_ATTACHMENT_INFO,
+                .imageView = tex_vk->view,
+                .imageLayout = tex_vk->layout,
+                .loadOp = pass->params.load_target
+                            ? VK_ATTACHMENT_LOAD_OP_LOAD
+                            : VK_ATTACHMENT_LOAD_OP_DONT_CARE,
+                .storeOp = VK_ATTACHMENT_STORE_OP_STORE,
+            },
+        });
 
         if (index) {
             vk->CmdDrawIndexed(cmd->buf, params->vertex_count, 1, 0, 0, 0);
@@ -987,10 +979,7 @@ void vk_pass_run(pl_gpu gpu, const struct pl_pass_run_params *params)
             vk->CmdDraw(cmd->buf, params->vertex_count, 1, 0, 0);
         }
 
-        vk->CmdEndRenderPass(cmd->buf);
-
-        // The renderPass implicitly transitions the texture to this layout
-        tex_vk->layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
+        vk->CmdEndRenderingKHR(cmd->buf);
         break;
     }
     case PL_PASS_COMPUTE:
