@@ -18,7 +18,7 @@
 #include "gpu.h"
 
 void vk_tex_barrier(pl_gpu gpu, struct vk_cmd *cmd, pl_tex tex,
-                    VkPipelineStageFlags stage, VkAccessFlags access,
+                    VkPipelineStageFlags2 stage, VkAccessFlags2 access,
                     VkImageLayout layout, uint32_t qf)
 {
     struct pl_vk *p = PL_PRIV(gpu);
@@ -41,14 +41,16 @@ void vk_tex_barrier(pl_gpu gpu, struct vk_cmd *cmd, pl_tex tex,
     bool is_trans = layout != tex_vk->layout, is_xfer = qf != tex_vk->qf;
     last = vk_sem_barrier(vk, cmd, &tex_vk->sem, stage, access, is_trans || is_xfer);
 
-    VkImageMemoryBarrier barr = {
-        .sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,
+    VkImageMemoryBarrier2 barr = {
+        .sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER_2,
+        .srcStageMask = last.stage,
+        .srcAccessMask = last.access,
+        .dstStageMask = stage,
+        .dstAccessMask = access,
         .oldLayout = tex_vk->layout,
         .newLayout = layout,
         .srcQueueFamilyIndex = tex_vk->qf,
         .dstQueueFamilyIndex = qf,
-        .srcAccessMask = last.access,
-        .dstAccessMask = access,
         .image = tex_vk->img,
         .subresourceRange = {
             .aspectMask = tex_vk->aspect,
@@ -63,8 +65,11 @@ void vk_tex_barrier(pl_gpu gpu, struct vk_cmd *cmd, pl_tex tex,
     }
 
     if (last.access || is_trans || is_xfer) {
-        vk->CmdPipelineBarrier(cmd->buf, last.stage, stage, 0, 0, NULL,
-                               0, NULL, 1, &barr);
+        vk->CmdPipelineBarrier2KHR(cmd->buf, &(VkDependencyInfo) {
+            .sType = VK_STRUCTURE_TYPE_DEPENDENCY_INFO,
+            .imageMemoryBarrierCount = 1,
+            .pImageMemoryBarriers = &barr,
+        });
     }
 
     tex_vk->qf = qf;
@@ -91,7 +96,6 @@ static void vk_tex_destroy(pl_gpu gpu, struct pl_tex_t *tex)
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
 
     vk_sync_deref(gpu, tex_vk->ext_sync);
-    vk->DestroyFramebuffer(vk->dev, tex_vk->framebuffer, PL_VK_ALLOC);
     vk->DestroyImageView(vk->dev, tex_vk->view, PL_VK_ALLOC);
     for (int i = 0; i < tex_vk->num_planes; i++)
         vk_tex_deref(gpu, tex->planes[i]);
@@ -114,7 +118,7 @@ void vk_tex_deref(pl_gpu gpu, pl_tex tex)
 }
 
 
-// Initializes non-VkImage values like the image view, framebuffers, etc.
+// Initializes non-VkImage values like the image view and metadata
 static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
 {
     struct pl_vk *p = PL_PRIV(gpu);
@@ -127,7 +131,6 @@ static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
     pl_rc_init(&tex_vk->rc);
     if (tex_vk->num_planes)
         return true;
-    vk_sem_init(&tex_vk->sem);
     tex_vk->layout = VK_IMAGE_LAYOUT_UNDEFINED;
     tex_vk->transfer_queue = GRAPHICS;
     tex_vk->qf = VK_QUEUE_FAMILY_IGNORED; // will be set on first use, if needed
@@ -141,9 +144,6 @@ static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
     if (params->format->emulated)
         tex_vk->transfer_queue = COMPUTE;
 
-    bool ret = false;
-    VkRenderPass dummyPass = VK_NULL_HANDLE;
-
     if (params->sampleable || params->renderable || params->storable) {
         static const VkImageViewType viewType[] = {
             [VK_IMAGE_TYPE_1D] = VK_IMAGE_VIEW_TYPE_1D,
@@ -167,64 +167,10 @@ static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
         PL_VK_NAME(IMAGE_VIEW, tex_vk->view, debug_tag);
     }
 
-    if (params->renderable) {
-        // Framebuffers need to be created against a specific render pass
-        // layout, so we need to temporarily create a skeleton/dummy render
-        // pass for vulkan to figure out the compatibility
-        VkRenderPassCreateInfo rinfo = {
-            .sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO,
-            .attachmentCount = 1,
-            .pAttachments = &(VkAttachmentDescription) {
-                .format = tex_vk->img_fmt,
-                .samples = VK_SAMPLE_COUNT_1_BIT,
-                .loadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE,
-                .storeOp = VK_ATTACHMENT_STORE_OP_STORE,
-                .initialLayout = VK_IMAGE_LAYOUT_UNDEFINED,
-                .finalLayout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL,
-            },
-            .subpassCount = 1,
-            .pSubpasses = &(VkSubpassDescription) {
-                .pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS,
-                .colorAttachmentCount = 1,
-                .pColorAttachments = &(VkAttachmentReference) {
-                    .attachment = 0,
-                    .layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL,
-                },
-            },
-        };
-
-        VK(vk->CreateRenderPass(vk->dev, &rinfo, PL_VK_ALLOC, &dummyPass));
-
-        VkFramebufferCreateInfo finfo = {
-            .sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO,
-            .renderPass = dummyPass,
-            .attachmentCount = 1,
-            .pAttachments = &tex_vk->view,
-            .width = tex->params.w,
-            .height = tex->params.h,
-            .layers = 1,
-        };
-
-        if (finfo.width > vk->limits.maxFramebufferWidth ||
-            finfo.height > vk->limits.maxFramebufferHeight)
-        {
-            PL_ERR(gpu, "Framebuffer of size %dx%d exceeds the maximum allowed "
-                   "dimensions: %dx%d", finfo.width, finfo.height,
-                   vk->limits.maxFramebufferWidth,
-                   vk->limits.maxFramebufferHeight);
-            goto error;
-        }
-
-        VK(vk->CreateFramebuffer(vk->dev, &finfo, PL_VK_ALLOC,
-                                 &tex_vk->framebuffer));
-        PL_VK_NAME(FRAMEBUFFER, tex_vk->framebuffer, debug_tag);
-    }
-
-    ret = true;
+    return true;
 
 error:
-    vk->DestroyRenderPass(vk->dev, dummyPass, PL_VK_ALLOC);
-    return ret;
+    return false;
 }
 
 pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
@@ -325,6 +271,7 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
     // will belong to, we're forced to share all of our images between all
     // command pools.
     uint32_t qfs[3] = {0};
+    pl_assert(vk->pools.num <= PL_ARRAY_SIZE(qfs));
     for (int i = 0; i < vk->pools.num; i++)
         qfs[i] = vk->pools.elem[i]->qf;
 
@@ -411,6 +358,7 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
 #endif
 
     if (params->export_handle == PL_HANDLE_DMA_BUF) {
+        pl_assert(drm_list.drmFormatModifierCount > 0);
         vk_link_struct(&iinfo, &drm_list);
         iinfo.tiling = VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT;
     }
@@ -418,7 +366,6 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
     // Double-check physical image format limits and fail if invalid
     VkPhysicalDeviceImageDrmFormatModifierInfoEXT drm_pinfo = {
         .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_DRM_FORMAT_MODIFIER_INFO_EXT,
-        .drmFormatModifier = drm_explicit.drmFormatModifier,
         .sharingMode = iinfo.sharingMode,
         .queueFamilyIndexCount = iinfo.queueFamilyIndexCount,
         .pQueueFamilyIndices = iinfo.pQueueFamilyIndices,
@@ -429,8 +376,19 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         .handleType = ext_info.handleTypes,
     };
 
-    if (handle_type == PL_HANDLE_DMA_BUF)
+    if (handle_type == PL_HANDLE_DMA_BUF) {
+        if (params->import_handle) {
+            // On import, we know exactly which format modifier to test
+            drm_pinfo.drmFormatModifier = drm_explicit.drmFormatModifier;
+        } else {
+            // On export, the choice of format modifier is ambiguous, because
+            // we offer the implementation a whole list to choose from. In
+            // principle, we must check *all* supported drm format modifiers,
+            // but in practice it should hopefully suffice to just check one
+            drm_pinfo.drmFormatModifier = drm_list.pDrmFormatModifiers[0];
+        }
         vk_link_struct(&ext_pinfo, &drm_pinfo);
+    }
 
     VkPhysicalDeviceImageFormatInfo2KHR pinfo = {
         .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_FORMAT_INFO_2_KHR,
@@ -675,8 +633,8 @@ void vk_tex_clear_ex(pl_gpu gpu, pl_tex tex, const union pl_clear_color color)
     if (!cmd)
         return;
 
-    vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                   VK_ACCESS_TRANSFER_WRITE_BIT,
+    vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_2_CLEAR_BIT,
+                   VK_ACCESS_2_TRANSFER_WRITE_BIT,
                    VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
                    VK_QUEUE_FAMILY_IGNORED);
 
@@ -708,7 +666,7 @@ void vk_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params)
     bool planar_fallback = src_vk->aspect != VK_IMAGE_ASPECT_COLOR_BIT ||
                            dst_vk->aspect != VK_IMAGE_ASPECT_COLOR_BIT;
 
-    struct pl_rect3d src_rc = params->src_rc, dst_rc = params->dst_rc;
+    pl_rect3d src_rc = params->src_rc, dst_rc = params->dst_rc;
     bool requires_scaling = !pl_rect3d_eq(src_rc, dst_rc);
     if ((requires_scaling && blit_emulated) || planar_fallback) {
         if (!pl_tex_blit_compute(gpu, params))
@@ -720,19 +678,19 @@ void vk_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params)
     if (!cmd)
         return;
 
-    vk_tex_barrier(gpu, cmd, params->src, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                   VK_ACCESS_TRANSFER_READ_BIT,
-                   VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
-                   VK_QUEUE_FAMILY_IGNORED);
-
-    vk_tex_barrier(gpu, cmd, params->dst, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                   VK_ACCESS_TRANSFER_WRITE_BIT,
-                   VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
-                   VK_QUEUE_FAMILY_IGNORED);
-
     // When the blit operation doesn't require scaling, we can use the more
     // efficient vkCmdCopyImage instead of vkCmdBlitImage
     if (!requires_scaling) {
+        vk_tex_barrier(gpu, cmd, params->src, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT,
+                       VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
+                       VK_QUEUE_FAMILY_IGNORED);
+
+        vk_tex_barrier(gpu, cmd, params->dst, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT,
+                       VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
+                       VK_QUEUE_FAMILY_IGNORED);
+
         pl_rect3d_normalize(&src_rc);
 
         VkImageCopy region = {
@@ -756,6 +714,16 @@ void vk_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params)
         vk->CmdCopyImage(cmd->buf, src_vk->img, src_vk->layout,
                          dst_vk->img, dst_vk->layout, 1, &region);
     } else {
+        vk_tex_barrier(gpu, cmd, params->src, VK_PIPELINE_STAGE_2_BLIT_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT,
+                       VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
+                       VK_QUEUE_FAMILY_IGNORED);
+
+        vk_tex_barrier(gpu, cmd, params->dst, VK_PIPELINE_STAGE_2_BLIT_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT,
+                       VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
+                       VK_QUEUE_FAMILY_IGNORED);
+
         VkImageBlit region = {
             .srcSubresource = {
                 .aspectMask = src_vk->aspect,
@@ -854,7 +822,7 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
 
     pl_buf buf = params->buf;
     struct pl_buf_vk *buf_vk = PL_PRIV(buf);
-    struct pl_rect3d rc = params->rc;
+    pl_rect3d rc = params->rc;
     size_t size = pl_tex_transfer_size(params);
 
     size_t buf_offset = buf_vk->mem.offset + params->buf_offset;
@@ -902,11 +870,11 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
             .size = size,
         };
 
-        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_READ_BIT, params->buf_offset, size,
+        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT, params->buf_offset, size,
                        false);
-        vk_buf_barrier(gpu, cmd, tbuf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_WRITE_BIT, 0, size, false);
+        vk_buf_barrier(gpu, cmd, tbuf, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT, 0, size, false);
         vk->CmdCopyBuffer(cmd->buf, buf_vk->mem.buf, tbuf_vk->mem.buf,
                           1, &region);
 
@@ -945,11 +913,11 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
         if (!cmd)
             goto error;
 
-        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_READ_BIT, params->buf_offset, size,
+        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT, params->buf_offset, size,
                        false);
-        vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_WRITE_BIT,
+        vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT,
                        VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
                        VK_QUEUE_FAMILY_IGNORED);
         vk->CmdCopyBufferToImage(cmd->buf, buf_vk->mem.buf, tex_vk->img,
@@ -980,7 +948,7 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
 
     pl_buf buf = params->buf;
     struct pl_buf_vk *buf_vk = PL_PRIV(buf);
-    struct pl_rect3d rc = params->rc;
+    pl_rect3d rc = params->rc;
     size_t size = pl_tex_transfer_size(params);
 
     size_t buf_offset = buf_vk->mem.offset + params->buf_offset;
@@ -1027,10 +995,10 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
             .size = size,
         };
 
-        vk_buf_barrier(gpu, cmd, tbuf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_READ_BIT, 0, size, false);
-        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_WRITE_BIT, params->buf_offset, size,
+        vk_buf_barrier(gpu, cmd, tbuf, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT, 0, size, false);
+        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT, params->buf_offset, size,
                        false);
         vk->CmdCopyBuffer(cmd->buf, tbuf_vk->mem.buf, buf_vk->mem.buf,
                           1, &region);
@@ -1064,11 +1032,11 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
         if (!cmd)
             goto error;
 
-        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_WRITE_BIT, params->buf_offset, size,
+        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT, params->buf_offset, size,
                        false);
-        vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_READ_BIT,
+        vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT,
                        VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
                        VK_QUEUE_FAMILY_IGNORED);
         vk->CmdCopyImageToBuffer(cmd->buf, tex_vk->img, tex_vk->layout,
@@ -1131,13 +1099,13 @@ bool vk_tex_export(pl_gpu gpu, pl_tex tex, pl_sync sync)
     if (!cmd)
         goto error;
 
-    vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
+    vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_2_NONE,
                    0, VK_IMAGE_LAYOUT_GENERAL, VK_QUEUE_FAMILY_EXTERNAL);
 
     // Make the next barrier appear as though coming from a different queue
     tex_vk->sem.write.queue = tex_vk->sem.read.queue = NULL;
 
-    vk_cmd_sig(cmd, (pl_vulkan_sem){ sync_vk->wait });
+    vk_cmd_sig(cmd, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, (pl_vulkan_sem){ sync_vk->wait });
     if (!CMD_SUBMIT(&cmd))
         goto error;
 
@@ -1229,7 +1197,13 @@ pl_tex pl_vulkan_wrap(pl_gpu gpu, const struct pl_vulkan_wrap_params *params)
     tex_vk->img_fmt = params->format;
     tex_vk->num_planes = fmt->num_planes;
     tex_vk->usage_flags = usage;
-    tex_vk->aspect = PL_DEF(params->aspect, VK_IMAGE_ASPECT_COLOR_BIT);
+    tex_vk->aspect = params->aspect;
+
+    if (!tex_vk->aspect) {
+        for (int i = 0; i < tex_vk->num_planes; i++)
+            tex_vk->aspect |= VK_IMAGE_ASPECT_PLANE_0_BIT << i;
+        tex_vk->aspect = PL_DEF(tex_vk->aspect, VK_IMAGE_ASPECT_COLOR_BIT);
+    }
 
     // Blitting to planar images requires fallback via compute shaders
     if (tex_vk->aspect != VK_IMAGE_ASPECT_COLOR_BIT) {
@@ -1244,10 +1218,10 @@ pl_tex pl_vulkan_wrap(pl_gpu gpu, const struct pl_vulkan_wrap_params *params)
     for (int i = 0; i < tex_vk->num_planes; i++) {
         struct pl_tex_t *plane;
         VkImageAspectFlags aspect = VK_IMAGE_ASPECT_PLANE_0_BIT << i;
-        if (!(aspect & params->aspect)) {
+        if (!(aspect & tex_vk->aspect)) {
             PL_INFO(gpu, "Not wrapping plane %d due to aspect bit 0x%x not "
                     "being contained in supplied params->aspect 0x%x!",
-                    i, (unsigned) aspect, (unsigned) params->aspect);
+                    i, (unsigned) aspect, (unsigned) tex_vk->aspect);
             continue;
         }
 
@@ -1327,19 +1301,17 @@ bool pl_vulkan_hold_ex(pl_gpu gpu, const struct pl_vulkan_hold_params *params)
     bool may_invalidate = true;
     if (!tex_vk->num_planes) {
         may_invalidate &= tex_vk->may_invalidate;
-        vk_tex_barrier(gpu, cmd, params->tex,
-                       VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
+        vk_tex_barrier(gpu, cmd, params->tex, VK_PIPELINE_STAGE_2_NONE,
                        0, layout, params->qf);
     }
 
     for (int i = 0; i < tex_vk->num_planes; i++) {
         may_invalidate &= tex_vk->planes[i]->may_invalidate;
         vk_tex_barrier(gpu, cmd, params->tex->planes[i],
-                       VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
-                       0, layout, params->qf);
+                       VK_PIPELINE_STAGE_2_NONE, 0, layout, params->qf);
     }
 
-    vk_cmd_sig(cmd, params->semaphore);
+    vk_cmd_sig(cmd, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, params->semaphore);
     bool ok = CMD_SUBMIT(&cmd);
 
     if (!tex_vk->num_planes) {
@@ -1383,3 +1355,37 @@ void pl_vulkan_release_ex(pl_gpu gpu, const struct pl_vulkan_release_params *par
     tex_vk->layout = params->layout;
     tex_vk->held = false;
 }
+
+bool pl_vulkan_hold(pl_gpu gpu, pl_tex tex, VkImageLayout layout,
+                    pl_vulkan_sem sem_out)
+{
+    return pl_vulkan_hold_ex(gpu, pl_vulkan_hold_params(
+        .tex        = tex,
+        .layout     = layout,
+        .semaphore  = sem_out,
+        .qf         = VK_QUEUE_FAMILY_IGNORED,
+    ));
+}
+
+bool pl_vulkan_hold_raw(pl_gpu gpu, pl_tex tex,
+                        VkImageLayout *out_layout,
+                        pl_vulkan_sem sem_out)
+{
+    return pl_vulkan_hold_ex(gpu, pl_vulkan_hold_params(
+        .tex        = tex,
+        .out_layout = out_layout,
+        .semaphore  = sem_out,
+        .qf         = VK_QUEUE_FAMILY_IGNORED,
+    ));
+}
+
+void pl_vulkan_release(pl_gpu gpu, pl_tex tex, VkImageLayout layout,
+                       pl_vulkan_sem sem_in)
+{
+    pl_vulkan_release_ex(gpu, pl_vulkan_release_params(
+        .tex        = tex,
+        .layout     = layout,
+        .semaphore  = sem_in,
+        .qf         = VK_QUEUE_FAMILY_IGNORED,
+    ));
+}
