@@ -38,8 +38,10 @@
 #define MAXIMUM_PAGE_COUNT (sizeof(uint64_t) * 8)
 
 // Controls the maximum page size. Any allocations above this threshold
-// will be served by dedicated allocations. (Default: 64 MB)
-#define MAXIMUM_PAGE_SIZE (1LLU << 26)
+// (absolute size or fraction of VRAM, whichever is higher) will be served by
+// dedicated allocations. (Default: 64 MB or 1/16 of VRAM)
+#define MAXIMUM_PAGE_SIZE_ABSOLUTE (1LLU << 26)
+#define MAXIMUM_PAGE_SIZE_RELATIVE 16
 
 // Controls the minimum slab size, to avoid excessive re-allocation of very
 // small slabs. (Default: 256 KB)
@@ -58,6 +60,7 @@
 // each of which contains a list of slabs of differing page sizes.
 struct vk_slab {
     pl_mutex lock;
+    pl_debug_tag debug_tag; // debug tag of the triggering allocation
     VkDeviceMemory mem;     // underlying device allocation
     VkDeviceSize size;      // total allocated size of `mem`
     VkMemoryType mtype;     // underlying memory type
@@ -97,6 +100,7 @@ struct vk_malloc {
     struct vk_ctx *vk;
     pl_mutex lock;
     VkPhysicalDeviceMemoryProperties props;
+    size_t maximum_page_size;
     PL_ARRAY(struct vk_pool) pools;
     uint64_t age;
 };
@@ -174,11 +178,12 @@ void vk_malloc_print_stats(struct vk_malloc *ma, enum pl_log_level lev)
             size_t slab_res = slab->size - avail;
 
             PL_MSG(vk, lev, "    Slab %2d: %8"PRIx64" x %s: "
-                   "%s used %s res %s alloc from heap %d, efficiency %.2f%%",
+                   "%s used %s res %s alloc from heap %d, efficiency %.2f%%  [%s]",
                    j, slab->spacemap, PRINT_SIZE(slab->pagesize),
                    PRINT_SIZE(slab->used), PRINT_SIZE(slab_res),
                    PRINT_SIZE(slab->size), (int) slab->mtype.heapIndex,
-                   efficiency(slab->used, slab_res));
+                   efficiency(slab->used, slab_res),
+                   PL_DEF(slab->debug_tag, "unknown"));
 
             pool_size += slab->size;
             pool_used += slab->used;
@@ -199,10 +204,11 @@ void vk_malloc_print_stats(struct vk_malloc *ma, enum pl_log_level lev)
     pl_mutex_unlock(&ma->lock);
 
     PL_MSG(vk, lev, "Memory summary: %s used %s res %s alloc, "
-           "efficiency %.2f%%, utilization %.2f%%",
+           "efficiency %.2f%%, utilization %.2f%%, max page: %s",
            PRINT_SIZE(total_used), PRINT_SIZE(total_res),
            PRINT_SIZE(total_size), efficiency(total_used, total_res),
-           efficiency(total_res, total_size));
+           efficiency(total_res, total_size),
+           PRINT_SIZE(ma->maximum_page_size));
 }
 
 static void slab_free(struct vk_ctx *vk, struct vk_slab *slab)
@@ -212,10 +218,14 @@ static void slab_free(struct vk_ctx *vk, struct vk_slab *slab)
 
 #ifndef NDEBUG
     if (!slab->dedicated && slab->used > 0) {
-        fprintf(stderr, "!!! libplacebo: leaked %zu bytes of vulkan memory\n"
-                "!!! slab total size: %zu bytes, heap: %d, flags: 0x%"PRIX64"\n",
-                slab->used, (size_t) slab->size, (int) slab->mtype.heapIndex,
+        PL_WARN(vk, "Leaked %zu bytes of vulkan memory!", slab->used);
+        PL_WARN(vk, "slab total size: %zu bytes, heap: %d, flags: 0x%"PRIX64,
+                (size_t) slab->size, (int) slab->mtype.heapIndex,
                 (uint64_t) slab->mtype.propertyFlags);
+        if (slab->debug_tag)
+            PL_WARN(vk, "last used for: %s", slab->debug_tag);
+        pl_log_stack_trace(vk->log, PL_LOG_WARN);
+        pl_debug_abort();
     }
 #endif
 
@@ -361,6 +371,7 @@ static struct vk_slab *slab_alloc(struct vk_malloc *ma,
         .age = ma->age,
         .size = params->reqs.size,
         .handle_type = params->export_handle,
+        .debug_tag = params->debug_tag,
     };
     pl_mutex_init(&slab->lock);
 
@@ -390,6 +401,7 @@ static struct vk_slab *slab_alloc(struct vk_malloc *ma,
         // Queue family sharing modes don't matter for buffers, so we just
         // set them as concurrent and stop worrying about it.
         uint32_t qfs[3] = {0};
+        pl_assert(vk->pools.num <= PL_ARRAY_SIZE(qfs));
         for (int i = 0; i < vk->pools.num; i++)
             qfs[i] = vk->pools.elem[i]->qf;
 
@@ -447,9 +459,10 @@ static struct vk_slab *slab_alloc(struct vk_malloc *ma,
         goto error;
 
     const VkMemoryType *mtype = &ma->props.memoryTypes[minfo.memoryTypeIndex];
-    PL_DEBUG(vk, "Allocating %zu memory of type 0x%x (id %d) in heap %d",
+    PL_DEBUG(vk, "Allocating %zu memory of type 0x%x (id %d) in heap %d: %s",
              (size_t) slab->size, (unsigned) mtype->propertyFlags,
-             (int) minfo.memoryTypeIndex, (int) mtype->heapIndex);
+             (int) minfo.memoryTypeIndex, (int) mtype->heapIndex,
+             PL_DEF(params->debug_tag, "unknown"));
 
     VkResult res = vk->AllocateMemory(vk->dev, &minfo, PL_VK_ALLOC, &slab->mem);
     switch (res) {
@@ -507,6 +520,8 @@ static struct vk_slab *slab_alloc(struct vk_malloc *ma,
     return slab;
 
 error:
+    if (params->debug_tag)
+        PL_ERR(vk, "  for malloc: %s", params->debug_tag);
     slab_free(vk, slab);
     return NULL;
 }
@@ -527,6 +542,16 @@ struct vk_malloc *vk_malloc_create(struct vk_ctx *vk)
     vk->GetPhysicalDeviceMemoryProperties(vk->physd, &ma->props);
     ma->vk = vk;
 
+    // Determine maximum page size
+    ma->maximum_page_size = MAXIMUM_PAGE_SIZE_ABSOLUTE;
+    for (int i = 0; i < ma->props.memoryHeapCount; i++) {
+        VkMemoryHeap heap = ma->props.memoryHeaps[i];
+        if (heap.flags & VK_MEMORY_HEAP_DEVICE_LOCAL_BIT) {
+            size_t size_max = heap.size / MAXIMUM_PAGE_SIZE_RELATIVE;
+            ma->maximum_page_size = PL_MAX(ma->maximum_page_size, size_max);
+        }
+    }
+
     vk_malloc_print_stats(ma, PL_LOG_INFO);
     return ma;
 }
@@ -537,6 +562,7 @@ void vk_malloc_destroy(struct vk_malloc **ma_ptr)
     if (!ma)
         return;
 
+    vk_malloc_print_stats(ma, PL_LOG_DEBUG);
     for (int i = 0; i < ma->pools.num; i++)
         pool_uninit(ma->vk, &ma->pools.elem[i]);
 
@@ -685,7 +711,6 @@ static struct vk_slab *pool_get_page(struct vk_malloc *ma, struct vk_pool *pool,
         }
 
         slab->spacemap ^= 0x1LLU << page_idx;
-        pl_mutex_unlock(&slab->lock);
         *offset = page_idx * slab->pagesize;
         return slab;
     }
@@ -757,6 +782,7 @@ static bool vk_malloc_import(struct vk_malloc *ma, struct vk_memslice *out,
 
     if (params->buf_usage) {
         uint32_t qfs[3] = {0};
+        pl_assert(vk->pools.num <= PL_ARRAY_SIZE(qfs));
         for (int i = 0; i < vk->pools.num; i++)
             qfs[i] = vk->pools.elem[i]->qf;
 
@@ -936,6 +962,8 @@ static bool vk_malloc_import(struct vk_malloc *ma, struct vk_memslice *out,
     return true;
 
 error:
+    if (params->debug_tag)
+        PL_ERR(vk, "  for malloc: %s", params->debug_tag);
     vk->DestroyBuffer(vk->dev, buffer, PL_VK_ALLOC);
 #ifdef PL_HAVE_UNIX
     if (fdinfo.fd > -1)
@@ -946,6 +974,19 @@ error:
     return false;
 }
 
+size_t vk_malloc_avail(struct vk_malloc *ma, VkMemoryPropertyFlags flags)
+{
+    size_t avail = 0;
+    for (int i = 0; i < ma->props.memoryTypeCount; i++) {
+        const VkMemoryType *mtype = &ma->props.memoryTypes[i];
+        if ((mtype->propertyFlags & flags) != flags)
+            continue;
+        avail = PL_MAX(avail, ma->props.memoryHeaps[mtype->heapIndex].size);
+    }
+
+    return avail;
+}
+
 bool vk_malloc_slice(struct vk_malloc *ma, struct vk_memslice *out,
                      const struct vk_malloc_params *params)
 {
@@ -957,13 +998,13 @@ bool vk_malloc_slice(struct vk_malloc *ma, struct vk_memslice *out,
     pl_assert(params->reqs.size);
     size_t size = params->reqs.size;
     size_t align = params->reqs.alignment;
-    align = pl_lcm(align, vk->limits.bufferImageGranularity);
-    align = pl_lcm(align, vk->limits.nonCoherentAtomSize);
+    align = pl_lcm(align, vk->props.limits.bufferImageGranularity);
+    align = pl_lcm(align, vk->props.limits.nonCoherentAtomSize);
 
     struct vk_slab *slab;
     VkDeviceSize offset;
 
-    if (params->ded_image || size > MAXIMUM_PAGE_SIZE) {
+    if (params->ded_image || size > ma->maximum_page_size) {
         slab = slab_alloc(ma, params);
         if (!slab)
             return false;
@@ -987,6 +1028,8 @@ bool vk_malloc_slice(struct vk_malloc *ma, struct vk_memslice *out,
         size = PL_ALIGN(size, align);
         slab->used += size;
         slab->age = ma->age;
+        if (params->debug_tag)
+            slab->debug_tag = params->debug_tag;
         pl_mutex_unlock(&slab->lock);
     }
 
