@@ -69,7 +69,6 @@ static bool map_color_space(VkColorSpaceKHR space, struct pl_color_space *out)
         *out = pl_color_space_monitor;
         return true;
 
-#ifdef VK_EXT_swapchain_colorspace
     case VK_COLOR_SPACE_BT709_NONLINEAR_EXT:
         *out = pl_color_space_monitor;
         return true;
@@ -137,7 +136,6 @@ static bool map_color_space(VkColorSpaceKHR space, struct pl_color_space *out)
     case VK_COLOR_SPACE_PASS_THROUGH_EXT:
         *out = pl_color_space_unknown;
         return true;
-#endif
 
 #ifdef VK_AMD_display_native_hdr
     case VK_COLOR_SPACE_DISPLAY_NATIVE_AMD:
@@ -600,6 +598,13 @@ static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
              sinfo.imageExtent.width,
              sinfo.imageExtent.height);
 
+#ifdef PL_HAVE_UNIX
+    if (vk->props.vendorID == VK_VENDOR_ID_NVIDIA) {
+        vk->DeviceWaitIdle(vk->dev);
+        vk_wait_idle(vk);
+    }
+#endif
+
     // Calling `vkCreateSwapchainKHR` puts sinfo.oldSwapchain into a retired
     // state whether the call succeeds or not, so we always need to garbage
     // collect it afterwards - asynchronously as it may still be in use
